#!BPY

"""
Name: 'Ultimate 3D model file export (*.u3d)'
Blender: 249b
Group: 'Export'
Tip: 'Exports models (geometry, materials and animations) to the Ultimate 3D model file format (*.u3d).'
"""

__author__="Christoph Peters";
__url__="http://Ultimate3D.org";
__version__="2.0.5";
__bpydoc__="""Exports geometry, materials and armatures with animations and constraints of models in Blender to the Ultimate 3D model file format.""";

#GNU-GPL-START
# ----------------------------- LICENSE -----------------------------
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#GNU-GPL-END


# Import the export module

# export_u3d_shared START
# This source code has been inserted into this module from the 
# module export_u3d_shared
# Import the exporting utility module, which collects the needed data from 
# Blender

# export_utility START
# This source code has been inserted into this module from the 
# module export_utility
# Include the needed components of the Blender API

# export_utility_shared START
# This source code has been inserted into this module from the 
# module export_utility_shared
# Defined in Blender version specific modules:
# def CleanPath(Path):
#     """This little utility function brings a path into a normalized form by removing
#        any redundant use of ".." or ".". The result is returned. As directory
#        separator it uses "/". E.g.:
#        CleanPath("c:\\Test/..\\Abc")=="c:/Abc" and
#        CleanPath("Directory/NoDirectory\\..")=="Directory/" """
# Defined in Blender version specific modules:
# def MakeRelativePath(Path,StartPath):
#     """This little utility function makes the given path relative to StartPath. The
#        resulting path is cleaned automatically (see CleanPath(...))."""
# Defined in Blender version specific modules:
# def MakeAbsolutePath(Path):
#     """This function makes the given path absolute and cleans it. If it is relative 
#        it is made absolute using the directory of the current Blend file."""

# Defined in Blender version specific modules:
# def GetBlenderVersion():
#     """This function returns a human readable string giving the used version of 
#        Blender (e.g. "2.49" or "2.58a")."""

# Defined in Blender version specific modules:
# def GetCurrentBlendFile():
#     """This function returns the absolute path and filename of the *.blend file, 
#        which is currently opened in the Blender instance, which executes this 
#        script or an empty string, if no *.blend file is opened."""


def Choose(Condition,Value1,Value2):
    """This function returns Value1, if the given Condition evaluates true or
    Value2 otherwise."""
    if(Condition):
        return Value1;
    else:
        return Value2;


def Lerp(Value1,Value2,LerpFactor):
    """This function performs a simple linear interpolation between Value1 and Value2
       using LerpFactor and returns the result. At 0.0 it returns Value1, at 1.0 it
       returns Value2."""
    return Value1+LerpFactor*(Value2-Value1);


class C2DVector:
    """This class represents simple 2D vectors"""

    def __init__(self,x,y):
        """Constructs the vector from two components"""
        self.x=x;
        self.y=y;

    def Copy(self,RHS):
        """This function makes self become a deep-copy of RHS where RHS has to
           support __getitem__(...) for 0, and 1."""
        self.x=RHS[0];
        self.y=RHS[1];

    def Clone(self):
        """This function returns a deep-copy of this object."""
        return C2DVector(self.x,self.y);

    def __getitem__(self,iComponent):
        """This function returns the component of this vector with the given integer
           index. This means self[0]==x and self[1]==y."""
        if(iComponent==0):
            return self.x;
        elif(iComponent==1):
            return self.y;
        else:
            raise IndexError;

    def __setitem__(self,iComponent,NewValue):
        """This function sets the component of this vector with the given integer
           index to the given value."""
        if(iComponent==0):
            self.x=NewValue;
        elif(iComponent==1):
            self.y=NewValue;
        else:
            raise IndexError;

    def __str__(self):
        """This function returns a string giving the current values of this vector in 
           an easily readable form."""
        return "x="+str(self.x)+" y="+str(self.y);

    def __eq__(self,RHS):
        """This operator returns true, if and only if the given vector is 
           identical to this vector in every component."""
        return isinstance(RHS,C2DVector) and self.x==RHS.x and self.y==RHS.y;

    def __add__(self,RHS):
        """This operator implements component-wise vector addition. RHS has to
           support __getitem__ for integer values 0 and 1, otherwise the function 
           fails with an exception."""
        return C2DVector(self.x+RHS[0],self.y+RHS[1]);

    def __sub__(self,RHS):
        """This operator implements component-wise vector subtraction. RHS has to
           support __getitem__ for integer values 0 and 1, otherwise the function
           fails with an exception."""
        return C2DVector(self.x-RHS[0],self.y-RHS[1]);

    def __rmul__(self,LHS):
        """This operator returns the result of the component-wise multiplication of
           this vector by the given left-hand-side scalar."""
        return C2DVector(LHS*self.x,LHS*self.y);

    def __iadd__(self,RHS):
        """This additon-assignment operator assigns self+RHS to self and returns
           self. RHS has to be a vector-type with __getitem__(...)."""
        self.Copy(self+RHS);
        return self;

    def __isub__(self,RHS):
        """This subtraction-assignment operator assigns self-RHS to self and returns
           self. RHS has to be a vector-type with __getitem__(...)."""
        self.Copy(self-RHS);
        return self;

    def __neg__(self):
        """This operator returns the additive inverse of this vector. This means that 
           it returns a copy of this vector, where the sign of every component is
           negated."""
        return C2DVector(-self.x,-self.y);

    x=0.0;
    y=0.0;

def MakeC2DVector(Vector):
    """This function takes an arbitrary object, which behaves like a tuple of two  
       floats, constructs a C2DVector from it and returns the result. It can be used 
       to cast Blender vectors to C2DVector."""
    return C2DVector(Vector[0],Vector[1]);


class C3DVector:
    """This class represents simple 3D vectors"""

    def __init__(self,x,y,z):
        """Constructs the vector from three components"""
        self.x=x;
        self.y=y;
        self.z=z;

    def Copy(self,RHS):
        """This function makes self become a deep-copy of RHS where RHS has to
           support __getitem__(...) for 0, 1 and 2."""
        self.x=RHS[0];
        self.y=RHS[1];
        self.z=RHS[2];

    def Clone(self):
        """This function returns a deep-copy of this object."""
        return C3DVector(self.x,self.y,self.z);

    def Normalize(self):
        """Normalizes this vector"""
        Length=(self.x*self.x+self.y*self.y+self.z*self.z)**0.5;
        if(Length>0.0):
            InvLength=1.0/Length;
            self.x*=InvLength;
            self.y*=InvLength;
            self.z*=InvLength;

    def Minimize(self,RHS):
        """This function chooses the minimum of this vector and the given other
           vector on per-component base."""
        self.x=min(self.x,RHS.x);
        self.y=min(self.y,RHS.y);
        self.z=min(self.z,RHS.z);
    def Maximize(self,RHS):
        """This function chooses the maximum of this vector and the given other
           vector on per-component base."""
        self.x=max(self.x,RHS.x);
        self.y=max(self.y,RHS.y);
        self.z=max(self.z,RHS.z);

    def __getitem__(self,iComponent):
        """This function returns the component of this vector with the given integer
           index. This means self[0]==x, self[1]==y and self[2]==z."""
        if(iComponent==0):
            return self.x;
        elif(iComponent==1):
            return self.y;
        elif(iComponent==2):
            return self.z;
        else:
            raise IndexError;

    def __setitem__(self,iComponent,NewValue):
        """This function sets the component of this vector with the given integer
           index to the given value."""
        if(iComponent==0):
            self.x=NewValue;
        elif(iComponent==1):
            self.y=NewValue;
        elif(iComponent==2):
            self.z=NewValue;
        else:
            raise IndexError;

    def __str__(self):
        """This function returns a string giving the current values of this vector in 
           an easily readable form."""
        return "x="+str(self.x)+" y="+str(self.y)+" z="+str(self.z);
        
    def __eq__(self,RHS):
        """This operator returns true, if and only if the given vector is 
           identical to this vector in every component."""
        return isinstance(RHS,C3DVector) and self.x==RHS.x and self.y==RHS.y and self.z==RHS.z;

    def __add__(self,RHS):
        """This operator implements component-wise vector addition. RHS has to
           support __getitem__ for integer values from 0 to 2, otherwise the function
           fails with an exception."""
        return C3DVector(self.x+RHS[0],self.y+RHS[1],self.z+RHS[2]);

    def __sub__(self,RHS):
        """This operator implements component-wise vector subtraction. RHS has to
           support __getitem__ for integer values from 0 to 2, otherwise the function
           fails with an exception."""
        return C3DVector(self.x-RHS[0],self.y-RHS[1],self.z-RHS[2]);

    def __rmul__(self,LHS):
        """This operator returns the result of the component-wise multiplication of
           this vector by the given left-hand-side scalar."""
        return C3DVector(LHS*self.x,LHS*self.y,LHS*self.z);

    def __iadd__(self,RHS):
        """This additon-assignment operator assigns self+RHS to self and returns
           self. RHS has to be a vector-type with __getitem__(...)."""
        self.Copy(self+RHS);
        return self;

    def __isub__(self,RHS):
        """This subtraction-assignment operator assigns self-RHS to self and returns
           self. RHS has to be a vector-type with __getitem__(...)."""
        self.Copy(self-RHS);
        return self;

    def __neg__(self):
        """This operator returns the additive inverse of this vector. This means that 
           it returns a copy of this vector, where the sign of every component is
           negated."""
        return C3DVector(-self.x,-self.y,-self.z);

    x=0.0;
    y=0.0;
    z=0.0;


def MakeC3DVector(Vector):
    """This function takes an arbitrary object, which behaves like a tuple of three 
       floats, constructs a C3DVector from it and returns the result. It can be used 
       to cast Blender vectors to C3DVector."""
    return C3DVector(Vector[0],Vector[1],Vector[2]);


class CQuaternion:
    """This class handles a quaternion describing a rotation."""

    def __init__(self,x=0.0,y=0.0,z=0.0,w=1.0):
        """This constructor initializes this quaternion with the given entries
           and normalizes it. If x, y and z are zero an identity rotation is
           created."""
        LengthSq=x*x+y*y+z*z+w*w;
        if(LengthSq>0.0):
            InvLength=(LengthSq)**(-0.5);
            self.x=x*InvLength;
            self.y=y*InvLength;
            self.z=z*InvLength;
            self.w=w*InvLength;
        else:
            self.x=0.0;
            self.y=0.0;
            self.z=0.0;
            self.w=1.0;

    def Copy(self,RHS):
        """This function makes self become a deep-copy of RHS where RHS has to be a 
           quaternion type with members x, y, z and w as float."""
        self.x=RHS.x;
        self.y=RHS.y;
        self.z=RHS.z;
        self.w=RHS.w;

    def Clone(self):
        """This function returns a deep-copy of this object."""
        return CQuaternion(self.x,self.y,self.z,self.w);

    def __getitem__(self,iComponent):
        """This function returns the component of this quaternion with the 
           given integer index (according to alphabetical order). This means 
           self[0]==w, self[1]==x, self[2]==y and self[3]==z."""
        if(iComponent==0):
            return self.w;
        elif(iComponent==1):
            return self.x;
        elif(iComponent==2):
            return self.y;
        elif(iComponent==3):
            return self.z;
        else:
            raise IndexError;

    def __setitem__(self,iComponent,NewValue):
        """This function sets the component of this vector with the given integer
           index to the given value (see __getitem__(...))."""
        if(iComponent==0):
            self.w=NewValue;
        elif(iComponent==1):
            self.x=NewValue;
        elif(iComponent==2):
            self.y=NewValue;
        elif(iComponent==3):
            self.z=NewValue;
        else:
            raise IndexError;

    def AppendRotation(self,RHS):
        """This function multiplies this quaternion by the given quaternion.
           Transforming a vector with the resulting quaternion will be
           equivalent to transforming it first with this quaternion and after
           that with the given quaternion."""
        self*=RHS;
        
    # def MakeRotationAlongAxis(self,Axis,Angle)
    # Defined in non-shared module.

    # def Slerp(self,Q2,Factor)
    # Defined in non-shared module.

    def RotateVector(self,VectorX,VectorY,VectorZ):
        """This function returns a version of the given vector, to which the
           rotation described by this vector has been applied."""
        return C3DVector(
            2.0*(VectorX*(0.5-(self.y*self.y+self.z*self.z))+VectorY*(self.x*self.y-self.z*self.w)+VectorZ*(self.x*self.z+self.y*self.w)),
            2.0*(VectorX*(self.x*self.y+self.z*self.w)+VectorY*(0.5-(self.x*self.x+self.z*self.z))+VectorZ*(self.y*self.z-self.x*self.w)),
            2.0*(VectorX*(self.x*self.z-self.y*self.w)+VectorY*(self.y*self.z+self.x*self.w)+VectorZ*(0.5-(self.x*self.x+self.y*self.y)))
    );

    def ToMatrix(self):
        """This function returns a CMatrix4x4 giving the rotation described by this
           quaternion."""
        xx=self.x*self.x;
        xy=self.x*self.y;
        xz=self.x*self.z;
        xw=self.x*self.w;
        yy=self.y*self.y;
        yz=self.y*self.z;
        yw=self.y*self.w;
        zz=self.z*self.z;
        zw=self.z*self.w;
        return CMatrix4x4(
                1.0-2.0*(yy+zz),2.0*(xy+zw),2.0*(xz-yw),0.0,
                2.0*(xy-zw),1.0-2.0*(xx+zz),2.0*(yz+xw),0.0,
                2.0*(xz+yw),2.0*(yz-xw),1.0-2.0*(xx+yy),0.0,
                0.0,0.0,0.0,1.0);

    def Normalize(self):
        """This function normalizes this quaternion."""
        Factor=self.x*self.x+self.y*self.y+self.z*self.z+self.w*self.w;
        if(Factor>0.00000001):
            Factor=Factor**(-0.5);
            self.x*=Factor;
            self.y*=Factor;
            self.z*=Factor;
            self.w*=Factor;

    def __str__(self):
        """This function returns a string giving the current values of this
           quaternion in an easily readable format."""
        return "x="+str(self.x)+" y="+str(self.y)+" z="+str(self.z)+" w="+str(self.w);

    def __mul__(self,RHS):
        """This operator returns the result of the CQuaternion RHS, getting multiplied
           from the left using this quaternion. The resulting quaternion describes
           the rotation of this quaternion followed by the rotation of RHS. For
           incompatible right-hand-side types it returns NotImplemented."""
        if(isinstance(RHS,CQuaternion)):
            return CQuaternion(
               RHS.w*self.x+self.w*RHS.x+RHS.y*self.z-RHS.z*self.y,
               RHS.w*self.y+self.w*RHS.y+RHS.z*self.x-RHS.x*self.z,
               RHS.w*self.z+self.w*RHS.z+RHS.x*self.y-RHS.y*self.x,
               RHS.w*self.w-RHS.x*self.x-RHS.y*self.y-RHS.z*self.z);
        else:
            return NotImplemented;
        
    def __eq__(self,RHS):
        """This operator returns true, if and only if the given quaternion is 
           identical to this quaternion in every component."""
        return isinstance(RHS,CQuaternion) and self.w==RHS.w and self.x==RHS.x and self.y==RHS.y and self.z==RHS.z;

    def __rmul__(self,LHS):
        """This operator returns the result of the CQuaternion, CMatrix4x4 or
           C3DVector LHS, getting multiplied from the right using this quaternion.
           For a transformation matrix the result is a matrix that applies first the
           transformation in LHS and then the rotation in self. For a quaternion the
           result is LHS*self. For a vector the result is the vector rotated using
           this quaternion. For incompatible left-hand-side types it returns
           NotImplemented."""
        if(isinstance(LHS,CMatrix4x4)):
            RHS=CMatrix4x4();
            RHS.ApplyRotation(self);
            return RHS.TransformMatrix(LHS);
        elif(isinstance(LHS,CQuaternion)):
            return LHS*self;
        elif(isinstance(LHS,C3DVector)):
            return self.RotateVector(LHS.x,LHS.y,LHS.z);
        else:
            return NotImplemented;

    def __imul__(self,RHS):
        """This multiplication-assignment operator assigns self*RHS to self and
           returns self or it returns NotImplemented and does nothing, if RHS is not
           a CQuaternion."""
        if(isinstance(RHS,CQuaternion)):
            self.Copy(self*RHS);
            return self;
        else:
            return NotImplemented;

    def __neg__(self):
        """This operator returns the additive inverse of this quaternion. It defines
           the same rotation, but slerping from another quaternion to the inverted
           quaternion will use the other rotational direction."""
        return CQuaternion(-self.x,-self.y,-self.z,-self.w);

    def __invert__(self):
        """This operator returns the multiplicative inverse of this quaternion. The
           multiplicative inverse defines the inverse rotation."""
        return CQuaternion(self.x,self.y,self.z,-self.w);

    # The member variables, describing an identity rotation
    x=0.0;
    y=0.0;
    z=0.0;
    w=1.0;


class CMatrix4x4:
    """This class represents a 4x4 transformation matrix."""

    def __init__(self,_11=1.0,_12=0.0,_13=0.0,_14=0.0,_21=0.0,_22=1.0,_23=0.0,_24=0.0,_31=0.0,_32=0.0,_33=1.0,_34=0.0,_41=0.0,_42=0.0,_43=0.0,_44=1.0):
        """This constructor initializes this matrix with the given entries."""
        self._11=_11; self._12=_12; self._13=_13; self._14=_14;
        self._21=_21; self._22=_22; self._23=_23; self._24=_24;
        self._31=_31; self._32=_32; self._33=_33; self._34=_34;
        self._41=_41; self._42=_42; self._43=_43; self._44=_44;

    def Clone(self):
        """This function returns a deep-copy of this object."""
        return CMatrix4x4(
            self._11,self._12,self._13,self._14,
            self._21,self._22,self._23,self._24,
            self._31,self._32,self._33,self._34,
            self._41,self._42,self._43,self._44);

    def ApplyRotation(self,Quaternion):
        """This function makes the 3x3 part of this matrix become a rotation matrix,
           which matches the given CQuaternion. The current scaling is lost."""
        xx=Quaternion.x*Quaternion.x;
        xy=Quaternion.x*Quaternion.y;
        xz=Quaternion.x*Quaternion.z;
        xw=Quaternion.x*Quaternion.w;
        yy=Quaternion.y*Quaternion.y;
        yz=Quaternion.y*Quaternion.z;
        yw=Quaternion.y*Quaternion.w;
        zz=Quaternion.z*Quaternion.z;
        zw=Quaternion.z*Quaternion.w;
        
        self._11=1.0-2.0*(yy+zz);
        self._12=2.0*(xy+zw);
        self._13=2.0*(xz-yw);
        
        self._21=2.0*(xy-zw);
        self._22=1.0-2.0*(xx+zz);
        self._23=2.0*(yz+xw);

        self._31=2.0*(xz+yw);
        self._32=2.0*(yz-xw);
        self._33=1.0-2.0*(xx+yy);

        
    def ApplyTranslation(self,Translation):
        """This function sets the translation described by this matrix to the value
           in the given 3-tuple, Blender vector or C3DVector."""
        self._41=Translation[0];
        self._42=Translation[1];
        self._43=Translation[2];

    def ApplyScaling(self,Scaling):
        """This function sets the scaling described by this matrix to the values in 
           in the given 3-tuple, Blender vector or C3DVector."""
        self._11*=Scaling[0];
        self._12*=Scaling[0];
        self._13*=Scaling[0];

        self._21*=Scaling[1];
        self._22*=Scaling[1];
        self._23*=Scaling[1];

        self._31*=Scaling[2];
        self._32*=Scaling[2];
        self._33*=Scaling[2];

    def TransformVector(self,Vector):
        """This function returns a transformed version of the given vector. The
           result written with matrix multiplication is Vector*self, where Vector is
           interpreted as line-vector."""
        return C3DVector(Vector.x*self._11+Vector.y*self._21+Vector.z*self._31+self._41,Vector.x*self._12+Vector.y*self._22+Vector.z*self._32+self._42,Vector.x*self._13+Vector.y*self._23+Vector.z*self._33+self._43);

    def TransformMatrix(self,Matrix):
        """This function transforms the given matrix by this matrix. The
           resulting matrix will describe the transformation, that can be
           achieved by using first the given matrix and after that this
           matrix. In other words the result is Matrix*self."""
        return CMatrix4x4(
            self._11*Matrix._11+self._21*Matrix._12+self._31*Matrix._13+self._41*Matrix._14,
            self._12*Matrix._11+self._22*Matrix._12+self._32*Matrix._13+self._42*Matrix._14,
            self._13*Matrix._11+self._23*Matrix._12+self._33*Matrix._13+self._43*Matrix._14,
            self._14*Matrix._11+self._24*Matrix._12+self._34*Matrix._13+self._44*Matrix._14,
            
            self._11*Matrix._21+self._21*Matrix._22+self._31*Matrix._23+self._41*Matrix._24,
            self._12*Matrix._21+self._22*Matrix._22+self._32*Matrix._23+self._42*Matrix._24,
            self._13*Matrix._21+self._23*Matrix._22+self._33*Matrix._23+self._43*Matrix._24,
            self._14*Matrix._21+self._24*Matrix._22+self._34*Matrix._23+self._44*Matrix._24,
            
            self._11*Matrix._31+self._21*Matrix._32+self._31*Matrix._33+self._41*Matrix._34,
            self._12*Matrix._31+self._22*Matrix._32+self._32*Matrix._33+self._42*Matrix._34,
            self._13*Matrix._31+self._23*Matrix._32+self._33*Matrix._33+self._43*Matrix._34,
            self._14*Matrix._31+self._24*Matrix._32+self._34*Matrix._33+self._44*Matrix._34,
            
            self._11*Matrix._41+self._21*Matrix._42+self._31*Matrix._43+self._41*Matrix._44,
            self._12*Matrix._41+self._22*Matrix._42+self._32*Matrix._43+self._42*Matrix._44,
            self._13*Matrix._41+self._23*Matrix._42+self._33*Matrix._43+self._43*Matrix._44,
            self._14*Matrix._41+self._24*Matrix._42+self._34*Matrix._43+self._44*Matrix._44
        );

    def ComputeInverse(self):
        """This function computes the multiplicative inverse of this matrix assuming
           that the last column is (0,0,0,1). If the operation failed due to a
           singular matrix the function returns None and prints a warning, otherwise
           it returns the inverse matrix."""
        # Calculate the inverse determinant of the 3*3 matrix
        InvDeterminant=self.Compute3x3Determinant();
        if(InvDeterminant!=0.0):
            InvDeterminant=1.0/InvDeterminant;
            _11=(self._22*self._33-self._23*self._32)*InvDeterminant;
            _12=(self._12*self._33-self._32*self._13)*-InvDeterminant;
            _13=(self._12*self._23-self._22*self._13)*InvDeterminant;
            _21=(self._21*self._33-self._23*self._31)*-InvDeterminant;
            _22=(self._11*self._33-self._31*self._13)*InvDeterminant;
            _23=(self._11*self._23-self._21*self._13)*-InvDeterminant;
            _31=(self._21*self._32-self._31*self._22)*InvDeterminant;
            _32=(self._11*self._32-self._31*self._12)*-InvDeterminant;
            _33=(self._11*self._22-self._12*self._21)*InvDeterminant;
            return CMatrix4x4(
                _11,_12,_13,0.0,
                _21,_22,_23,0.0,
                _31,_32,_33,0.0,
                -(self._41*_11+self._42*_21+self._43*_31),
                -(self._41*_12+self._42*_22+self._43*_32),
                -(self._41*_13+self._42*_23+self._43*_33),
                1.0
            );
        print("WARNING: There has been an attempt to invert a singular matrix (matrix with scaling zero). This can cause serious problems.");
        return None;

    def ComputeTranspose(self):
        """This function returns the transpose of this matrix. You can find self._ij
           at Result._ji. For rotation matrices this is the inverse matrix."""
        return CMatrix4x4(
            self._11,self._12,self._13,self._14,
            self._21,self._22,self._23,self._24,
            self._31,self._32,self._33,self._34,
            self._41,self._42,self._43,self._44);

    def Compute3x3Determinant(self):
        """This function computes the determinant of the 3x3 sub-matrix of
           this matrix and returns it."""
        return self._11*(self._22*self._33-self._32*self._23)-self._12*(self._21*self._33-self._31*self._23)+self._13*(self._21*self._32-self._31*self._22);

    def RotateAndScaleVector(self,Vector):
        """This function returns a rotated and scaled version of the given
           vector. Translation is not applied."""
        return C3DVector(Vector.x*self._11+Vector.y*self._21+Vector.z*self._31,Vector.x*self._12+Vector.y*self._22+Vector.z*self._32,Vector.x*self._13+Vector.y*self._23+Vector.z*self._33);

    def GetScalingX(self):
        """This function returns the x component of the scaling of this matrix."""
        return (self._11*self._11+self._12*self._12+self._13*self._13)**0.5;
    def GetScalingY(self):
        """This function returns the y component of the scaling of this matrix."""
        return (self._21*self._21+self._22*self._22+self._23*self._23)**0.5;
    def GetScalingZ(self):
        """This function returns the z component of the scaling of this matrix."""
        return (self._31*self._31+self._32*self._32+self._33*self._33)**0.5;
    def GetAverageScaling(self):
        """This function returns the arithmetic mean of the scaling for the three
           components."""
        return (self.GetScalingX()+self.GetScalingY()+self.GetScalingZ())/3.0;
    def ScalingToVector(self):
        """This function returns the scaling along the three axes of this matrix as
           C3DVector."""
        return C3DVector(self.GetScalingX(),self.GetScalingY(),self.GetScalingZ());
    def CheckHomogenousScaling(self,Tolerance=1.01):
        """This function returns true, if and only if the relative difference 
           between the scaling along the different axes in this matrix, is 
           equal according to the given tolerance. Additionally it checks, 
           whether the 3x3 part of this matrix has a non-negative determinant,
           i.e. it checks if this matrix mirrors."""
        Scaling=self.ScalingToVector();
        InvTolerance=1.0/Tolerance;
        return (Scaling.x*Tolerance>Scaling.y and Scaling.x*InvTolerance<Scaling.y
            and Scaling.x*Tolerance>Scaling.z and Scaling.x*InvTolerance<Scaling.z
            and self.Compute3x3Determinant()>=0.0);

    def TranslationToVector(self):
        """This function returns the translation described by this matrix as
           C3DVector."""
        return C3DVector(self._41,self._42,self._43);

    def RotationToQuaternion(self):
        """This function computes a CQuaternion, which matches this matrix and
           returns it. For a meaningful result the 3x3 part of this matrix has to be
           orthogonal."""
        InvScalingX=self.GetScalingX();
        InvScalingY=self.GetScalingY();
        InvScalingZ=self.GetScalingZ();
        if(InvScalingX>=0.000001 and InvScalingY>=0.000001 and InvScalingZ>=0.000001):
            InvScalingX=1.0/InvScalingX;
            InvScalingY=1.0/InvScalingY;
            InvScalingZ=1.0/InvScalingZ;
            if(self.Compute3x3Determinant()<0.0):
                InvScalingY=-InvScalingY;
            _11=self._11*InvScalingX;
            _12=self._12*InvScalingX;
            _13=self._13*InvScalingX;
            _21=self._21*InvScalingY;
            _22=self._22*InvScalingY;
            _23=self._23*InvScalingY;
            _31=self._31*InvScalingZ;
            _32=self._32*InvScalingZ;
            _33=self._33*InvScalingZ;
            Result=CQuaternion(
                (max(0.0, _11-_22-_33+1.0)**0.5)*0.5,
                (max(0.0,-_11+_22-_33+1.0)**0.5)*0.5,
                (max(0.0,-_11-_22+_33+1.0)**0.5)*0.5,
                (max(0.0, _11+_22+_33+1.0)**0.5)*0.5);
            # Except for the signs the result is completely known now. One sign can
            # be chosen, three are needed. This part is very sensitive to rounding
            # errors. A simple approach works only for bigger values of w. In other
            # cases a big vector component is chosen to have a positive sign and the
            # other components are chosen accordingly.
            if(Result.w>0.02):
                if(_23<_32):
                    Result.x=-Result.x;
                if(_31<_13):
                    Result.y=-Result.y;
                if(_12<_21):
                    Result.z=-Result.z;
            elif(Result.x>0.4):
                if(_12<0.0):
                    Result.y=-Result.y;
                if(_13<0.0):
                    Result.z=-Result.z;
                if(_23<_32):
                    Result.w=-Result.w;
            elif(Result.y>0.4):
                if(_12<0.0):
                    Result.x=-Result.x;
                if(_23<0.0):
                    Result.z=-Result.z;
                if(_31<_13):
                    Result.w=-Result.w;
            else:
                if(_13<0.0):
                    Result.x=-Result.x;
                if(_23<0.0):
                    Result.y=-Result.y;
                if(_12<_21):
                    Result.w=-Result.w;
            Result.Normalize();
            return Result;
        else:
            return CQuaternion(0.0,0.0,0.0,1.0);

    def IsIdentity(self,Epsilon=0.00001):
        """This function checks whether this matrix equals the identity matrix using
           the given additive tolerance."""
        L0=-Epsilon;
        U0=Epsilon;
        L1=1.0-Epsilon;
        U1=1.0+Epsilon;
        return (
            self._11>L1 and self._11<U1 and self._12>L0 and self._12<U0 and self._13>L0 and self._13<U0 and self._14>L0 and self._14<U0
        and self._21>L0 and self._21<U0 and self._22>L1 and self._22<U1 and self._23>L0 and self._23<U0 and self._24>L0 and self._24<U0
        and self._31>L0 and self._31<U0 and self._32>L0 and self._32<U0 and self._33>L1 and self._33<U1 and self._34>L0 and self._34<U0
        and self._41>L0 and self._41<U0 and self._42>L0 and self._42<U0 and self._43>L0 and self._43<U0 and self._44>L1 and self._44<U1
        );

    def __str__(self):
        """This function returns a string giving the current values of this matrix
           in an easy to read format."""
        return ("_11="+str(self._11)+" _12="+str(self._12)+" _13="+str(self._13)+" _14="+str(self._14)+"\n"+
                "_21="+str(self._21)+" _22="+str(self._22)+" _23="+str(self._23)+" _24="+str(self._24)+"\n"+
                "_31="+str(self._31)+" _32="+str(self._32)+" _33="+str(self._33)+" _34="+str(self._34)+"\n"+
                "_41="+str(self._41)+" _42="+str(self._42)+" _43="+str(self._43)+" _44="+str(self._44)+"\n");

    def __eq__(self,RHS):
        """This operator returns true, if and only if the given vector is 
           identical to this vector in every component."""
        return (isinstance(RHS,CMatrix4x4) and 
                self._11==RHS._11 and self._12==RHS._12 and self._13==RHS._13 and self._14==RHS._14 and
                self._21==RHS._21 and self._22==RHS._22 and self._23==RHS._23 and self._24==RHS._24 and
                self._31==RHS._31 and self._32==RHS._32 and self._33==RHS._33 and self._34==RHS._34 and
                self._41==RHS._41 and self._42==RHS._42 and self._43==RHS._43 and self._44==RHS._44);

    def __add__(self,RHS):
        """This operator implements component-wise matrix addition. RHS has to be
           another CMatrix4x4 otherwise NotImplemented is returned."""
        if(isinstance(RHS,CMatrix4x4)):
            return CMatrix4x4(
                self._11+RHS._11,self._12+RHS._12,self._13+RHS._13,self._14+RHS._14,
                self._21+RHS._21,self._22+RHS._22,self._23+RHS._23,self._24+RHS._24,
                self._31+RHS._31,self._32+RHS._32,self._33+RHS._33,self._34+RHS._34,
                self._41+RHS._41,self._42+RHS._42,self._43+RHS._43,self._44+RHS._44);
        else:
            return NotImplemented;

    def __sub__(self,RHS):
        """This operator implements component-wise matrix subtraction. RHS has to be
           another CMatrix4x4 otherwise NotImplemented is returned."""
        if(isinstance(RHS,CMatrix4x4)):
            return CMatrix4x4(
                self._11-RHS._11,self._12-RHS._12,self._13-RHS._13,self._14-RHS._14,
                self._21-RHS._21,self._22-RHS._22,self._23-RHS._23,self._24-RHS._24,
                self._31-RHS._31,self._32-RHS._32,self._33-RHS._33,self._34-RHS._34,
                self._41-RHS._41,self._42-RHS._42,self._43-RHS._43,self._44-RHS._44);
        else:
            return NotImplemented;

    def __mul__(self,RHS):
        """This operator returns the result of the CMatrix4x4 RHS, getting multiplied
           from the left using this matrix. It uses matrix multiplication so for a
           transformation matrix the result is a matrix that applies first the
           transformation in self and then the transformation in RHS. For
           incompatible right-hand-side types it returns NotImplemented."""
        if(isinstance(RHS,CMatrix4x4)):
            return RHS.TransformMatrix(self);
        else:
            return NotImplemented;

    def __rmul__(self,LHS):
        """This operator returns the result of the CMatrix4x4 or C3DVector LHS,
           getting multiplied from the right using this matrix. It uses matrix
           multiplication so for a transformation matrix the result is a matrix
           that applies first the transformation in LHS and then the transformation
           in self. For incompatible left-hand-side types it returns
           NotImplemented."""
        if(isinstance(LHS,CMatrix4x4)):
            return self.TransformMatrix(LHS);
        elif(isinstance(LHS,C3DVector)):
            return self.TransformVector(LHS);
        else:
            return NotImplemented;

    def __iadd__(self,RHS):
        """This additon-assignment operator assigns self+RHS to self and returns self
           or it returns NotImplemented and does nothing, if RHS is not a CMatrix4x4."""
        if(isinstance(RHS,CMatrix4x4)):
            self.Copy(self+RHS);
            return self;
        else:
            return NotImplemented;

    def __isub__(self,RHS):
        """This subtraction-assignment operator assigns self-RHS to self and returns self
           or it returns NotImplemented and does nothing, if RHS is not a CMatrix4x4."""
        if(isinstance(RHS,CMatrix4x4)):
            self.Copy(self-RHS);
            return self;
        else:
            return NotImplemented;

    def __imul__(self,RHS):
        """This multiplication-assignment operator assigns self*RHS to self and
           returns self or it returns NotImplemented and does nothing, if RHS is not
           a CMatrix4x4."""
        if(isinstance(RHS,CMatrix4x4)):
            self.Copy(self*RHS);
            return self;
        else:
            return NotImplemented;

    def __neg__(self):
        """This operator returns the additive inverse of this matrix. This means that
           it returns a copy of this matrix, where the sign of every component is
           negated."""
        return CMatrix4x4(
            -self._11,-self._12,-self._13,-self._14,
            -self._21,-self._22,-self._23,-self._24,
            -self._31,-self._32,-self._33,-self._34,
            -self._41,-self._42,-self._43,-self._44);

    def __invert__(self):
        """This operator returns the multiplicative inverse of this matrix or None,
           if this matrix is singular. It assumes that the last column of this matrix
           has the entries 0.0, 0.0, 0.0, 1.0 (common for transformation matrices).
           The multiplicative inverse defines the inverse transformation."""
        return self.ComputeInverse();

    _11=1.0; _12=0.0; _13=0.0; _14=0.0;
    _21=0.0; _22=1.0; _23=0.0; _24=0.0;
    _31=0.0; _32=0.0; _33=1.0; _34=0.0;
    _41=0.0; _42=0.0; _43=0.0; _44=1.0;


def MakeCMatrix4x4(Matrix):
    """This function creates a new CMatrix4x4 and initializes it using the 
       Copy(...) function with the given object as parameter. Thus it is 
       suitable to cast other matrix types to CMatrix4x4."""
    Result=CMatrix4x4();
    Result.Copy(Matrix);
    return Result;


class CColor:
    """ This class represents an r, g, b, a color (four floats)."""

    def __init__(self,r,g,b,a):
        self.r=r;
        self.g=g;
        self.b=b;
        self.a=a;

    r=1.0;
    g=1.0;
    b=1.0;
    a=1.0;
    
    
def CompleteFrameList(FrameList,MaxFrameDistance=0.0,DefaultFrame=1.0):
    """This function takes a frame list (i.e. a list of floats) and ensures that it 
       is free of duplicates, consists of floats, is non-empty (empty lists are 
       replaced by the list containing only DefaultFrame) and that the difference 
       between subsequent frames is never greater than MaxFrameDistance (if 
       MaxFrameDistance is positive). The resulting list of floats is returned."""
    # Ensure that all entries are floats and get rid of duplicates
    FrameList=list(frozenset(FrameList));
    for iFrame,Frame in enumerate(FrameList):
        FrameList[iFrame]=float(Frame);
    # Sort the list and make it non-empty
    FrameList=sorted(FrameList);
    if(len(FrameList)==0):
        FrameList.append(float(DefaultFrame));
    # Add keys as implied by MaxFrameDistance
    MaxFrameDistance=float(MaxFrameDistance);
    if(MaxFrameDistance>0.0):
        i=0;
        while(i<len(FrameList)-1):
            # If a gap is too big
            if(FrameList[i+1]-FrameList[i]>MaxFrameDistance):
                # Add as many keys as necessary
                NewFrames=list();
                NewFrame=FrameList[i]+MaxFrameDistance;
                while(NewFrame<FrameList[i+1]):
                    NewFrames.append(NewFrame);
                    NewFrame+=MaxFrameDistance;
                FrameList=FrameList[0:i+1]+NewFrames+FrameList[i+1:];
                # And continue behind the gap
                i+=1+len(NewFrames);
            else:
                i+=1;
    return FrameList;


class CKey:
    """This is a base class for keys. It contains nothing but the time, because 
       the value type is specific to derived classes."""
       
    def __init__(self,Frame):
        """Initializes this key at the given frame."""
        self.Frame=float(Frame);
        
    def SameTime(self,OtherKey,Tolerance=0.0001):
        """This function returns true, if and only if the given key is at the 
           same time as this key. It uses a small tolerance for comparison."""
        return self.Frame-Tolerance<=OtherKey.Frame and self.Frame+Tolerance>=OtherKey.Frame;
    
    # The time of this key in frames. This might be any floating point value.
    Frame=0.0;


def GetKeyFrame(Key):
    """This tiny utility function returns the frame value, which is associated with 
       a given CKey. It can be used as parameter for sort."""
    return Key.Frame;

class CScalingKey(CKey):
    """This class handles a single scaling key."""

    def __init__(self,Frame,Scaling):
        """Initialize this key with the given data."""
        CKey.__init__(self,Frame);
        self.Scaling=Scaling;

    Scaling=C3DVector(0.0,0.0,0.0);


class CTranslationKey(CKey):
    """This class handles a single translation key."""

    def __init__(self,Frame,Translation):
        """Initialize this key with the given data."""
        CKey.__init__(self,Frame);
        self.Translation=Translation;

    Translation=C3DVector(0.0,0.0,0.0);


class CRotationKey(CKey):
    """This class handles a single rotation key."""

    def __init__(self,Frame,Rotation):
        """Initialize this key with the given data."""
        CKey.__init__(self,Frame);
        self.Rotation=Rotation;

    Rotation=CQuaternion(0.0,0.0,0.0,1.0);


class CKeyedTransformation:
    """This class describes a time-dependent transformation using key frames."""

    def __init__(self,Name,ParentTransformation=None):
        """This constructor initializes this keyed transformation with the given name
           and without any keys. The given other keyed transformation will be used as
           parent (if any). This is only relevant for the GetAbsolute...(...)
           functions."""
        self.Name=Name;
        self.pScalingKey=[];
        self.pRotationKey=[];
        self.pTranslationKey=[];
        self.MinFrame=0.0;
        self.MaxFrame=0.0;
        self.SourceList=[];
        self.Parent=ParentTransformation;


    def FixRotationKeys(self):
        """A slerp between two rotation keys can take two directions. Often it is not
           possible to find the correct one from the input data, but usually the
           shortest way is the right one. This function ensures that the keys always
           define the shortest way."""
        for i in range(len(self.pRotationKey)-1):
            Rotation=self.pRotationKey[i].Rotation;
            NextRotation=self.pRotationKey[i+1].Rotation;
            if(Rotation.x*NextRotation.x+Rotation.y*NextRotation.y+Rotation.z*NextRotation.z+Rotation.w*NextRotation.w<0.0):
                self.pRotationKey[i+1].Rotation=-NextRotation;


    def MakeDefaultKeys(self,Translation=C3DVector(0.0,0.0,0.0),Rotation=CQuaternion(0.0,0.0,0.0,1.0),Scaling=C3DVector(1.0,1.0,1.0)):
        """This function ensures that there is at least one key of every type. 
           If a key needs to be added the given transformation values are used 
           and self.MinFrame is used as time."""
        if(len(self.pTranslationKey)==0):
            self.pTranslationKey.append(CTranslationKey(self.MinFrame,Translation));
        if(len(self.pRotationKey)==0):
            self.pRotationKey.append(CRotationKey(self.MinFrame,Rotation));
        if(len(self.pScalingKey)==0):
            self.pScalingKey.append(CScalingKey(self.MinFrame,Scaling));

    def AddKeysFromActionCurves(self,ModelData,ArmatureData,ActionName,BoneName,BoneToParentSpace,MaxKeyDistance=0,InheritRotation=True,InheritScaling=True,IgnoreOwnTranslation=False,ArmatureSpaceTranslation=False):
        """This function adds key frames from the action with the given name in 
           the given CArmatureDataBase for the bone with the given name to this 
           keyed transformation. The given boolean flags are used to determine how 
           the transformation of the parent effects this transformation."""
        # Try to lock the curve data
        if(not ArmatureData.LockBoneAction(ActionName,BoneName)):
            return False;
        # Decompose the parent transformation, verify whether its valid
        BoneToParentTranslation=BoneToParentSpace.TranslationToVector();
        BoneToParentRotation=BoneToParentSpace.RotationToQuaternion();
        BoneToParentScaling=BoneToParentSpace.ScalingToVector();
        if(not BoneToParentSpace.CheckHomogenousScaling()):
            print("WARNING: The bone "+self.Name+" has an inhomogenous scaling in relation to its parent. Since inhomogenous scaling (scaling that's not the same for all axes) is not supported the scaling has been made homogenous.");
            print("The inhomogenous scaling is X="+str(BoneToParentScaling.x)+" Y="+str(BoneToParentScaling.y)+" Z="+str(BoneToParentScaling.z));
            BoneToParentScaling.x=BoneToParentScaling.y=BoneToParentScaling.z=(BoneToParentScaling.x+BoneToParentScaling.y+BoneToParentScaling.z)/3.0;
        # Get the key frame list and the range
        KeyFrameList=ModelData.GetActionFrameList(ActionName);
        SortedKeyFrameList=self.MakeSortedFrameList(KeyFrameList,MaxKeyDistance,not InheritRotation or not InheritScaling or ArmatureSpaceTranslation,self.MinFrame);
        # Create the keys
        for Frame in SortedKeyFrameList:
            # Get the key values
            KeyTranslation=ArmatureData.GetBoneActionTranslation(Frame);
            KeyScaling=ArmatureData.GetBoneActionScaling(Frame);
            KeyRotation=ArmatureData.GetBoneActionRotation(Frame);
            # Transform the key data
            if(KeyScaling!=None):
                KeyScaling=C3DVector(KeyScaling.x*BoneToParentScaling.x,KeyScaling.y*BoneToParentScaling.y,KeyScaling.z*BoneToParentScaling.z);
            if(KeyTranslation!=None):
                if(IgnoreOwnTranslation):
                    KeyTranslation=BoneToParentTranslation;
                elif(ArmatureSpaceTranslation):
                    if(self.Parent!=None):
                        KeyTranslation=KeyTranslation-self.Parent.GetAbsoluteFrameTranslation(Frame);
                else:
                    RotatedTranslation=KeyTranslation*BoneToParentRotation;
                    RotatedTranslation.x*=BoneToParentScaling.x;
                    RotatedTranslation.y*=BoneToParentScaling.y;
                    RotatedTranslation.z*=BoneToParentScaling.z;
                    KeyTranslation=RotatedTranslation+BoneToParentTranslation;
            if(KeyRotation!=None):
                KeyRotation*=BoneToParentRotation;
            if(self.Parent!=None):
                if(not InheritRotation and KeyRotation!=None):
                    KeyRotation*=~self.Parent.GetAbsoluteFrameRotation(Frame);
                if(not InheritScaling and KeyScaling!=None):
                    KeyScaling=(1.0/self.Parent.GetAbsoluteFrameScaling(Frame).x)*KeyScaling;
            # Save the transformed key data
            if(KeyTranslation!=None):
                self.pTranslationKey.append(CTranslationKey(Frame,KeyTranslation));
            if(KeyRotation!=None):
                self.pRotationKey.append(CRotationKey(Frame,KeyRotation));
            if(KeyScaling!=None):
                self.pScalingKey.append(CScalingKey(Frame,KeyScaling));
        ArmatureData.UnlockBoneAction();
        return True;


    def KeyCurrentPose(self,Frame,ArmatureData,BoneName):
        """This function creates scaling, rotation and translation keys for the
           current pose of the bone with the given name in the given armature 
           and associates them with the given frame. It returns true, if and only if 
           it added keys successfully."""
        # Retrieve the transformation in relation to the parent
        Transformation=ArmatureData.GetBonePoseTransformation(BoneName);
        if(Transformation==None):
            return False;
        Transformation=MakeCMatrix4x4(Transformation);
        ParentName=ArmatureData.GetBoneParent(BoneName);
        if(ParentName!=None):
            ParentTransformation=ArmatureData.GetBonePoseTransformation(ParentName);
            if(ParentTransformation==None):
                return True;
            ParentTransformation=MakeCMatrix4x4(ParentTransformation);
            Transformation*=ParentTransformation.ComputeInverse();
            # Check for an inhomogenuous parent scaling
            if(not ParentTransformation.CheckHomogenousScaling()):
                print("WARNING: The bone \""+ParentName+"\" is used as parent, but has an inhomogenuous scaling. This will have a negative effect on the animation of the child bones.");
        # Create the transformation keys
        self.pRotationKey.append(CRotationKey(Frame,Transformation.RotationToQuaternion()));
        self.pScalingKey.append(CScalingKey(Frame,Transformation.ScalingToVector()));
        self.pTranslationKey.append(CTranslationKey(Frame,Transformation.TranslationToVector()));
        return True;

    def AddTransformationKey(self,Frame,Transformation):
        """This function takes a this to parent transformation as CMatrix4x4. Then 
           it creates a corresponding rotation, scaling and translation key and 
           stores them at the given frame in this transformation."""
        Rotation=Transformation.RotationToQuaternion();
        self.pRotationKey.append(CRotationKey(Frame,Rotation));
        Translation=Transformation.TranslationToVector();
        self.pTranslationKey.append(CTranslationKey(Frame,Translation));
        Scaling=Transformation.ScalingToVector();
        self.pScalingKey.append(CScalingKey(Frame,Scaling));
        
    def AddKeys(self,Frame,Translation=None,Scaling=None,Rotation=None):
        """This function adds keys at the given frame using the given transformation 
           values. All keys are optional and are not created, if you pass None for 
           the corresponding parameter. Scaling and Translation have to be given as 
           C3DVector, Rotation as CQuaternion."""
        if(Rotation!=None):
            self.pRotationKey.append(CRotationKey(Frame,Rotation));
        if(Translation!=None):
            self.pTranslationKey.append(CTranslationKey(Frame,Translation));
        if(Scaling!=None):
            self.pScalingKey.append(CScalingKey(Frame,Scaling));
        

    def GetFrameList(self,IncludeParent=True):
        """This function returns a monotone list of all frames, which are keyed with
           at least one key for this bone. On request, keys of parents are 
           included."""
        FrameSet=set();
        for Key in self.pTranslationKey+self.pRotationKey+self.pScalingKey:
            FrameSet|=set([Key.Frame]);
        if(IncludeParent and self.Parent!=None):
            FrameSet|=set(self.Parent.GetFrameList(True));
        FrameList=list(FrameSet);
        FrameList.sort();
        return FrameList;


    def ValidateRange(self):
        """This function ensures that MinFrame and MaxFrame are valid by 
           iterating through all key frames and minimizing resp. maximizing 
           the values. It returns the new pair (self.MinFrame,self.MaxFrame)."""
        for Key in self.pTranslationKey+self.pRotationKey+self.pScalingKey:
            self.MinFrame=min(self.MinFrame,Key.Frame);
            self.MaxFrame=max(self.MaxFrame,Key.Frame);
        return (self.MinFrame,self.MaxFrame);


    def GetFrameTranslation(self,Frame):
        """This function returns the local translation, which has to be used for the
           given frame as C3DVector. It does not regard parents."""
        (Start,End,LerpFactor)=self.GetFrameKey(self.pTranslationKey,Frame);
        if(Start!=None):
            return C3DVector(
                Lerp(Start.Translation.x,End.Translation.x,LerpFactor),
                Lerp(Start.Translation.y,End.Translation.y,LerpFactor),
                Lerp(Start.Translation.z,End.Translation.z,LerpFactor));
        else:
            return C3DVector(0.0,0.0,0.0);

    def GetAbsoluteFrameTranslation(self,Frame):
        """This function returns the absolute translation, which has to be used for the
           given frame as C3DVector. All parent transformations are included."""
        Result=self.GetFrameTranslation(Frame);
        if(self.Parent!=None):
            Result*=self.Parent.GetAbsoluteFrameTransformation(Frame);
        return Result;

    def GetFrameRotation(self,Frame):
        """This function returns the local rotation, which has to be used for the
           given frame as CQuaternion. It does not regard parents."""
        (Start,End,LerpFactor)=self.GetFrameKey(self.pRotationKey,Frame);
        if(Start!=None):
            return Start.Rotation.Slerp(End.Rotation,LerpFactor);
        else:
            return CQuaternion(0.0,0.0,0.0,1.0);

    def GetAbsoluteFrameRotation(self,Frame):
        """This function returns a CQuaternion defining the absolute rotation at the
           given frame, including parent rotations."""
        Result=self.GetFrameRotation(Frame);
        if(self.Parent!=None):
            Result*=self.Parent.GetAbsoluteFrameRotation(Frame);
        return Result;

    def GetFrameScaling(self,Frame):
        """This function returns the local scaling, which has to be used for the
           given frame as C3DVector. It does not regard parents."""
        (Start,End,LerpFactor)=self.GetFrameKey(self.pScalingKey,Frame);
        if(Start!=None):
            return C3DVector(
                Lerp(Start.Scaling.x,End.Scaling.x,LerpFactor),
                Lerp(Start.Scaling.y,End.Scaling.y,LerpFactor),
                Lerp(Start.Scaling.z,End.Scaling.z,LerpFactor));
        else:
            return C3DVector(0.0,0.0,0.0);

    def GetAbsoluteFrameScaling(self,Frame):
        """This function returns a C3DVector defining the absolute scaling at the
           given frame. Parent scalings are assumed to be uniform."""
        Result=self.GetFrameScaling(Frame);
        if(self.Parent!=None):
            Result=self.Parent.GetAbsoluteFrameScaling(Frame).x*Result;
        return Result;

    def GetFrameTransformation(self,Frame):
        """This function returns a CMatrix4x4 defining the transformation at the
           given frame. It does not regard parents, so if there are parents the 
           returned transformation is the object to parent space transformation."""
        Result=CMatrix4x4();
        Result.ApplyRotation(self.GetFrameRotation(Frame));
        Result.ApplyTranslation(self.GetFrameTranslation(Frame));
        Result.ApplyScaling(self.GetFrameScaling(Frame));
        return Result;

    def GetAbsoluteFrameTransformation(self,Frame):
        """This function returns a CMatrix4x4 defining the absolute transformation at
           the given frame, including parent transformations. It gives an object to
           model space transformation."""
        Result=self.GetFrameTransformation(Frame);
        if(self.Parent!=None):
            Result*=self.Parent.GetAbsoluteFrameTransformation(Frame);
        return Result;

    def GetFrameKey(self,pKeyArray,Frame):
        """This function returns a triple of two elements of the given key array and
           a lerp factor between them for the given frame. The first key is the key
           with maximal index, which has a frame value smaller than the given frame
           value. The second one is the key with minimal index, which has a frame
           value bigger than the given frame value. If the key array is empty, it 
           returns (None,None,None)."""
        if(len(pKeyArray)==0):
            return (None,None,None);
        StartKey=pKeyArray[0];
        EndKey=pKeyArray[len(pKeyArray)-1];
        LerpFactor=0.0;
        if(EndKey.Frame<=Frame):
            StartKey=EndKey;
        elif(Frame<=StartKey.Frame):
            EndKey=StartKey;
        else:
            for i in range(len(pKeyArray)):
                if(pKeyArray[i].Frame>Frame):
                    StartKey=pKeyArray[i-1];
                    EndKey=pKeyArray[i];
                    LerpFactor=(Frame-StartKey.Frame)/(EndKey.Frame-StartKey.Frame);
                    break;
        return (StartKey,EndKey,LerpFactor);

    def Append(self,RHS,FrameOffset=None):
        """This function appends the given keyed transformation to this keyed
           transformation. This means that the keys get appended with an offset frame
           value such that the animation in the given keyed transformation starts
           right after the animation in this keyed transformation. Alternatively you
           can specify the offset explicitly. Any floating point offset is valid."""
        if(FrameOffset==None):
            FrameOffset=self.MaxFrame-RHS.MinFrame+1.0;
        self.SourceList+=RHS.SourceList;
        # Add the frames at the end
        for Scaling in RHS.pScalingKey:
            self.pScalingKey.append(CScalingKey(Scaling.Frame+FrameOffset,Scaling.Scaling));
        for Rotation in RHS.pRotationKey:
            self.pRotationKey.append(CRotationKey(Rotation.Frame+FrameOffset,Rotation.Rotation));
        for Translation in RHS.pTranslationKey:
            self.pTranslationKey.append(CTranslationKey(Translation.Frame+FrameOffset,Translation.Translation));
        # Ensure that the keys are sorted properly
        self.SortKeys();
        # Ensure that the lists are sorted
        self.MinFrame=min(self.MinFrame,RHS.MinFrame+FrameOffset);
        self.MaxFrame=max(self.MaxFrame,RHS.MaxFrame+FrameOffset);
        
    def SortKeys(self):
        """This function sorts all keys, which contribute to this transformation in 
           ascending order according to their frame value."""
        self.pScalingKey.sort(key=GetKeyFrame);
        self.pRotationKey.sort(key=GetKeyFrame);
        self.pTranslationKey.sort(key=GetKeyFrame);


    def TransformTime(self,FrameOffset,FrameFactor):
        """This function multiplies all frame values by the given (floating 
           point) time factor and adds the given offset after multiplication. 
           If the factor is negative it automatically reverts the key frame 
           order."""
        for Key in self.pTranslationKey+self.pRotationKey+self.pScalingKey:
            Key.Frame=Key.Frame*FrameFactor+FrameOffset;
        self.MinFrame=self.MinFrame*FrameFactor+FrameOffset;
        self.MaxFrame=self.MaxFrame*FrameFactor+FrameOffset;
        if(FrameFactor<0.0):
            self.pTranslationKey=list(reversed(self.pTranslationKey));
            self.pRotationKey=list(reversed(self.pRotationKey));
            self.pScalingKey=list(reversed(self.pScalingKey));
            MaxFrame=self.MinFrame;
            self.MinFrame=self.MaxFrame;
            self.MaxFrame=MaxFrame;

    def ApplyTransformation(self,Transformation):
        """This function multiplies the transformations of this keyed transformation
           from the right by the transformation of the given CMatrix4x4. The matrix
           must have an orthogonal 3x3 part with a uniform scaling."""
        # Decompose the given transformation
        Rotation=Transformation.RotationToQuaternion();
        LHRotation=CQuaternion(0.0,0.0,0.0,1.0);
        Scaling=Transformation.ScalingToVector();
        if(Transformation.Compute3x3Determinant()<0.0):
            Scaling.y*=-1.0;
            # The other two components get negated as well and this rotation gets
            # compensated through the left-hand rotation
            Scaling.x*=-1.0;
            Scaling.z*=-1.0;
            LHRotation.MakeRotationAlongAxis(C3DVector(0.0,1.0,0.0),3.14159265358979);
            Rotation=LHRotation*Rotation;
        # Ensure that there is at least one key of every type
        self.MakeDefaultKeys();
        # Transform every single key
        for RotationKey in self.pRotationKey:
            RotationKey.Rotation=RotationKey.Rotation*Rotation;
        for ScalingKey in self.pScalingKey:
            ScalingKey.Scaling.x*=Scaling.x;
            ScalingKey.Scaling.y*=Scaling.y;
            ScalingKey.Scaling.z*=Scaling.z;
        for TranslationKey in self.pTranslationKey:
            TranslationKey.Translation*=Transformation;


    def ApplyKeyedTransformation(self,RHS):
        """This function multiplies the transformations of this keyed transformation
           from the right by the transformations of the given keyed transformation.
           Note that it is not perfectly precise regarding translations."""
        # Create a list of frames at which keys are needed
        NewFrameList=list(set(self.GetFrameList()+RHS.GetFrameList()));
        NewFrameList.sort();
        if(len(NewFrameList)==0):
            NewFrameList=[self.MinFrame];
        # Create scaling and rotation keys for all of these frames
        pScalingKey=[];
        pRotationKey=[];
        for Frame in NewFrameList:
            FrameRotation=self.GetFrameRotation(Frame)*RHS.GetFrameRotation(Frame);
            pRotationKey.append(CRotationKey(Frame,FrameRotation));
            FrameScaling=RHS.GetFrameScaling(Frame).x*self.GetFrameScaling(Frame);
            pScalingKey.append(CScalingKey(Frame,FrameScaling));
        # For translations it is safer to key all frames, because the translation
        # between frames might lie on circular curves
        pTranslationKey=[];
        Frame=NewFrameList[0];
        while(Frame<NewFrameList[len(NewFrameList)-1]+1.0):
            FrameTranslation=self.GetFrameTranslation(Frame)*RHS.GetFrameTransformation(Frame);
            pTranslationKey.append(CTranslationKey(Frame,FrameTranslation));
            Frame+=1.0;
        # Save the new keys
        self.pScalingKey=pScalingKey;
        self.pRotationKey=pRotationKey;
        self.pTranslationKey=pTranslationKey;
        
    def MakeSortedFrameList(self,FrameList,MaxKeyDistance,AddParentFrames,DefaultFrame):
        """This utility function completes and sorts the given list of frames. 
           It might add frames of the parent and it adds frames as implied by 
           MaxKeyDistance. The output is a non-empty list of sorted floats in 
           every case. If necessary DefaultFrame is used as only value."""
        # Add parent frames, if necessary
        if(AddParentFrames and self.Parent!=None):
            FrameList=list(set(FrameList)|set(self.Parent.GetFrameList()));
        else:
            FrameList=list(set(FrameList));
        # Everything else is implemented in a global utility function
        return CompleteFrameList(FrameList,MaxKeyDistance,DefaultFrame);

    def __imul__(self,RHS):
        """This operator transforms this keyed transformation using either
           ApplyTransformation or ApplyKeyedTransformation(...) dependent on the
           right hand side."""
        if(isinstance(RHS,CMatrix4x4)):
            self.ApplyTransformation(RHS);
        elif(isinstance(RHS,CKeyedTransformation)):
            self.ApplyKeyedTransformation(RHS);
        else:
            raise TypeError("A keyed transformation is being multiplied by an incompatible type. Only CMatrix4x4 and CKeyedTransformation are allowed.");
        return self;
    
    def IsConstant(self):
        """This function returns true, if and only if all keys in this keyed 
           transformation have the same value, i.e. the transformation does not 
           change over time."""
        Result=True;
        if(len(self.pScalingKey)>0):
            for Key in self.pScalingKey:
                Result=(Result and Key.Scaling==self.pScalingKey[0].Scaling);
        if(len(self.pRotationKey)>0 and Result):
            for Key in self.pRotationKey:
                Result=(Result and Key.Rotation==self.pRotationKey[0].Rotation);
        if(len(self.pTranslationKey)>0 and Result):
            for Key in self.pTranslationKey:
                Result=(Result and Key.Translation==self.pTranslationKey[0].Translation);
        return Result;

    def Clone(self):
        """This function returns a new keyed transformation, which is an exact deep
           copy of this keyed transformation."""
        Result=CKeyedTransformation(self.Name,self.Parent);
        for Key in self.pScalingKey:
            Result.pScalingKey.append(CScalingKey(Key.Frame,Key.Scaling));
        for Key in self.pRotationKey:
            Result.pRotationKey.append(CRotationKey(Key.Frame,Key.Rotation));
        for Key in self.pTranslationKey:
            Result.pTranslationKey.append(CTranslationKey(Key.Frame,Key.Translation));
        Result.MinFrame=self.MinFrame;
        Result.MaxFrame=self.MaxFrame;
        Result.SourceList=list(self.SourceList);
        return Result;

    # The name of this keyed transformation. This is the name of its Ipo and of its
    # bone or object.
    Name="";

    # The scaling keys of this bone. If this class is in a valid state it contains at
    # least one CScalingKey.
    pScalingKey=[];
    # The translation keys of this bone. If this class is in a valid state it contains at
    # least one CRotationKey.
    pRotationKey=[];
    # The rotation keys of this bone. If this class is in a valid state it contains at
    # least one CTranslationKey.
    pTranslationKey=[];

    # This is either another keyed transformation, which functions as parent in the
    # GetAbsolute...(...) functions, or None, if there is no parent
    Parent=None;

    # These two floats save the range of this animation
    # The minimal frame value that occurs in the keys
    MinFrame=0.0;
    # The maximal frame value that occurs in the keys
    MaxFrame=0.0;

    # This is a list of Blender actions, which have been used as data source for this
    # object
    SourceList=[];


class CKeyedTransformationDict(dict):
    """This dictionary handles a list of transformations for different actions. It 
       maps action names or other strings to keyed transformations."""

    def __init__(self):
        """This constructor initializes this list with no keyed transformations."""
        pass;
    
    def GetKeyedTransformation(self,Key,ParentTransformation=None):
        """This function returns the keyed, transformation, which is stored using 
           the given string as key. If there is no such keyed transformation it 
           creates a new transformation, stores it in the dictionary and returns it. 
           Optionally you can define a parent transformation for the newly created 
           transformation. It has to be a CKeyedTransformation."""
        if(Key in self):
            return self[Key];
        else:
            KeyedTransformation=CKeyedTransformation(Key,ParentTransformation);
            self[Key]=KeyedTransformation;
            return KeyedTransformation;

    def CombineTransformations(self,OffsetDict=None):
        """This function simplifies this transformation list by appending all keyed
           transformations to a new keyed transformation named "". The other 
           transformations get deleted and the new transformation is added with an 
           empty string as key. The resulting keyed transformation is also returned. 
           You can pass a dictionary, which maps transformation names to 
           corresponding offsets. If so these offsets are applied to the respective 
           transformations before appending them. Transformations, where no offset 
           is defined, are appended at the end."""
        # Create the new keyed transformation
        CombinedTransformation=CKeyedTransformation("");
        for Transformation in self:
            if(OffsetDict!=None and Transformation in OffsetDict):
                CombinedTransformation.Append(self[Transformation],OffsetDict[Transformation]);
            else:
                CombinedTransformation.Append(self[Transformation]);
            CombinedTransformation.Parent=self[Transformation].Parent;
        self.clear();
        self[""]=CombinedTransformation;
        return CombinedTransformation;

    def TransformTime(self,FrameOffset,FrameFactor):
        """This function invokes TransformTime(...) on all transformations in 
           this dictionary."""
        for Transformation in self.values():
            Transformation.TransformTime(FrameOffset,FrameFactor);
    
    def MakeDefaultKeys(self,DefaultTransformation,DefaultName="",DefaultParentTransformation=None):
        """This function ensures that this dictionary contains at least one 
           keyed transformation and that every keyed transformation has at 
           least one key of every type. The created keys use the transformation 
           of DefaultTransformation as values. If a new keyed transformation 
           gets created its key and its name are DefaultName and its parent 
           transformation is DefaultParentTransformation."""
        if(len(self)==0):
            self[DefaultName]=CKeyedTransformation(DefaultName,DefaultParentTransformation);
        Translation=DefaultTransformation.TranslationToVector();
        Scaling=DefaultTransformation.ScalingToVector();
        Rotation=DefaultTransformation.RotationToQuaternion();
        for Transformation in self.values():
            Transformation.MakeDefaultKeys(Translation,Rotation,Scaling);

    def IsConstant(self):
        """This function returns true, if and only if all keyed transformations in 
           this dictionary are constant. See CKeyedTransformation.IsConstant(...)."""
        Result=True;
        for Transformation in self.values():
            Result=(Result and Transformation.IsConstant());
        return Result;
            
    def FixRotationKeys(self):
        """This function invokes FixRotationKeys() on all keyed transformations in 
           this dictionary."""
        for Transformation in self.values():
            Transformation.FixRotationKeys();
            
    def Clone(self):
        """This function creates and returns a complete deep copy of this dictionary 
           of transformations."""
        Result=CKeyedTransformationDict();
        for TransformationName in self:
            Result[TransformationName]=self[TransformationName].Clone();
        return Result;


class CTriangleIndices:
    """This class holds three vertex indices, describing a triangle."""

    def __init__(self,A,B,C):
        """Constructs the triangle index set from three indices"""
        self.A=A;
        self.B=B;
        self.C=C;

    A=0;
    B=0;
    C=0;


class CVertex:
    """This class represents a single vertex. It always contains a position and a
       normal. Additionally it may contain skin weights with skinning indices and an
       arbitrary number of texture coordinates."""

    def __init__(self,VertexPosition=C3DVector(0.0,0.0,0.0),VertexNormal=C3DVector(0.0,0.0,0.0),VertexTextureCoordinate=None):
        """This constructor initializes this vertex with the given position, normal
           and optionally a single texture coordinate (may be None)."""
        self.Position=VertexPosition;
        self.Normal=VertexNormal;
        self.pTexCoord=list();
        if(VertexTextureCoordinate!=None):
            self.pTexCoord.add(VertexTextureCoordinate);
        self.pSkinWeight=list();
        self.piSkinBone=list();

    def Compare(self,RHS):
        """This function returns true, if and only if all data held by the given
           vertex matches the data of this vertex exactly."""
        Result=(self.Position.x==RHS.Position.x and self.Position.y==RHS.Position.y and self.Position.z==RHS.Position.z);
        Result=Result and (self.Normal.x==RHS.Normal.x and self.Normal.y==RHS.Normal.y and self.Normal.z==RHS.Normal.z);
        Result=Result and (len(self.pTexCoord)==len(RHS.pTexCoord));
        Result=Result and (len(self.pSkinWeight)==len(RHS.pSkinWeight));
        Result=Result and (len(self.piSkinBone)==len(RHS.piSkinBone));
        if(Result):
            for i in range(len(self.pTexCoord)):
                Result=Result and (self.pTexCoord[i].x==RHS.pTexCoord[i].x and self.pTexCoord[i].y==RHS.pTexCoord[i].y);
            for i in range(len(self.pSkinWeight)):
                Result=Result and (self.pSkinWeight[i]==RHS.pSkinWeight[i]);
            for i in range(len(self.piSkinBone)):
                Result=Result and (self.piSkinBone[i]==RHS.piSkinBone[i]);
        return Result;


    def ProcessInfluences(self,pInfluence,BoneList,MissingBoneSet=None):
        """This function processes the influence pairs in the given list (the 
           pairs have the form (BoneName,Weight)). It adds the influences to 
           the weight and index lists of this vertex using a descending weight 
           order and indices, which refer into the given CBone list. The given 
           influence list also gets sorted with descending weights. All weights 
           that have already existed, get deleted."""
        self.piSkinBone=list();
        self.pSkinWeight=list();
        # Sort the list of influences descending by the weights
        def GetInfluenceKey(Influence):
            return -Influence[1];
        pInfluence.sort(key=GetInfluenceKey);
        # Remove influences with a weight of zero
        for j in range(len(pInfluence)):
            if(pInfluence[j][1]<0.0001):
                pInfluence=pInfluence[0:j];
                break;
        # Add the influences to the weighted vertex
        for Influence in pInfluence:
            # Find the index of the right bone
            iBone=-1;
            for j in range(len(BoneList)):
                if(BoneList[j].Name==Influence[0]):
                    iBone=j;
                    break;
            if(iBone>=0):
                self.pSkinWeight.append(Influence[1]);
                self.piSkinBone.append(iBone);
            elif(MissingBoneSet!=None):
                MissingBoneSet|=frozenset([Influence[0]]);
        # Make sure that there is at least one weight
        if(len(self.pSkinWeight)==0):
            self.pSkinWeight.append(1.0);
            self.piSkinBone.append(0);
        # Normalize the weights
        InvTotalWeight=1.0/sum(self.pSkinWeight);
        for j in range(len(self.pSkinWeight)):
            self.pSkinWeight[j]*=InvTotalWeight;


    def LimitWeightCount(self,nMinWeight,nMaxWeight,iDefaultBone=0):
        """This function ensures that every vertex has at least the given minimal
           number of weights (by adding weights without weight and the given bone
           index) and at most the given maximal number of weights (by removing the
           weights with minimal influence)."""
        for i in range(len(self.pSkinWeight),nMinWeight):
            self.pSkinWeight.append(0.0);
            self.piSkinBone.append(iDefaultBone);
        if(len(self.pSkinWeight)>nMaxWeight):
            self.pSkinWeight=self.pSkinWeight[0:nMaxWeight];
            self.piSkinBone=self.piSkinBone[0:nMaxWeight];


    def Clone(self):
        """This function returns a complete deep copy of this vertex."""
        Result=CVertex(self.Position.Clone(),self.Normal.Clone());
        Result.pSkinWeight=list(self.pSkinWeight);
        Result.piSkinBone=list(self.piSkinBone);
        Result.pTexCoord=list();
        for TexCoord in self.pTexCoord:
            Result.pTexCoord.append(TexCoord.Clone());
        return Result;

    # The position of this vertex in local coordinates
    Position=C3DVector(0.0,0.0,0.0);
    # The normal vector of this vertex in local coordinates
    Normal=C3DVector(0.0,1.0,0.0);
    # An arbitrary number of skin weights for this vertex. The weights are normalized
    # so sum(pSkinWeight)==1.0. Besides they are sorted descending.
    pSkinWeight=list();
    # One bone index for every weight in pSkinWeight
    piSkinBone=list();
    # An arbitrary number of texture coordinates for this vertex (using 
    # C2DVector). The coordinate system used for texture coordinates differs 
    # from the one used by Blender internally. The coordinate (0.0,0.0) 
    # corresponds to the left top, (1.0,1.0) is the right bottom.  
    pTexCoord=list();


class CBranch:
    """This class represents a branch of an octree. A branch covers a cubic area and
       it holds either a list of appended vertices (if it is a leave branch) or a
       list of eight child branches."""

    def __init__(self,CubeCenterIn,CubeSizeIn,DepthIn=0):
        """This constructor initializes this branch without child branches or
           appended vertices. It makes it cover the given axis aligned cubic area."""
        # Save the input data
        self.CubeCenter=CubeCenterIn;
        self.CubeSize=CubeSizeIn;
        self.Depth=DepthIn;
        # Initialize the branch with no children and no vertices
        self.pChildBranch=[];
        self.piVertex=[];

    def Add(self,iVertex,pVertex):
        """This function adds the vertex with the given index to this branch. It
           fails changing nothing, if this branch is not a leaf branch! If the
           maximal number of vertices per branch is exceeded through this step, the
           branch is split up. For this reason it may need the vertex list."""
        if(not self.IsLeaf()):
            return False;
        # Add the vertex
        self.piVertex.append(iVertex);
        # If necessary split the branch
        if(len(self.piVertex)>16):
            self.Split(pVertex);
        return True;

    def Split(self,pVertex):
        """This function splits this branch, meaning that it converts it from a leaf
           branch to a normal branch. Its vertices are moved to the newly created
           child branches. For non-leaf branches it does nothing. If the maximal
           depth would get exceeded by another split, it does nothing. It needs the
           vertex list to look up vertex positions."""
        if(not self.IsLeaf() or self.Depth>=7):
            return False;
        # Create the eight child branches
        HalfCubeSize=self.CubeSize*0.5;
        P=C3DVector(self.CubeCenter.x+HalfCubeSize,self.CubeCenter.y+HalfCubeSize,self.CubeCenter.z+HalfCubeSize);
        N=C3DVector(self.CubeCenter.x-HalfCubeSize,self.CubeCenter.y-HalfCubeSize,self.CubeCenter.z-HalfCubeSize);
        self.pChildBranch=[
            CBranch(C3DVector(N.x,N.y,N.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(N.x,N.y,P.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(N.x,P.y,N.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(N.x,P.y,P.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(P.x,N.y,N.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(P.x,N.y,P.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(P.x,P.y,N.z),HalfCubeSize,self.Depth+1),
            CBranch(C3DVector(P.x,P.y,P.z),HalfCubeSize,self.Depth+1)
        ];
        # Pop the appended vertices from this branch and append them to the right
        # child branch
        while(len(self.piVertex)>0):
            iVertex=self.piVertex.pop();
            self.GetChildBranch(pVertex[iVertex].Position).Add(iVertex,pVertex);
        return True;

    def GetChildBranch(self,Position):
        """This function returns the child branch, which contains the given position.
           It does not test, whether the position is in this branch, so the result
           may be meaningless. For leaf branches it causes an exception!"""
        XPositive=Position.x>self.CubeCenter.x;
        YPositive=Position.y>self.CubeCenter.y;
        ZPositive=Position.z>self.CubeCenter.z;
        return self.pChildBranch[4*XPositive+2*YPositive+1*ZPositive];

    def IsLeaf(self):
        """This function returns True, if and only if this branch is a leaf branch."""
        return len(self.pChildBranch)==0;

    # This C3DVector gives the center of the cubic area covered by this branch
    CubeCenter=C3DVector(0.0,0.0,0.0);
    # This floating point value gives the edge length of the cubic area covered by
    # this branch
    CubeSize=0.0;
    # This integer saves the depth of this branch. It is needed to avoid endless
    # splitting.
    Depth=0;
    # This is a list of the children of this branch. For leaf branches it is empty
    # otherwise it gives exactly eight child branches.
    pChildBranch=[];
    # This is a list of the indices of the vertices which are appended to this
    # branch. It should be empty, if this branch is not a leaf branch.
    piVertex=[];


class COctree:
    """This class is used to sort vertices by their position. This makes it
       possible to find double vertices more efficiently. It is roughly
       an increasement from O(n^2) to O(n*log(n))."""

    def __init__(self,CubeCenter,CubeSize):
        """This constructor initializes this octree so that it can cover an axis 
           aligned cubic area with edge length CubeSize and its center at the given
           C3DVector. It associates it with the given vertex list. Vertex indices
           always refer to this list."""
        # Initialize the root branch to cover the given area
        self.Root=CBranch(CubeCenter,CubeSize);

    def FindBranch(self,Position):
        """This function finds the leaf branch, which covers the given position and
           returns it. It does not verify that the position is in this octree."""
        Result=self.Root;
        while(not Result.IsLeaf()):
            Result=Result.GetChildBranch(Position);
        return Result;

    # This is the root branch of the octree
    Root=None;


def GetDefaultTexCoord(Position,Normal,MinPosition,MaxPosition,iTexCoord,iVertex):
    """This function computes a default texture coordinate and returns it as
       C2DVector. Position is the vertex position as C3DVector, Normal is the normal
       vector of the vertex as C3DVector, MinPosition and MaxPosition define the
       bounding box of the mesh. iTexCoord gives the texture coordinate set index and
       iVertex gives the index of the vertex within the mesh."""
    LeftTop=C2DVector(MinPosition.x,MaxPosition.z);
    RightBottom=C2DVector(MaxPosition.x,MinPosition.z);
    Difference=RightBottom-LeftTop;
    Difference.x=max(0.001,abs(Difference.x));
    Difference.y=max(0.001,abs(Difference.y));
    return C2DVector(abs(Position.x-LeftTop.x)/Difference.x,abs(Position.y-LeftTop.y)/Difference.y);


def IdentityVertexIndex(iVertex,MeshData):
    """This function just returns its first parameter, i.e. it is the identity 
       function. It is used as index map for 
       CMeshData.CollectTextureCoordinates(...)."""
    return iVertex;
    
def ListedVertexIndex(iFace,iVertex,MeshData):
    """This function describes the mapping from index pairs to vertex indices, 
       that is achieved, by saving the vertices of one face after the other. It 
       is used as index map for CMeshData.CollectTextureCoordinates(...)."""
    return MeshData.piFaceFirstVertex[iFace]+iVertex;

class CMeshDataBase:
    """This is an abstract helper class of CMesh, which is used to provide 
       access to mesh data in a way that is independent from a particular 
       Blender version."""
    
    def __init__(self):
        """This constructor initializes this mesh data in an unlocked state 
           with default data."""
        self.nLock=0;
        self.piFaceFirstVertex=None;
        self.SourceList=list();

    def Lock(self):
        """This function binds this object to the associated Blender object. 
           It may cause temporary changes. Do not forget to call 
           UnlockMeshData() to revert these changes! The functions count their 
           calls. It returns true, if and only if the operation was 
           successful."""
        Result=True;
        if(self.nLock==0):
            Result=self.DoLock();
        if(Result):
            self.nLock+=1;
            self.PrepareVertexIndexList();
        return Result;
        
    def DoLock(self):
        """Implements behavior of LockMeshData() that is specific for derived 
           classes. It returns true, if and only if the operation was 
           successful."""
        self.MarkAbstract();
    
    def Unlock(self):
        """This function unbinds this object from the associated Blender 
           object. All temporary changes are reverted."""
        self.nLock-=1;
        if(self.nLock==0):
            self.DoUnlock();
        self.nLock=max(0,self.nLock);

    def DoUnlock(self):
        """Implements behavior of UnlockMeshData() that is specific for derived 
           classes."""
        self.MarkAbstract();
        
    def GetName(self):
        """This function returns the name of the associated mesh."""
        self.MarkAbstract();
    
    def UsesSkinning(self):
        """This function returns true, if and only if the associated mesh uses 
           skinning."""
        self.MarkAbstract();
    
    def UsesFaceVertices(self):
        """This function returns true, if and only if it is necessary to assume 
           that a single vertex may have different values (normals and/or 
           texture coordinates) for different faces. If it returns true you 
           should use GetFaceVertex(...) instead of GetVertex(...)."""
        self.MarkAbstract();

    def GetTexCoordSetCount(self):
        """This function returns the number of texture coordinate sets that can 
           be accessed through this object."""
        self.MarkAbstract();

    def GetVertexCount(self):
        """This function returns the number of vertices in this mesh. It is 
           barely relevant if UsesFaceVertices() returns true."""
        self.MarkAbstract();

    def GetVertex(self,iVertex):
        """This function returns a new CVertex holding the data of the vertex with 
           the given index in this mesh. Skinning weights, skinning indices and 
           texture coordinates are NOT included."""
        self.MarkAbstract();
        
    def GetFaceVertexIndex(self,iFace,iFaceVertex):
        """Given the indices of a face and one of the vertices of this face 
           this function returns the index (for GetVertex(...)) at which the 
           corresponding vertex may be found. If USesFaceVertices(...) returns 
           true the return value is irrelevant."""
        self.MarkAbstract();
    
    def GetFaceCount(self):
        """This function returns the number of faces in this mesh."""
        self.MarkAbstract();

    def GetFaceVertexCount(self,iFace):
        """This function returns the number of vertices of the face with the 
           given index."""
        self.MarkAbstract();

    def GetFaceVertex(self,iFace,iFaceVertex):
        """This function returns a new CVertex holding the data of the vertex with 
           the given index of the face with the given index in this mesh. Skinning 
           weights, skinning indices and texture coordinates are NOT included."""
        self.MarkAbstract();
        
    def GetFaceMaterial(self,iFace):
        """This function returns the index of the material that is to be used 
           for the face with the given index."""
        self.MarkAbstract();
    
    def CollectTextureCoordinates(self,pVertex,GetVertexIndex=None):
        """This function adds the texture coordinates of this mesh to the 
           vertices in pVertex. The parameter GetVertexIndex defines how 
           indices for the array pVertex are to be computed. If 
           self.UsesFaceVertices() returns True it has to map integer 
           parameters iFace and iVertex and self to a pVertex index. Otherwise 
           it has to map a single iVertex integer and self to a pVertex index. 
           In any case the function may return -1 to indicate a non-existent 
           vertex. If you pass None either IdentityVertexIndex(iVertex,...) or 
           ListedVertexIndex(iFace,iVertex,...) will be used (dependent on 
           self.UsesFaceVertices())."""
        if(GetVertexIndex==None):
            if(self.UsesFaceVertices()):
                GetVertexIndex=ListedVertexIndex;
            else:
                GetVertexIndex=IdentityVertexIndex;
        self.DoCollectTextureCoordinates(pVertex,GetVertexIndex);
    
    def DoCollectTextureCoordinates(self,pVertex,GetVertexIndex):
        """Implements CollectTextureCoordinates(...). GetVertexIndex must not 
           be None."""
        self.MarkAbstract();
        
    def CollectWeights(self,pVertex,GetVertexIndex=None):
        """If necessary this function collects weights of this mesh and adds 
           them to the data of the vertices in pVertex. The GetVertexIndex 
           parameter is equivalent to the one of 
           CollectTextureCoordinates(...). It prints warnings, if it does not 
           find some of the referred bones."""
        if(GetVertexIndex==None):
            if(self.UsesFaceVertices()):
                GetVertexIndex=ListedVertexIndex;
            else:
                GetVertexIndex=IdentityVertexIndex;
        MissingBoneSet=set();
        self.DoCollectWeights(pVertex,GetVertexIndex,MissingBoneSet);
        if(len(MissingBoneSet)>0):
            print("WARNING: The following bones are referenced by weights of the mesh \""+self.GetName()+"\", but could not be found. The weights have been ignored.");
            for MissingBone in MissingBoneSet:
                print(MissingBone);

    def DoCollectWeights(self,pVertex,GetVertexIndex,MissingBoneSet=None):
        """Implements CollectWeights(...). GetVertexIndex must not be None. 
           MissingBoneSet may be a set that will be filled with the names of 
           bones, that are referenced, but were not found."""
        self.MarkAbstract();
    
    def MarkAbstract(self):
        """This function throws a not implemented exception. It is used to 
           indicate "abstract" functions."""
        raise NotImplemented("CMeshDataBase is abstract, use derived classes to represent actual mesh data.");
    
    def PrepareVertexIndexList(self):
        """This function creates self.piFaceFirstVertex, if this has not 
           happened yet."""
        if(self.piFaceFirstVertex==None):
            self.piFaceFirstVertex=[0]*self.GetFaceCount();
            iVertex=0;
            for iFace in range(self.GetFaceCount()):
                self.piFaceFirstVertex[iFace]=iVertex;
                iVertex+=self.GetFaceVertexCount(iFace);
    
    # This function counts how many calls to Lock() have not been followed by 
    # calls to Unlock()
    nLock=0;
    # This list maps face indices to the number of vertices in faces with lower 
    # index
    piFaceFirstVertex=None;
    
    # A list of Blender objects, which function as data source for this mesh 
    # data
    SourceList=list();


class CMesh:
    """This class is used to collect and export mesh data"""

    def __init__(self,iMesh,MeshData):
        """This constructor collects mesh data from the given object (instance of 
           CMeshData). Double vertices are removed, the given index is saved. If 
           the given data does not lead to a mesh with geometry this object is 
           placed in an invalid state. See self.IsValid()."""
        self.CleanUp();
        # Lock the mesh data
        if(MeshData==None or not MeshData.Lock()):
            return;
        # Check whether there are any triangles. Otherwise abort.
        HasTriangle=False;
        for iFace in range(MeshData.GetFaceCount()):
            if(MeshData.GetFaceVertexCount(iFace)>=3):
                HasTriangle=True;
                break;
        if(not HasTriangle):
            return;
        # Copy some meta data
        self.Name=MeshData.GetName();
        self.SkinningUsed=MeshData.UsesSkinning();
        self.nTexCoordSet=MeshData.GetTexCoordSetCount();
        # Remember the source mesh
        self.SourceList.extend(MeshData.SourceList);
        # Save the mesh index
        self.iMesh=iMesh;
        # There are two different cases when it comes to mapping Blender vertex
        # indices to the indices used in the output:
        # 1) They are identical (if MeshData.UsesFaceVertices()==False)
        # 2) The output indices are the unique face vertex indices. Face vertex
        #    indices can be obtained by checking how many vertices are used by faces
        #    with smaller index or earlier in the same face. To get to indices of
        #    unique face vertices you need to map face vertex indices through
        #    piFaceVertex. Multiple face vertex indices may correspond to the same 
        #    Blender vertex index and a single unique face vertex may refer to
        #    multiple Blender vertex indices (if MeshData.UsesFaceVertices()==True).
        # All of this is necessary because different faces may associate different
        # texture coordinates and normals with a single Blender vertex.
        piFaceVertex=[];
    
        # The bounding box may be needed to remove double vertices efficiently using
        # an octree. Initialize vectors to determine it.
        BigValue=1.0e20;
        MinVec=C3DVector( BigValue, BigValue, BigValue);
        MaxVec=C3DVector(-BigValue,-BigValue,-BigValue);
    
        # If the texture coordinates and the normals are saved per vertex the
        # process is simple
        if(not MeshData.UsesFaceVertices()):
            for i in range(MeshData.GetVertexCount()):
                # Collect the data of the vertex and add it to the list
                NewVertex=MeshData.GetVertex(i);
                self.pVertex.append(NewVertex);
                # Update the bounding box
                MinVec.Minimize(NewVertex.Position);
                MaxVec.Maximize(NewVertex.Position);
            self.nVertex=len(self.pVertex);
            # Get the texture coordinates
            MeshData.CollectTextureCoordinates(self.pVertex);
    
        # If the texture coordinates or the normals are saved on per triangle
        # base get them from there
        else:
            # This case is a little complicated. One vertex gets created for every 
            # face vertex. Then the list gets reduced to unique combinations of
            # positions, texture coordinates and normals. During this process a
            # lookup table for creating the index buffer gets created.
    
            # For all faces iterate through all vertices and create a list of them
            pFaceVertex=[];
            for i in range(MeshData.GetFaceCount()):
                for j in range(MeshData.GetFaceVertexCount(i)):
                    NewVertex=MeshData.GetFaceVertex(i,j);
                    pFaceVertex.append(NewVertex);
                    # Update the bounding box
                    MinVec.Minimize(NewVertex.Position);
                    MaxVec.Maximize(NewVertex.Position);
            # Get the texture coordinates from all relevant UV layers
            MeshData.CollectTextureCoordinates(pFaceVertex);
            # Now remove double vertices
            # Inform the client
            #print("Removing double vertices for the mesh \""+self.Name+"\".");
            # Create an octree, which can contain the entire bounding box. All
            # unique vertices are added to it. This way it is possible to
            # determine uniqueness of vertices fast.
            BBoxSize=C3DVector(MaxVec.x-MinVec.x,MaxVec.y-MinVec.y,MaxVec.z-MinVec.z);
            Octree=COctree(C3DVector(MinVec.x+BBoxSize.x*0.5,MinVec.y+BBoxSize.y*0.5,MinVec.z+BBoxSize.z*0.5),max(BBoxSize.x,BBoxSize.y,BBoxSize.z));
            # Cycle through all vertices and check whether they equal one of the
            # unique vertices found so far. If they do not append them.
            nUniqueVertex=0;
            for FaceVertex in pFaceVertex:
                # Search the vertex in the octree
                Found=False;
                Branch=Octree.FindBranch(FaceVertex.Position);
                for i in Branch.piVertex:
                    if(FaceVertex.Compare(pFaceVertex[i])):
                        # Use the existing identical vertex
                        piFaceVertex.append(i);
                        Found=True;
                        break;
                if(not Found):
                    # A new unique vertex has been found. Add it to the list and
                    # to the octree. The right branch is already known.
                    pFaceVertex[nUniqueVertex]=FaceVertex;
                    piFaceVertex.append(nUniqueVertex);
                    Branch.Add(nUniqueVertex,pFaceVertex);
                    nUniqueVertex+=1;
            #print("Removed "+str(len(pFaceVertex)-nUniqueVertex)+" vertices ("+str(100-int((100.0*nUniqueVertex)/len(pFaceVertex)))+"% and "+str(nUniqueVertex)+" remain).");
            # Reducing the list has been done in the list. Slice away the
            # remaining duplicate vertices and set it up as mesh vertex list.
            self.pVertex=pFaceVertex[0:nUniqueVertex];
            self.nVertex=nUniqueVertex;
    
        # Get the triangle count and the triangles with their material indices
        iFaceVertex=0;
        UsesFaceVertices=MeshData.UsesFaceVertices();
        for iFace in range(MeshData.GetFaceCount()):
            nFaceVertex=MeshData.GetFaceVertexCount(iFace);
            # See the comment of piFaceVertex
            piVertex=[];
            if(not UsesFaceVertices):
                for iFaceVertex in range(nFaceVertex):
                    # Use Blender indices
                    piVertex.append(MeshData.GetFaceVertexIndex(iFace,iFaceVertex));
            else:
                for i in range(nFaceVertex):
                    # Use indices of unique face vertices
                    piVertex.append(piFaceVertex[iFaceVertex+i]);
            # Put the three or four face vertex indices together to one or two
            # triangles (the code is ready for faces with more vertices)
            iMaterial=MeshData.GetFaceMaterial(iFace);
            for i in range(1,nFaceVertex-1):
                self.pTriangle.append(CTriangleIndices(piVertex[0],piVertex[i],piVertex[i+1]));
                self.piTriangleMaterial.append(iMaterial);
                self.nTriangle+=1;
            iFaceVertex+=nFaceVertex;
    
        # Get the skinning weights if necessary
        if(self.SkinningUsed):
            # Define the relation of indices
            GetVertexIndex=None;
            if(UsesFaceVertices):
                def GetFaceVertexIndex(iFace,iFaceVertex,MeshData):
                    return piFaceVertex[MeshData.piFaceFirstVertex[iFace]+iFaceVertex];
                GetVertexIndex=GetFaceVertexIndex;
            # Get the weight data
            MeshData.CollectWeights(self.pVertex,GetVertexIndex);
            # Find the maximal used bone index
            self.iMaxBone=0;
            for Vertex in self.pVertex:
                for iBone in Vertex.piSkinBone:
                    self.iMaxBone=max(iBone,self.iMaxBone);
        # Unlock the mesh data
        MeshData.Unlock();

    def ApplyTransformation(self,Transformation):
        """This function applies the given CMatrix4x4 transformation to the position
           and normal of every single vertex in this mesh. If the 3x3 part of the
           matrix has a negative determinant it also flips all triangles."""
        for Vertex in self.pVertex:
            Vertex.Position*=Transformation;
            Vertex.Normal=Transformation.RotateAndScaleVector(Vertex.Normal);
            Vertex.Normal.Normalize();
        if(Transformation.Compute3x3Determinant()<0.0):
            self.FlipTriangles();

    def FlipTriangles(self):
        """This function reverts the order of triangle indices in all triangles
           of this mesh. The normals remain uneffect and thus become inconsistent
           with the triangles. It is recommended to use this function, if you
           intend to use the mesh with a negative scaling."""
        for Triangle in self.pTriangle:
            A=Triangle.A;
            Triangle.A=Triangle.C;
            Triangle.C=A;

    def AppendMesh(self,MeshObject):
        """This function appends the given mesh object to this mesh object. It does
           NOT create deep copies of the vertex data and the triangle indices, so
           changing the data of MeshObject will effect the corresponding part of this
           mesh."""
        self.SourceList+=MeshObject.SourceList;
        # Combine the names
        self.Name=self.Name+"+"+MeshObject.Name;
        # Combine the triangle lists
        for i in range(MeshObject.nTriangle):
            self.pTriangle.append(CTriangleIndices(MeshObject.pTriangle[i].A+self.nVertex,MeshObject.pTriangle[i].B+self.nVertex,MeshObject.pTriangle[i].C+self.nVertex));
            self.piTriangleMaterial.append(MeshObject.piTriangleMaterial[i]);
        self.nTriangle+=MeshObject.nTriangle;
        # Combine the vertex lists
        self.pVertex.extend(MeshObject.pVertex);
        self.nVertex+=MeshObject.nVertex;


    def UpdateSkinningIndices(self,piBone):
        """This function updates the skinning indices in all vertices of this mesh.
           The index i is replaced by the index piBone[i]. If piBone has negative
           entries the corresponding weights get deleted."""
        for Vertex in self.pVertex:
            i=0;
            while(i<len(Vertex.piSkinBone)):
                if(piBone[Vertex.piSkinBone[i]]<0):
                    del Vertex.piSkinBone[i];
                    del Vertex.pSkinWeight[i];
                else:
                    Vertex.piSkinBone[i]=piBone[Vertex.piSkinBone[i]];
                    i+=1;
            # Weights might have been deleted. Renormalize the weights.
            TotalWeight=sum(Vertex.pSkinWeight);
            if(TotalWeight>0.0001):
                for i in range(len(Vertex.pSkinWeight)):
                    Vertex.pSkinWeight[i]/=TotalWeight;
        self.iMaxBone=0;
        if(len(piBone)>0):
            self.iMaxBone=max(piBone);


    def SkinToSingleBone(self,iBone):
        """This function gives every vertex exactly one weight with strength 1.0 and
           bone index iBone. Old weights (if any) are overwritten. SkinningUsed is
           set to True."""
        self.SkinningUsed=True;
        self.iMaxBone=iBone;
        for Vertex in self.pVertex:
            Vertex.pSkinWeight=[1.0];
            Vertex.piSkinBone=[iBone];


    def MakeDefaultTexCoords(self,nTexCoordSet,TexCoordGetter=GetDefaultTexCoord):
        """This function adds default texture coordinates to this mesh until it has
           the given number of texture coordinate sets. If the mesh already has more
           texture coordinate sets it does not do anything. The given function is
           used to compute texture coordinate sets. See GetDefaultTexCoord(...) for
           the signature."""
        if(nTexCoordSet>self.nTexCoordSet):
            # Get the bounding box
            MinVec=C3DVector( 1.175494351e-38, 1.175494351e-38, 1.175494351e-38);
            MaxVec=C3DVector(-1.175494351e-38,-1.175494351e-38,-1.175494351e-38);
            for Vertex in self.pVertex:
                MinVec.Minimize(Vertex.Position);
                MaxVec.Maximize(Vertex.Position);
            # Add the texture coordinate sets
            for i in range(self.nTexCoordSet,nTexCoordSet):
                iVertex=0;
                for Vertex in self.pVertex:
                    Vertex.pTexCoord.append(TexCoordGetter(Vertex.Position,Vertex.Normal,MinVec,MaxVec,i,iVertex));
                    iVertex+=1;
            self.nTexCoordSet=max(self.nTexCoordSet,nTexCoordSet);


    def Clone(self):
        """This function creates a new CMesh, which is completely identical to this
           one. The result is returned. It is a complete deep copy."""
        # Create a new, empty, invalid mesh for the result
        Result=CMesh(None,None);
        Result.SourceList=list(self.SourceList);
        # Copy the contents
        Result.iMesh=self.iMesh;
        Result.Name=self.Name;
        Result.nVertex=self.nVertex;
        Result.nTriangle=self.nTriangle;
        Result.nTexCoordSet=self.nTexCoordSet;
        for Vertex in self.pVertex:
            Result.pVertex.append(Vertex.Clone());
        Result.SkinningUsed=self.SkinningUsed;
        Result.iMaxBone=self.iMaxBone;
        for Triangle in self.pTriangle:
            Result.pTriangle.append(CTriangleIndices(Triangle.A,Triangle.B,Triangle.C));
        for iMaterial in self.piTriangleMaterial:
            Result.piTriangleMaterial.append(iMaterial);
        return Result;


    def LimitWeightCount(self,nMinWeight,nMaxWeight,iDefaultBone=0):
        """This function ensures that every vertex has at least the given minimal
           number of weights (by adding weights without weight and the given bone
           index) and at most the given maximal number of weights (by removing the
           weights with minimal influence)."""
        if(nMinWeight>nMaxWeight):
            print("Invalid call to CMesh.LimitWeightCount(...).");
            return;
        for Vertex in self.pVertex:
            Vertex.LimitWeightCount(nMinWeight,nMaxWeight,iDefaultBone);


    def GetMaxBoneWeightList(self):
        """This function returns a list holding self.iMaxBone float objects in the
           range from 0.0 to 1.0. For every bone index it gives the maximal influence
           that this bone has on a vertex in this mesh."""
        pBoneMaxWeight=[0.0]*(self.iMaxBone+1);
        for Vertex in self.pVertex:
            for i in range(len(Vertex.piSkinBone)):
                pBoneMaxWeight[Vertex.piSkinBone[i]]=max(pBoneMaxWeight[Vertex.piSkinBone[i]],Vertex.pSkinWeight[i]);
        return pBoneMaxWeight;


    def CleanUp(self):
        """This function overwrites all data of this mesh with new objects holding
           default values. This way it makes the instance become independent from
           other instances of the same class."""
        self.iMesh=0;
        self.Name="";
        self.nVertex=self.nTriangle=0;
        self.nTexCoordSet=0;
        self.pVertex=[];
        self.pTriangle=[];
        self.piTriangleMaterial=[];
        self.SkinningUsed=False;
        self.iMaxBone=0;
        self.SourceList=[];

    def IsValid(self):
        """This function returns True if and only if this object describes a mesh
           with at least one triangle in a valid way."""
        return (self.iMesh>=0 and self.nVertex>0 and self.nTriangle>0 and self.nTexCoordSet>=0
            and self.pVertex!=None and self.pTriangle!=None and self.piTriangleMaterial!=None
            and len(self.pVertex)>=self.nVertex and len(self.pTriangle)>=self.nTriangle
            and len(self.piTriangleMaterial)>=self.nTriangle);

    # The index of this mesh
    iMesh=0;
    # The name of this mesh
    Name="";

    # The number of vertices and triangles
    nVertex=0;
    nTriangle=0;

    # This integer saves the number of texture coordinate sets available in this mesh
    nTexCoordSet=0;

    # The array of vertices
    pVertex=[];
    # If this boolean is True the vertices have three skinning weights and four
    # skinning indices. Otherwise they have neither weights nor indices.
    SkinningUsed=False;
    # This integer saves the highest referenced bone index (according to the skinning
    # indices)
    iMaxBone=0;

    # The triangle index array
    pTriangle=[];
    # The triangle material index array
    piTriangleMaterial=[];

    # This is a list of Blender Mesh objects, which function as source for the data
    # in this object
    SourceList=[];


class CObject:
    """This class represents an object in a scene. It is not necessarily visible. It
       corresponds directly to a Blender Object. The object might have a still or an
       animated transformation and a parent."""

    # This function is implemented in both Blender specific modules
    #def __init__(self,Object,ParentObject=None,SkeletonForSkinning=True,DuplicationTransformation=CMatrix4x4()):
    #    """This function creates a CObject, from a Blender Object. It uses the 
    #       given CObject as parent. The name of the Blender Object is used and the 
    #       transformation is initialized using the world space transformation of 
    #       the Object with regards to the given duplication transformation. 
    #       SkeletonForSkinning indicates whether a CSkeleton, that is parent should 
    #       or should not be used for skinning."""

    def GetTransformation(self,ActionName=None,Frame=0.0):
        """This function returns the complete object to model space transformation 
           of this object. This means that the returned transformation includes the
           transformation of all ancestors. If you pass a valid name of an action, 
           which has been read for this object, all transformations are evaluated 
           for this action at the given frame. Otherwise the current transformation 
           in Blender is used."""
        if(ActionName!=None and ActionName in self.TransformationDict):
            Result=self.TransformationDict[ActionName].GetFrameTransformation(Frame);
        else:
            Result=self.Transformation.Clone();
        if(self.Parent!=None):
            if(self.ParentBone==None):
                Result*=self.Parent.GetTransformation(ActionName,Frame);
            else:
                Result*=self.ParentBone.GetTransformation(self.Parent,ActionName,Frame);
        return Result;
    
    def IsAnimated(self):
        """This function returns true, if and only if the keyed transformation of 
           this object is not constant and it does not use skinning."""
        TransformationAnimated=(not self.TransformationDict.IsConstant());
        return (TransformationAnimated 
                and (not self.ParentForSkinning or not isinstance(self.Parent,CSkeleton)));

    def SetParentBone(self,ParentBone):
        """This function sets up a parent bone for this object. It is important to
           use it instead of setting self.ParentBone directly, because it makes the
           transformation relative to the bone. The parent bone has to be part of the
           skeleton, which is the parent of this object. ParentBone has to be a valid
           bone. It returns True on success and False on failure."""
        if(not isinstance(self.Parent,CSkeleton)):
            print("WARNING: Trying to set up a parent bone, for an object which does not have a skeleton as parent.");
            return False;
        if(self.ParentBone!=None and ParentBone==None):
            print("WARNING: Trying to remove a parent bone.");
            return False;
        # Get the current model space transformation of the object and the bone
        CurrentTransformation=self.GetTransformation();
        InvCurrentBoneTransformation=ParentBone.GetTransformation().ComputeInverse();
        if(InvCurrentBoneTransformation==None):
            print("WARNING: A bone that is being used as parent has a singular transformation (scaling zero). It will not be used as parent bone.");
            return False;
        # Compute the new transformation and set the parent
        self.Transformation=CurrentTransformation*InvCurrentBoneTransformation;
        self.ParentBone=ParentBone;
        if(self.ParentBone!=None):
            self.ParentForSkinning=False;
        return True;
    
    def AddKeysFromState(self,ActionName,Frame):
        """This function adds keys, which describe the current transformation of the 
           associated Blender Object correctly. The keys are stored in 
           self.TransformationDict using ActionName and Frame as time value."""
        # Get the current object to parent transformation of this object 
        CurrentTransformation=self.GetCurrentTransformation();
        if(CurrentTransformation==None):
            CurrentTransformation=self.Transformation;
        # Get the keyed transformation of this object and create the keys
        KeyedTransformation=self.GetKeyedTransformation(ActionName);
        KeyedTransformation.AddTransformationKey(Frame,CurrentTransformation);
        
    def GetKeyedTransformation(self,TransformationKey):
        """This function returns the keyed transformation for the object to parent 
           space transformation of this object, which is stored under the given key. 
           If it does not exist, it is created."""
        # Get the keyed transformation of this object and create the keys
        return self.TransformationDict.GetKeyedTransformation(TransformationKey);
        
    def FixRotationKeys(self):
        """This function forwards to FixRotationKeys() of the keyed transformation 
           dictionary of this object."""
        self.TransformationDict.FixRotationKeys();

    # This function is implemented in both Blender specific modules
    #def GetCurrentTransformation(self):
    #    """This function returns the current object to parent transformation of the 
    #       associated Blender Object (first SourceList entry) of this object as 
    #       CMatrix4x4. On failure (e.g. if the SourceList is empty) it returns 
    #       None."""

    # This function is implemented in both Blender specific modules
    #def AddKeysFromActionCurves(self,ActionName,MaxKeyDistance=0.0):
    #    """This function tries to locate animation curves for the object 
    #       transformation in an action with the given name. If it finds them, it 
    #       generates corresponding keys for this object and stores them in the 
    #       transformation dictionary using the ActionName as key. If it fails to 
    #       find (some) curves it generates keys, which match the default 
    #       transformation instead. The created keys have at least the given 
    #       distance (unless the distance is not positive)."""

    # This function is implemented in both Blender specific modules
    #def AddKeysFromAllActionCurves(self,MaxKeyDistance=0.0):
    #    """This function forwards to AddKeysFromActionCurves(...) using all 
    #       existing action names as parameters."""

    def ApplyTransformation(self):
        """This function applies the current local transformation of this object to 
           the associated data and sets the transformation to identity. This should 
           not be used if this object has a child or an animation."""
        self.ApplyDataTransformation(self.Transformation);
        self.Transformation=CMatrix4x4();

    def ApplyDataTransformation(self,Transformation):
        """This function applies the given transformation to the data that is
           associated with this object."""
        pass;

    def CopyObject(self,RHS):
        """This function makes the CObject part of self become an exact deep copy of
           the CObject part of RHS."""
        self.Name=RHS.Name;
        self.Transformation=RHS.Transformation.Clone();
        self.TransformationDict=RHS.TransformationDict.Clone();
        self.Parent=RHS.Parent;
        self.ParentBone=RHS.ParentBone;
        self.ParentForSkinning=RHS.ParentForSkinning;
        self.SourceList=list(RHS.SourceList);

    def CloneObject(self):
        """This function creates a deep copy of the CObject part of this object and
           returns a copy to it. The returned object is always just a CObject, even
           if the function is invoked on a derived class. It is useful for
           eliminating objects, which are still needed as parent."""
        Result=CObject(None);
        Result.CopyObject(self);
        return Result;

    def Clone(self):
        """This function creates a deep copy of this object and returns it. Derived
           classes have to implement it."""
        return self.CloneObject();

    def CompareType(self,Object):
        """This function returns true if and only if the given object has the same
           type as this object."""
        return isinstance(Object,CObject);

    def IsPureObject(self):
        """This function returns True if and only if this object is only a CObject
           not an instance of some derived type."""
        return True;

    def AppendObjectData(self,Object):
        """This function makes this object become an object that also contains the
           data of the given other object (needed as CObject). For this to work the
           relative transformation between self and Object must not be animated.
           Besides the type of self has to be the same as the type of Object. If the 
           process fails the return value is False, otherwise it is True. Note that 
           this function may create new meshes and bones. Call CModel.UpdateMeshList
           after using it."""
        # Check whether the types are compatible
        if(not self.CompareType(Object)):
            return False;
        # Find the first shared ancestor
        AncestorSet=set([self]);
        Ancestor=self;
        while(Ancestor.Parent!=None):
            Ancestor=Ancestor.Parent;
            AncestorSet|=set([Ancestor]);
        Ancestor=Object;
        while(Ancestor!=None and not (Ancestor in AncestorSet)):
            Ancestor=Ancestor.Parent;
        # Ancestor is the shared Ancestor now. Check whether the way to it contains
        # a parent bone. If so it is animated and the function has to return false.
        NextAncestor=self;
        while(NextAncestor!=Ancestor):
            if(NextAncestor.ParentBone!=None or isinstance(NextAncestor.Parent,CSkeleton)):
                return False;
            NextAncestor=NextAncestor.Parent;
        NextAncestor=Object;
        while(NextAncestor!=Ancestor):
            if(NextAncestor.ParentBone!=None):
                return False;
            NextAncestor=NextAncestor.Parent;
        # Get the (constant) relative transformation between the two objects.
        WorldToSelfSpace=self.GetTransformation().ComputeInverse();
        if(WorldToSelfSpace==None):
            print("WARNING: Merging objects failed, because of a singular transformation (scaling zero) in the object \""+self.Name+"\".");
            return False;
        ObjectToSelfSpace=Object.GetTransformation()*WorldToSelfSpace;
        # Do the actual appending
        if(self.DoAppendObjectData(self,Object,ObjectToSelfSpace)):
            # Update the source list
            self.SourceList.append(Object.SourceList);
            return True;
        return False;

    def DoAppendObjectData(self,Object,ObjectToSelfSpace):
        """This function implements AppendObjectData(...) by performing object type 
           specific steps."""
        # For pure CObject instances there is nothing to do. Derived classes
        # override this function.
        pass;


    def CleanUp(self):
        """This function brings this instance into an invalid, "empty" state."""
        self.Name="";
        self.Transformation=CMatrix4x4();
        self.TransformationDict=CKeyedTransformationDict();
        self.Parent=None;
        self.ParentBone=None;
        self.ParentForSkinning=False;
        self.SourceList=list();

    # This is the name of this object
    Name="";
    # This CMatrix4x4 saves the object to parent space transformation of this 
    # object. If there is no parent it is the object to world space transformation.
    # Note that this matrix only gives the transformation for the current state of 
    # the object in Blender. The animated transformation is stored in 
    # TransformationDict.
    Transformation=CMatrix4x4();
    # This CKeyedTransformationDict holds keyed transformations for different 
    # animations of this object. The transformations are object to parent or object 
    # to world space transformations (dependent on the existence of a parent). They 
    # are NOT linked to parent transformations properly.
    TransformationDict=CKeyedTransformationDict();
    # If this object has a parent object this is a reference to the corresponding
    # CObject. If the parent is a bone, then this is a reference to the CObject for 
    # the skeleton, that holds this bone.
    Parent=None;
    # If the parent of this object is a bone this is a reference to the corresponding
    # CBone. Otherwise it is None.
    ParentBone=None;
    # If the parent is a CSkeleton and there is no parent bone, the skeleton might
    # either be used for skinning or it might be used as usual parent such that only
    # the CObject part is regarded. This boolean is True, if and only if it is used
    # for skinning.
    ParentForSkinning=False;

    # This is a list containing the objects, which function as source for this 
    # instance
    SourceList=[];
    

class CMeshObject(CObject):
    """This class represents an instance of a mesh. It is useful, because a single
       mesh can be used by multiple objects."""

    def ApplyDataTransformation(self,Transformation):
        """This function applies the given transformation to the mesh of this object."""
        if(self.Mesh!=None):
            self.Mesh.ApplyTransformation(Transformation);

    def CompareType(self,Object):
        """This function returns true if and only if the given object has the same
           type as this object."""
        return isinstance(Object,CMeshObject);

    def IsPureObject(self):
        """This function returns True if and only if this object is only a CObject
           not an instance of some derived type."""
        return False;

    def DoAppendObjectData(self,Object,ObjectToSelfSpace):
        """This function makes this object become an object that also contains the
           data of the given other object (needed as CMeshObject). To do so it
           creates a new mesh with the name of this mesh object. The relative
           transformation between self and Object has to be constant and it has to be
           given as CMatrix4x4. If the process fails the return value is False,
           otherwise it is True."""
        self.Mesh=self.Mesh.Clone();
        self.Mesh.Name=self.Name;
        self.Mesh.AppendMesh(Object.Mesh.Clone());
    
    def Clone(self):
        """This function creates a deep copy of this object and returns it."""
        Result=CMeshObject(None,None,None,None,None,None);
        Result.CopyObject(self);
        Result.Mesh=self.Mesh.Clone();
        return Result;

    # This is a reference to the CMesh, which is used for this mesh object
    Mesh=None;


class CBlendMode:
    """This class provides a representation of blend modes that does is not 
       specific to a particular version of Blender. It supports """

    def __init__(self,BlendModeNameOrBlendMode):
        """This function initializes this blend mode from the given value. It 
           has to be either the name of a blend mode in capital letters or 
           another blend mode that is to be copied. If it is not the result 
           will be an invalid blend mode."""
        RHS=BlendModeNameOrBlendMode;
        if(isinstance(RHS,CBlendMode)):
            self.iBlendMode=RHS.iBlendMode;
        else:
            try:
                self.iBlendMode=self.BlendModeList.index(RHS);
            except:
                self.iBlendMode=-1;
                
    def GetName(self):
        """This function returns the name that is associated with this blend 
           mode or None, if this blend mode is invalid."""
        if(self.IsValid()):
            return self.BlendModeList[self.iBlendMode];
        
    def __str__(self):
        """Same as GetName(...), but returns an empty string instead of None."""
        if(self.IsValid()):
            return self.BlendModeList[self.iBlendMode];
        else:
            return "";

    def __eq__(self,RHS):
        """This operator returns true, if and only if this blend mode is valid 
           and refers the same blend mode as RHS."""
        return self.IsValid() and isinstance(RHS,CBlendMode) and self.iBlendMode==RHS.iBlendMode;

    def IsValid(self):
        """This function returns true, if and only if this object describes a 
            blend mode."""
        return self.iBlendMode>=0 and self.iBlendMode<len(self.BlendModeList);
    
    # The index for this blend mode
    iBlendMode=-1;
    # This is a list of all existing blend modes in all supported versions of 
    # Blender. It is a static member.
    BlendModeList=list([
        "MIX","ADD","SUBTRACT","MULTIPLY",
        "SCREEN","OVERLAY","DIFFERENCE","DIVIDE",
        "DARKEN","LIGHTEN",
        "HUE","SATURATION","VALUE",
        "COLOR","SOFT LIGHT","LINEAR LIGHT"]);


class CMaterialDataBase:
    """This is class gives access to material data of Blender 2.49."""
       
    def __init__(self):
        """This function initializes this base class with default data."""
        self.SourceList=list();
        self.nLock=0;
        self.iLockedTexture=-1;
    
    def GetName(self):
        """This function retrieves the name of the associated material."""
        self.MarkAbstract();
    
    def GetColorInfo(self):
        """This function returns a 5-tuple of the form (AmbientColor, 
           DiffuseColor, EmissiveColor, SpecularColor, SpecularPower). This way 
           it gives the colors of this material. The colors are given as 
           CColor, SpecularPower is a float."""
        self.MarkAbstract();
        
    def GetTextureCount(self):
        """This function returns the number of textures that are used by this 
           material to effect the color."""
        self.MarkAbstract();
        
    def LockTexture(self,iTexture):
        """This function prepares the texture data for being saved. Every call has 
           to be followed by precisely one UnlockTexture(...) call. You can not lock 
           two textures simultaneously. It returns True, if and only if locking 
           succeeded."""
        if(self.nLock>0 and iTexture!=self.iLockedTexture):
            return False;
        Result=True;
        if(self.nLock==0):
            Result=self.DoLockTexture(iTexture);
        if(Result):
            self.nLock+=1;
            self.iLockedTexture=iTexture;
        return Result;
    
    def DoLockTexture(self,iTexture):
        """This function implements LockTexture. It returns True, if and only 
           if locking succeeded."""
        self.MarkAbstract();
        
    def GetBlendMode(self):
        """This function returns the blend mode for the currently locked 
           texture as CBlendMode."""
        self.MarkAbstract();
    
    def GetTexCoordSet(self):
        """This function returns the index of the texture coordinate set that 
           is to be used for the currently locked texture."""
        self.MarkAbstract();
        
    def GetTextureFile(self):
        """This function returns the absolute path to the image file that is to 
           be used for the currently locked texture."""
        self.MarkAbstract();
    
    def UnlockTexture(self):
        """This function undoes the operations performed by LockTexture(...)."""
        if(self.nLock==1):
            self.DoUnlockTexture();
        self.nLock-=1;
        self.nLock=max(self.nLock,0);
        
    def DoUnlockTexture(self):
        """This function implements UnlockTexture()."""
        self.MarkAbstract();
    
    def MarkAbstract(self):
        """This function throws a not implemented exception. It is used to 
           indicate "abstract" functions."""
        raise NotImplemented("CMaterialDataBase is abstract, use derived classes to represent actual material data.");
    
    # A list of Blender objects, which function as data source
    SourceList=list();
    # The number of valid calls to LockTexture(...) which have not been 
    # followed by a call to UnlockTexture()
    nLock=0;
    # The index of the currently locked texture. If nLock is 0 this is 
    # irrelevant.
    iLockedTexture=-1;


class CMaterial:
    """This class is used to collect data about materials."""

    def __init__(self,iMaterial=None,MaterialData=None):
        """This constructor collects data from the given material data. If None is
           passed for MaterialData or iMaterial parameter a default material is 
           created."""
        if(MaterialData==None or iMaterial==None):
            self.MakeDefaultMaterial();
            return;
        self.CleanUp();
        self.SourceList.extend(MaterialData.SourceList);
        # Get the material index and the name
        self.iMaterial=iMaterial;
        self.Name=MaterialData.GetName();
        # Get the material colors
        (self.AmbientColor,self.DiffuseColor,self.EmissiveColor,self.SpecularColor,self.SpecularPower)=MaterialData.GetColorInfo();
        # Determine the number of used textures and get the list of texture
        # files
        self.nTexture=MaterialData.GetTextureCount();
        for iTextureData in range(self.nTexture):
            if(not MaterialData.LockTexture(iTextureData)):
                print("Locking a texture failed for the material \""+self.Name+"\".");
                continue;
            # Get the texture coordinate set and the blend mode
            self.piTexCoordSet.append(MaterialData.GetTexCoordSet());
            self.pBlendMode.append(MaterialData.GetBlendMode());
            # Remove slashes from the beginning of the file name with path
            TextureFile=MaterialData.GetTextureFile();
            TextureFileName=TextureFile[max(TextureFile.rfind("/"),TextureFile.rfind("\\"))+1:];
            self.pTextureFile.append(TextureFile);
            self.pTextureFileName.append(TextureFileName);
            # Unlock the texture
            MaterialData.UnlockTexture();
        self.nTexture=len(self.pTextureFile);


    def MakeDefaultMaterial(self):
        """This function makes this material become a white, untextured default
           material with index 0."""
        self.CleanUp();
        self.iMaterial=0;
        self.Name="Default";
        # The material colors
        self.AmbientColor=CColor(1.0,1.0,1.0,1.0);
        self.DiffuseColor=CColor(1.0,1.0,1.0,1.0);
        self.SpecularColor=CColor(0.0,0.0,0.0,1.0);
        self.EmissiveColor=CColor(0.0,0.0,0.0,1.0);
        self.SpecularPower=0.0;
        # Use no textures at all
        self.nTexture=0;
        self.pTextureFile=[];
        self.pTextureFileName=[];
        self.pBlendMode=[];
        self.piTexCoordSet=[];

    def CleanUp(self):
        """This function overwrites all data of this material with new objects holding
           default values. This way it makes the instance become independent from
           other instances of the same class."""
        self.iMaterial=0;
        self.Name="";
        self.pTextureFile=[];
        self.pTextureFileName=[];
        self.pBlendMode=[];
        self.piTexCoordSet=[];
        self.SourceList=[];

    # The index of this material
    iMaterial=0;
    # The name of this material
    Name="";
    # The material colors
    AmbientColor=CColor(1.0,1.0,1.0,1.0);
    DiffuseColor=CColor(1.0,1.0,1.0,1.0);
    SpecularColor=CColor(0.0,0.0,0.0,1.0);
    EmissiveColor=CColor(0.0,0.0,0.0,1.0);
    SpecularPower=0.0;
    # The number of used textures
    nTexture=0;
    # A list of strings containing texture file names with path
    pTextureFile=[];
    # A list of strings containing texture file names without a path
    pTextureFileName=[];
    # A list with nTexture entries giving the Blender texture operation for every
    # texture
    pBlendMode=[];
    # A list with nTexture entries giving the texture coordinate set for every
    # texture. The index is determined by iterating through UVLayerSet.
    piTexCoordSet=[];
    # This is a list holding any Blender Material object, which has been a source for
    # this object. Typically it contains either zero or one entries.
    SourceList=[];
    
    
class CArmatureDataBase:
    """This is an abstract base class that encapsulates Blender armature 
       objects and makes their data accessible through a version independent 
       interface."""
    
    def __init__(self,ArmatureName):
        """This function initalizes the base part of this class with the given 
           data."""
        # Save the passed data
        self.Name=ArmatureName;
        # Initialize all members
        self.SourceList=list();
        self.LockActionUsed=False;
        self.nLock=0;
        self.LockedActionName="";
        self.LockedBoneName="";
    
    def GetName(self):
        """This function returns the name of the associated armature."""
        return self.Name;
        
    def GetArmatureTransformation(self):
        """This function returns the armature to world space transformation of 
           the associated armature as CMatrix4x4."""
        self.MarkAbstract();

    def GetBoneNameList(self):
        """This function returns a list of strings giving the names of all bones in 
           this armature data."""
        self.MarkAbstract();
    
    def BoneExists(self,BoneName):
        """This function returns true, if and only if a bone with the given name is 
           part of this armature."""
        self.MarkAbstract();
    
    def GetBoneParent(self,BoneName):
        """This function returns the name of the parent of the bone with the 
           given name or None, if a bone with the given name does not exist, or if 
           the armature functions as parent."""
        self.MarkAbstract();
    
    def GetBoneTransformationInherit(self,BoneName):
        """This function returns a tuple of four booleans describing how the 
           transformation of the bone with the given name is effected by the 
           transformation of its parent. If the bone name is invalid it returns 
           (None,None,None,None) to indicate failure.
           The first boolean is true if and only if the rotation of the parent 
           bone effects the transformation.
           The second boolean is true if and only if the scaling of the parent 
           bone effects the transformation.
           The third boolean is true if and only if the translation of the bone is 
           disregarded completely, regarding only the parent translation.
           The fourth boolean is true, if and only if the translation of the bone 
           is given in armature space.
           In Blender 2.5 this corresponds to the properties use_inherit_rotation, 
           use_inherit_scale, use_connect and not use_local_location."""
        self.MarkAbstract();
    
    def GetBoneSource(self,BoneName):
        """This function returns the Blender object that belongs to the bone 
           with the given name or None, if there is no such bone. It is meant to be 
           written into SourceLists."""
        self.MarkAbstract();

    def GetBoneRestTransformation(self,BoneName):
        """This function returns the bone to armature space transformation of 
           the bone with the given index in its rest pose as CMatrix4x4 or None if 
           there is no such bone."""
        self.MarkAbstract();
    
    def ActionExists(self,ActionName):
        """This function returns true, if and only if an action with the given name 
           exists."""
        self.MarkAbstract();

    def LockBoneAction(self,ActionName,BoneName):
        """This function locks the action with the given index for the bone 
           with the given name, such that the animation of this bone can be read 
           from the curves using the GetBoneAction...(...) functions. Every call to 
           LockBoneAction(...) has to be followed by exactly one call to 
           UnlockBoneAction(...). You can not lock two different actions or 
           bones at once. It returns true, if and only if locking succeeded."""
        if(not self.BoneExists(BoneName)):
            return False;
        Result=not self.LockActionUsed and (self.nLock==0 or (self.LockedActionName==ActionName and self.LockedBoneName==BoneName));
        if(self.nLock==0):
            Result=self.DoLockBoneAction(ActionName,BoneName);
        if(Result):
            self.nLock+=1;
        return Result;
        
    def DoLockBoneAction(self,ActionName,BoneName):
        """This function implements LockBoneAction(...). It returns True if and 
           only if locking succeeded."""
        self.MarkAbstract();

    def UnlockBoneAction(self):
        """This function undoes the changes caused by the last successful call 
           to LockBoneAction(...)."""
        if(not self.LockActionUsed):
            self.nLock-=1;
            if(self.nLock==0):
                self.DoUnlockBoneAction();
                self.LockedActionName="";
                self.LockedBoneName="";
            self.nLock=max(0,self.nLock);

    def DoUnlockBoneAction(self):
        """This function implements UnlockBoneAction(...)."""
        self.MarkAbstract();
    
    def GetBoneActionTranslation(self,Frame):
        """This function returns the change of the translation of the currently 
           locked bone at the given frame of the locked action relative to the 
           rest pose in the coordinate system of the parent bone or the 
           armature as C3DVector. This is precisely the key data. If no 
           translation keys should be created it returns None."""
        self.MarkAbstract();
        
    def GetBoneActionRotation(self,Frame):
        """This function returns the change of the rotation of the currently 
           locked bone at the given frame of the locked action relative to the 
           rest pose in the coordinate system of the parent bone or the 
           armature as CQuaternion. This is precisely the key data. If no 
           rotaiton keys should be created it returns None."""
        self.MarkAbstract();
        
    def GetBoneActionScaling(self,Frame):
        """This function returns the change of the scaling of the currently 
           locked bone at the given frame of the locked action relative to the 
           rest pose in the coordinate system of the parent bone or the 
           armature as C3DVector. This is precisely the key data. If no 
           scaling keys should be created it returns None."""
        self.MarkAbstract();
    
    def GetBonePoseTransformation(self,BoneName):
        """This function returns the bone to armature space transformation of 
           the bone with the given name in the current pose as CMatrix4x4 or 
           as Blender matrix. On failure it returns None and prints a message."""
        self.MarkAbstract();
    
    def IsValid(self):
        """This function returns true, if and only if this object has access to 
           valid data."""
        self.MarkAbstract();
        
    def MarkAbstract(self):
        """This function throws a not implemented exception. It is used to 
           indicate "abstract" functions."""
        raise NotImplemented("CArmatureDataBase is abstract. Use derived classes to get access to model data in a particular Blender version.");

    # A list of Blender objects, which function as source for this armature 
    # data object
    SourceList=list(); 
    # The name of the associated armature
    Name="";
    
    # This boolean is True, if and only if the armature data is currently 
    # locked through LockBoneAction(...)
    LockActionUsed=False;
    # The lock count for both LockBoneAction(...)
    nLock=0;
    # The name of the currently locked action
    LockedActionName=-1;
    # The name of the bone that is currently locked through LockBoneAction(...)
    LockedBoneName="";


class CBone:
    """This class is used to collect data about a single bone."""

    def __init__(self,ParentBone,iBone,BoneName,ModelData,ArmatureData,ExportThroughCurves,ExportActions,MaxKeyDistance):
        """This constructor collects data from the bone with the given data name 
           in the given armature data. The index of the bone gets set to iBone and 
           the CBone ParentBone is set as parent (without further verification). The 
           given skeleton is stored as skeleton, which uses this bone. If 
           ExportThroughCurves is True keys are read. Otherwise keys may be added 
           later. In every case the keyed transformations are created."""
        self.CleanUp();
        if(ArmatureData==None or ModelData==None):
            return;
        self.SourceList.append(ArmatureData.GetBoneSource(BoneName));
        # Save the name, the index and the parent
        self.Name=BoneName;
        self.iBone=iBone;
        self.Parent=ParentBone;
        # Determine how transformations of the parent need to be used
        self.InheritRotation,self.InheritScaling,self.IgnoreOwnTranslation,self.ArmatureSpaceTranslation=ArmatureData.GetBoneTransformationInherit(BoneName);
        # Get the current transformation
        ArmatureToWorldSpace=MakeCMatrix4x4(ArmatureData.GetArmatureTransformation());
        self.CurrentTransformation=MakeCMatrix4x4(ArmatureData.GetBonePoseTransformation(BoneName))*ArmatureToWorldSpace;
        # Get the mesh transformation
        BoneToArmatureSpace=MakeCMatrix4x4(ArmatureData.GetBoneRestTransformation(BoneName));
        self.MeshTransformation=(BoneToArmatureSpace*ArmatureToWorldSpace).ComputeInverse();
        # Determine the default bone to parent bone space matrix
        ParentName=ArmatureData.GetBoneParent(BoneName);
        if(ParentName!=None):
            ParentBoneRest=MakeCMatrix4x4(ArmatureData.GetBoneRestTransformation(ParentName));
            self.DefaultBoneToParentSpace=BoneToArmatureSpace*ParentBoneRest.ComputeInverse();
        else:
            self.DefaultBoneToParentSpace=BoneToArmatureSpace;
        # Create the keys
        if(ExportActions):
            for ActionName in ModelData.GetActionNameList():
                # Create a keyed transformation for the action
                ActionTransformation=self.GetKeyedTransformation(ActionName);
                ActionTransformation.SourceList.append(ModelData.GetActionSource(ActionName));
                MinFrame,MaxFrame=ModelData.GetActionRange(ActionName);
                ActionTransformation.MinFrame=float(MinFrame);
                ActionTransformation.MaxFrame=float(MaxFrame);
                if(ExportThroughCurves):
                    ActionTransformation.AddKeysFromActionCurves(ModelData,ArmatureData,ActionName,BoneName,self.DefaultBoneToParentSpace,MaxKeyDistance,self.InheritRotation,self.InheritScaling,self.IgnoreOwnTranslation,self.ArmatureSpaceTranslation);
        # Ensure that there is at least one key of every type to have the bone to
        # parent transformation included
        if(ExportActions and ExportThroughCurves):
            BoneToParentTranslation=self.DefaultBoneToParentSpace.TranslationToVector();
            BoneToParentRotation=self.DefaultBoneToParentSpace.RotationToQuaternion();
            BoneToParentScaling=self.DefaultBoneToParentSpace.ScalingToVector();
            for Action in self.KeyDict.keys():
                self.KeyDict[Action].MakeDefaultKeys(
                    BoneToParentTranslation,BoneToParentRotation,BoneToParentScaling);

                    
    def MakeDummyBone(self,BoneTransformationDict,BoneCurrentTransformation=CMatrix4x4(),BoneName="DummyBone"):
        """This function creates a dummy bone. See CSkeleton.MakeDummySkeleton(...) 
           for details. All current data of this bone is overwritten."""
        self.CleanUp();
        self.Name=BoneName;
        self.MeshTransformation=CMatrix4x4();
        self.DefaultBoneToParentSpace=BoneCurrentTransformation.Clone();
        self.CurrentTransformation=BoneCurrentTransformation.Clone();
        self.KeyDict=BoneTransformationDict.Clone();


    def GetKeyedTransformation(self,TransformationKey):
        """This function returns the keyed transformation for the bone to parent 
           space transformation of this bone, which is stored under the given key. 
           If it does not exist, it is created (including ancestor 
           transformations)."""
        # Get the corresponding keyed transformation of the parent (if any)
        ParentTransformation=None;
        if(self.Parent!=None):
            ParentTransformation=self.Parent.GetKeyedTransformation(TransformationKey);
        # Get the keyed transformation of this object and create the keys
        return self.KeyDict.GetKeyedTransformation(TransformationKey,ParentTransformation);


    def Visit(self,piDFS):
        """This function visits this bone and all of its child bones (unless this has
           already happened). It uses the only element of the given list to count up
           the DFS number."""
        # Check whether this bone is unvisited
        if(self.iDFS==-1):
            # Mark it as visited and assign a DFS number
            self.iDFS=piDFS[0];
            piDFS[0]+=1;
            # Visit its children
            for Child in self.pChild:
                Child.Visit(piDFS);


    def GetTransformation(self,Skeleton=None,ActionName=None,Frame=0.0):
        """This function returns a bone to model space transformation for this
           bone. If you pass a valid action name and the CSkeleton, which makes use 
           of this bone, the transformation at the given frame in the action with 
           the given name is returned. Otherwise the transformation of the bone in  
           Blender at the creation of this object is returned."""
        if(ActionName==None or Skeleton==None or not ActionName in self.KeyDict):
            return self.CurrentTransformation.Clone();
        else:
            BoneToSkeletonSpace=self.KeyDict[ActionName].GetAbsoluteFrameTransformation(Frame);
            SkeletonToWorldSpace=Skeleton.GetTransformation(ActionName,Frame);
            return BoneToSkeletonSpace*SkeletonToWorldSpace;


    def ApplyTransformation(self,Transformation):
        """This function applies the given transformation to all keyed
           transformations of this bone."""
        ParentDefaultTransformation=None;
        if(self.Parent!=None):
            if("" in self.Parent.KeyDict):
                ParentDefaultTransformation=self.Parent.KeyDict[""];
        self.KeyDict.MakeDefaultKeys(self.DefaultBoneToParentSpace,"",ParentDefaultTransformation);
        self.DefaultBoneToParentSpace*=Transformation;
        for KeyedTransformation in self.KeyDict.values():
            KeyedTransformation*=Transformation;


    def Clone(self,BoneList):
        """This function returns a new CBone object, which holds exactly the same
           data as this bone with the only exception that the parent will be an
           object from the corresponding index in the given list."""
        Result=CBone(None,None,None,None,None,None,None,None);
        Result.iBone=self.iBone;
        Result.Name=self.Name;
        Result.Parent=None;
        if(self.Parent!=None):
            Result.Parent=BoneList[self.Parent.iBone];
        Result.MeshTransformation.Copy(self.MeshTransformation);
        Result.DefaultBoneToParentSpace.Copy(self.DefaultBoneToParentSpace);
        Result.CurrentTransformation.Copy(self.CurrentTransformation);
        Result.InheritRotation=self.InheritRotation;
        Result.InheritScaling=self.InheritScaling;
        Result.IgnoreOwnTranslation=self.IgnoreOwnTranslation;
        Result.ArmatureSpaceTranslation=self.ArmatureSpaceTranslation;
        for Key in self.KeyDict:
            TransformationClone=self.KeyDict[Key].Clone();
            TransformationClone.Parent=None;
            if(Result.Parent!=None):
                TransformationClone.Parent=Result.Parent.GetKeyedTransformation(Key);
            Result.KeyDict[Key]=TransformationClone;
        Result.SourceList=list(self.SourceList);
        return Result;


    def CleanUp(self):
        """This function overwrites all data of this bone with new objects holding
           default values. This way it makes the instance become independent from
           other instances of the same class."""
        self.iBone=0;
        self.Name="";
        self.Parent=None;
        self.MeshTransformation=CMatrix4x4();
        self.DefaultBoneToParentSpace=CMatrix4x4();
        self.CurrentTransformation=CMatrix4x4();
        self.InheritRotation=True;
        self.InheritScaling=True;
        self.IgnoreOwnTranslation=False;
        self.ArmatureSpaceTranslation=False;
        self.KeyDict=CKeyedTransformationDict();
        self.SourceList=[];

    # The index of this bone in the bone list of the corresponding CSkeleton
    iBone=0;
    # The name of this bone
    Name="";
    # The CBone that functions as parent or None for no parent
    Parent=None;
    # The transformation that is to be applied to the mesh space geometry before the 
    # final bone transformation (=mesh to bone space)
    MeshTransformation=CMatrix4x4();
    # The default transformation of this bone in relation to its parent. The default
    # transformation is the transformation of the bone, when the skeleton is in edit
    # mode. So this is a bone to parent bone space for this default pose. For roots
    # it is a bone to skeleton space transformation.
    DefaultBoneToParentSpace=CMatrix4x4();
    # This is the bone to model space transformation the bone had in the scene when
    # this object was created. It is used internally for making the transformations
    # of child objects relative.
    CurrentTransformation=CMatrix4x4();
    # This boolean is true if and only if the rotation of the parent bone effects 
    # the rotation of this bone
    InheritRotation=True;
    # This boolean is true if and only if the scaling of the parent bone effects 
    # the scaling of this bone
    InheritScaling=True;
    # This boolean is true if and only if the translation keys of this bone are 
    # ignored
    IgnoreOwnTranslation=False;
    # This boolean is true if and only if the translation keys of this bone are 
    # interpreted as armature space coordinates (irrelevant if IgnoreOwnTranslation 
    # is true)
    ArmatureSpaceTranslation=False;
    # This is an object, which maps Blender action names to the corresponding 
    # CKeyedTransformation giving the time-dependent bone to parent bone space. 
    # The transformations have proper parenting relations up to the root bones. 
    KeyDict=None;

    # This is a list holding all Blender Bone objects, which have been a source for
    # this object. Typically it consists of either zero or one entries.
    SourceList=[];


class CSkeleton(CObject):
    """Instances of this class correspond to one or more armature objects in Blender."""

    def __init__(self,ModelData,ArmatureData,ParentObject,ReadActionCurves,ExportActions,MaxKeyDistance,DuplicationTransformation=CMatrix4x4()):
        """This constructor initializes this skeleton from the given armature data 
           as part of the model described by the given model data. If 
           ReadActionCurves and ExportActions are True, it reads from curves to  
           create key frames using the given maximal key distance. Otherwise it does 
           not read keys at all, but if ExportActions is true it creates one keyed 
           transformation per action. If keys are not read it is recommended to use 
           AddKeysFromPose(...) after reading all other objects. See CObject for a 
           description of DuplicationTransformation."""
        if(ArmatureData!=None):
            CObject.__init__(self,ArmatureData.SourceList[0],ParentObject,False,DuplicationTransformation);
        else:
            CObject.__init__(self,None,ParentObject,False,DuplicationTransformation);
        self.BoneList=[];
        self.ActionList=[];
        if(ArmatureData==None or not ArmatureData.IsValid()):
            return;
        # Here it gets a little complicated. It is necessary to create parent bones
        # before their children, because of options like hinge. Performing a DFS 
        # through the bone list gives such an order, so this is done here.
        BoneNameList=ArmatureData.GetBoneNameList();
        # Create a list of child bone names for every bone and save it in a 
        # dictionary
        pBoneChildDict=dict();
        for BoneName in BoneNameList:
            pBoneChildDict[BoneName]=list();
        for BoneName in BoneNameList:
            ParentName=ArmatureData.GetBoneParent(BoneName);
            if(ParentName!=None):
                pBoneChildDict[ParentName].append(BoneName);
        # Start the DFS using the child lists. Creating the bones is done during the
        # DFS and they get added to self.BoneList.
        self.BoneList=[];
        for RootBoneName in BoneNameList:
            if(ArmatureData.GetBoneParent(RootBoneName)==None):
                self.CreateBone(RootBoneName,ModelData,ArmatureData,None,pBoneChildDict,ReadActionCurves,ExportActions,MaxKeyDistance);
        # Create a list of all used actions
        self.ActionList=[];
        for ActionName in ModelData.GetActionNameList():
            self.ActionList.append(ActionName);

    
    def MakeDummySkeleton(self,BoneTransformationDict,BoneCurrentTransformation=CMatrix4x4(),SkeletonName="DummySkeleton",BoneName="DummyBone"):
        """This function overwrites the current data of this skeleton completely. It 
           fills it with new data, which describes a skeleton named SkeletonName 
           with a single root bone named BoneName. The skeleton does not have a 
           parent. Its only bone uses a clone of the given transformation dictionary 
           as animation and the given matrix as current transformation and default 
           bone to parent space transformation. All other transformations are the 
           identity. Thus the mesh to world space transformation of a mesh parented 
           to the given bone is the transformation in the given dictionary."""
        # Set the name
        self.Name=SkeletonName;
        # Set the transformation to identity
        self.Transformation=CMatrix4x4();
        self.TransformationDict=CKeyedTransformationDict();
        # Remove the parent
        self.Parent=None;
        self.ParentBone=None;
        self.ParentForSkinning=False;
        # A dummy has no sources
        self.SourceList=list();
        # The actions are the actions found in the transformation dictionary
        self.ActionList=list(BoneTransformationDict.keys());
        # Construct the dummy bone and set it up as only bone
        DummyBone=CBone(None,None,None,None,None,None,None,None);
        DummyBone.MakeDummyBone(BoneTransformationDict,BoneCurrentTransformation,BoneName);
        self.BoneList=[DummyBone];


    def CreateBone(self,BoneName,ModelData,ArmatureData,ParentBone,pBoneChildDict,ReadActionCurves,ExportActions,MaxKeyDistance):
        """This function creates a CBone for the bone with name BoneName in 
           ArmatureData and adds it to self.BoneList. Then it does the same for
           all children using pBoneChildDict. pBoneChildDict has to be a dictionary 
           mapping bone names to lists of child bone names. The given options are 
           used for creating the bone."""
        NewBone=CBone(ParentBone,len(self.BoneList),BoneName,ModelData,ArmatureData,ReadActionCurves,ExportActions,MaxKeyDistance);
        self.BoneList.append(NewBone);
        for ChildName in pBoneChildDict[BoneName]:
            self.CreateBone(ChildName,ModelData,ArmatureData,NewBone,pBoneChildDict,ReadActionCurves,ExportActions,MaxKeyDistance);


    def AddKeysFromPose(self,ActionName,Frame,ArmatureData):
        """This function generates keys for all bones, which resemble the current 
           pose of the armature object accessible through the given 
           CArmatureDataBase. The keys are added to the keyed transformations for 
           the action with the given name at the given frame. It returns true if and 
           only if all keys were generated successfully. On failure it might still 
           have created some keys."""
        # Create keys for every bone
        Result=True;
        for Bone in self.BoneList:
            # Get or create the keyed transformation
            KeyedTransformation=Bone.GetKeyedTransformation(ActionName);
            # Locate the bone data in the armature data
            if(not ArmatureData.BoneExists(Bone.Name)):
                print("Failed to find a bone called \""+Bone.Name+"\". Keys for the action \""+ActionName+"\" will be missing.");
            Frame=float(Frame);
            Result=Result and KeyedTransformation.KeyCurrentPose(Frame,ArmatureData,Bone.Name);
            KeyedTransformation.MinFrame=min(KeyedTransformation.MinFrame,Frame);
            KeyedTransformation.MaxFrame=max(KeyedTransformation.MaxFrame,Frame);
        return Result;
    

    def GetAnimationRange(self,Action=None):
        """This function returns the frame range of the animation of this 
           skeleton in the given action. More precisely it returns a tuple of 
           floats (MinFrame,MaxFrame), where MinFrame is the smallest frame 
           value that occurs in the action with the given name and MaxFrame is 
           the largest frame value. Passing None for Action means that all 
           actions will be regarded. If there is no key at all (None,None) is 
           returned."""
        MinFrame=1.0e20;
        MaxFrame=-1.0e20;
        if(Action!=None):
            for Bone in self.BoneList:
                if(Action in Bone.KeyDict):
                    MinFrame=min(MinFrame,Bone.KeyDict[Action].MinFrame);
                    MaxFrame=max(MaxFrame,Bone.KeyDict[Action].MaxFrame);
        else:
            for Bone in self.BoneList:
                for Action in Bone.KeyDict.keys():
                    MinFrame=min(MinFrame,Bone.KeyDict[Action].MinFrame);
                    MaxFrame=max(MaxFrame,Bone.KeyDict[Action].MaxFrame);
        if(MinFrame==1.0e20):
            return (None,None);
        else:
            return (MinFrame,MaxFrame);


    def EaseBoneDependencies(self,Model):
        """This function ensures that the bone list (BoneList) is sorted in a way
           where parents always have smaller indices than their children. It updates
           iBone indices accordingly. It needs the model to be able to update
           skinning indices. Pass None if you do not want this. The permutation of
           the bones is returned as list of new indices. This function will damage
           mesh skinning indices, if one mesh is used with multiple skeletons! Note
           that an unmodified bone list coming from __init__(...) already fulfills
           this criterium, so if you do not manipulate the order of the bone list,
           or add bones you do not need to call this function. If you do manipulate
           the order, you should call it, because some code relies on this order."""
        # Do a DFS on the bones. It writes DFS numbers to the iDFS member of the
        # bones in self.BoneList.
        self.DoBoneDFS();
        # Sorting with ascending DFS numbers in a tree gives a topological sorting.
        # In a topological sorting parents preceed children so this is what is needed.
        # Compute the new index for every bone.
        piBone=range(len(self.BoneList));
        for Bone in self.BoneList:
            piBone[Bone.iBone]=Bone.iDFS;
        # Move the bones to their new indices and update the indices they hold
        NewBoneList=[None]*len(self.BoneList);
        for Bone in self.BoneList:
            NewBoneList[piBone[Bone.iBone]]=Bone;
            Bone.iBone=piBone[Bone.iBone];
        # Check for a deadlock
        for Bone in self.BoneList:
            if(Bone.Parent!=None and Bone.Parent.iBone>Bone.iBone):
                print("WARNING: There is a deadlock in your bone dependencies. The bone \""+Bone.Parent.Name+"\" has the bone \""+Bone.Name+"\" as ancestor and as child. Here is the list of parent relations:");
                Relations=Bone.Name;
                Parent=Bone.Parent;
                while(Parent!=Bone):
                    Relations+=" -> "+Parent.Name;
                    if(Parent.Parent!=None):
                        Parent=Parent.Parent;
                    else:
                        break;
                Relations+=" -> "+Parent.Name;
                print(Relations);
        self.BoneList=NewBoneList;
        # Update skinning indices on all relevant meshes
        if(Model!=None):
            self.UpdateSkinningIndices(piBone,Model);
        return piBone;


    def UpdateSkinningIndices(self,piBone,Model):
        """Given an array that maps old bone indices to new bone indices, this
           function iterates through meshes in the given Model, which are animated
           through this skeleton and updates its skinning indices accordingly. If
           the new index of a bone is negative all weights referring to this bone get
           removed."""
        for Mesh in Model.pMesh:
            # Find out whether, this skeleton effects the mesh
            UsesThisSkeleton=False;
            for MeshObject in Model.pMeshObject:
                if(MeshObject.Mesh==Mesh):
                    UsesThisSkeleton=(MeshObject.Parent==self);
                    break;
            if(not UsesThisSkeleton):
                continue;
            # If so update the weights
            Mesh.UpdateSkinningIndices(piBone);


    def DoBoneDFS(self):
        """This function does a DFS on the bones of this skeleton starting only at
           root bones. It writes DFS numbers to the iDFS member of the bones in
           BoneList. It also adds pChild arrays."""
        # Add lists of children to all bones
        for Bone in self.BoneList:
            Bone.pChild=[];
        for Bone in self.BoneList:
            if(Bone.Parent!=None):
                Bone.Parent.pChild.append(Bone);
        # Initialize the DFS numbers with -1 to mark the bones as unvisited
        for Bone in self.BoneList:
            Bone.iDFS=-1;
        # Visit all bones. Keep the current DFS number in a list, to be able to 
        # change it from within the bones.
        piDFS=[0];
        for Bone in self.BoneList:
            if(Bone.Parent==None):
                Bone.Visit(piDFS);


    def RemoveUnusedBones(self,Model,MinWeight=0.05):
        """This function searches for bones, which do not effect any vertex with more
           than the given weight. Then it deletes all such bones. Animations get
           updated, but they can use a bit precision anyway. It can damage skinning
           indices, if this skeleton effects a mesh that is being used with multiple 
           skeletons! It only updates skinning indices, if Model is not None."""
        # Determine the maximal weight for every bone
        pBoneMaxWeight=[0.0]*len(self.BoneList);
        for Mesh in Model.pMesh:
            # Check whether this mesh is effected by this skeleton
            UsesThisSkeleton=False;
            for MeshObject in Model.pMeshObject:
                if(MeshObject.Mesh==Mesh):
                    UsesThisSkeleton=(MeshObject.Parent==self and MeshObject.ParentBone==None);
                    if(UsesThisSkeleton):
                        break;
            # If so regard its maximal weights
            if(UsesThisSkeleton):
                pBoneMeshMaxWeight=Mesh.GetMaxBoneWeightList();
                for i in range(min(len(self.BoneList),len(pBoneMeshMaxWeight))):
                    pBoneMaxWeight[i]=max(pBoneMaxWeight[i],pBoneMeshMaxWeight[i]);
        # Also check whether the bones are parents for objects
        for Object in Model.pObject:
            if(Object.ParentBone!=None):
                if(Object.Parent==self):
                    pBoneMaxWeight[Object.ParentBone.iBone]=1.0;
        # Create a new list of bones, which holds only used bones. Create another
        # list of deleted bones. Save the mapping of indices.
        NewBoneList=[];
        DeletedBoneList=[];
        piBone=[-1]*len(self.BoneList);
        print("The following bones can be eliminated in \""+self.Name+"\", because they do not effect geometry:");
        BoneNameList="";
        for i in range(len(self.BoneList)):
            if(pBoneMaxWeight[i]>MinWeight):
                piBone[i]=len(NewBoneList);
                NewBoneList.append(self.BoneList[i]);
            else:
                piBone[i]=-1;
                DeletedBoneList.append(self.BoneList[i]);
                BoneNameList+=self.BoneList[i].Name+"; ";
        print(BoneNameList[0:-2]);
        print("In total "+str(len(DeletedBoneList))+" bones get eliminated...");
        # For every remaining bone that has a parent ensure that this parent is 
        # a remaining bone as well. If its not find the next parent that does 
        # remain and apply the relative transformation.
        for Bone in NewBoneList:
            # Find the closest parent that is a remaining bone and find all 
            # parents on this way up, which do not remain
            RemainingParent=Bone.Parent;
            EliminatedParentList=list();
            while(RemainingParent!=None and pBoneMaxWeight[RemainingParent.iBone]<=MinWeight):
                EliminatedParentList.append(RemainingParent);
                RemainingParent=RemainingParent.Parent;
            # There may be nothing to do
            if(RemainingParent==Bone.Parent):
                continue;
            # Otherwise update the parent
            Bone.Parent=RemainingParent;
            # And apply the relative transformation to the transformation of 
            # the child
            for EliminatedParent in EliminatedParentList:
                # Update the default bone to parent space transformation
                Bone.DefaultBoneToParentSpace*=EliminatedParent.DefaultBoneToParentSpace;
                # Combine the animations
                for Action in Bone.KeyDict:
                    Bone.KeyDict[Action]*=EliminatedParent.KeyDict[Action];
        # Update the indices in the bones
        for i in range(len(self.BoneList)):
            self.BoneList[i].iBone=piBone[i];
        # Save the new bone list
        self.BoneList=NewBoneList;
        # Update skinning indices on all relevant meshes
        if(Model!=None):
            self.UpdateSkinningIndices(piBone,Model);


    def ApplyDataTransformation(self,Transformation):
        """This function applies the given transformation to the root bones of this
           object."""
        for Bone in self.BoneList:
            if(Bone.Parent==None):
                Bone.ApplyTransformation(Transformation);
                
                
    def TransformTime(self,FrameOffset,FrameFactor=1.0):
        """This function invokes TransformTime(...) on all keyed 
           transformations of all bones in this skeleton."""
        for Bone in self.BoneList:
            Bone.KeyDict.TransformTime(FrameOffset,FrameFactor);
            
            
    def FixRotationKeys(self):
        """This function invokes FixRotationKeys(...) on all keyed transformations, 
           which belong to this skeleton (including those of all bones and of the 
           object transformation)."""
        for Bone in self.BoneList:
            Bone.KeyDict.FixRotationKeys();

    def CompareType(self,Object):
        """This function returns true if and only if the given object has the same
           type as this object."""
        return isinstance(Object,CSkeleton);

    def IsPureObject(self):
        """This function returns True if and only if this object is only a CObject
           not an instance of some derived type."""
        return False;

    def DoAppendObjectData(self,Object,ObjectToSelfSpace):
        """This function makes this object become an object that also contains the
           data of the given other skeleton (needed as CSkeleton). For this to work
           the relative transformation between self and Object has to be constant and
           it has to be given as CMatrix4x4. If the process fails the return value is
           False, otherwise it is True."""
        # Create clones of all bones in the other object
        CloneList=[];
        for Bone in Object.BoneList:
            CloneList.append(Bone.Clone(CloneList));
        # Give them indices for the new list and transform them
        for Clone in CloneList:
            Clone.iBone+=len(self.BoneList);
            Clone.ApplyTransformation(ObjectToSelfSpace);
        # Append the clone list to the list of this object
        self.BoneList.append(CloneList);
        # Add actions which are not part of the list in this object
        for Action in Object.ActionList:
            if(not Action in self.ActionList):
                self.ActionList.append(Action);
        self.SourceList.append(Object.SourceList);
        return True;

    def Clone(self):
        """This function creates a deep copy of this object and returns it."""
        Result=CSkeleton(None,None,None,None,None,None);
        Result.CopyObject(self);
        # Copy the bone list using deep bone copies
        Result.BoneList=list();
        for Bone in self.BoneList:
            Result.BoneList.append(Bone.Clone(Result.BoneList));
        # Copy the action list
        Result.ActionList=list(self.ActionList);
        return Result;

    # A list of CBone objects for all bones in this skeleton. It is guaranteed that
    # the index of a parent bone is always smaller than the index of a child. After a
    # change this state can be restored using EaseBoneDependencies(...).
    BoneList=[];
    # This is a list holding the names of all actions that are used by this skeleton
    ActionList=[];


class CExportObjectBase:
    """This is a little abstract helper class, which is used to resolve groups 
       and find parents. Temporary instances of it get created in 
       CModel.__init__(...). Derived classes are specific to particular Blender 
       versions."""

    def __init__(self):
        """This function initializes the base part of this class with "empty" 
           default data."""
        self.BlenderObject=None;
        self.Object=None;
        self.BlenderParent=None;
        self.BlenderParentBone=None;
        self.DuplicationID="";
        self.DuplicationTransformation=CMatrix4x4();
        self.Parent=None;
        self.ParentBone=None;
        self.ObjectOnly=False;
        self.ChildList=[];
        self.GroupResolved=False;
        self.iDFS=-1;
        
    def IsChildOf(self,RHS):
        """This function returns true, if and only if this object has a parent set 
           in Blender and the right duplication of this parent is represented by the 
           given CExportObject."""
        if(self.BlenderParent!=None):
            return self.BlenderParent==RHS.BlenderObject and self.DuplicationID==RHS.DuplicationID;

    def DoDFS(self,piDFS=[0]):
        """This function starts a DFS from this export object and assigns DFS numbers
           to this bone and all of its children."""
        # Do not perform a DFS, if the DFS number is already known
        if(self.iDFS<0):
            # Assign the next DFS number
            self.iDFS=piDFS[0];
            piDFS[0]+=1;
            # Continue the DFS for the children
            for Child in self.ChildList:
                Child.DoDFS(piDFS);
                
    def GetObjectName(self):
        """This function returns the name of the associated Blender object."""
        self.MarkAbstract();

    def GetParentBoneName(self):
        """This function returns the name of the parent bone of this export 
           object, or None, if it does not have a parent bone."""
        self.MarkAbstract();
        
    def UsesSkinning(self):
        """This function returns true, iff this export object uses skinning."""
        self.MarkAbstract();
        
    def MarkAbstract(self):
        """This function throws a not implemented exception. It is used to 
           indicate "abstract" functions."""
        raise NotImplemented("CExportObjectBase is abstract. Use derived classes to get access to model data in a particular Blender version.");

    # This is a reference to the Blender object that is to be exported
    BlenderObject=None;
    # The CObject that corresponds to this object or None, if it is not yet created
    Object=None;
    # The Blender parent object or None, if there is no parent. If the parent is a
    # bone, this will be the corresponding armature.
    BlenderParent=None;
    # The Blender parent bone or None, if the parent of this object is not a bone.
    BlenderParentBone=None;
    # If this export object has been created by duplicating a group, this string 
    # gives an identifier for this duplication of the group. Otherwise it is empty.
    DuplicationID="";
    # This is the corresponding transformation, which needs to be applied to get 
    # from the original object to this duplication
    DuplicationTransformation=CMatrix4x4();
    # The parent of this bone as CExportObject or None, if there is no such object
    # yet
    Parent=None;
    # The CBone that represents the parent bone of this object or None, if this bone
    # has not been created yet
    ParentBone=None;
    # If this boolean is true only a CObject gets exported for BlenderObject.
    # Otherwise a CMeshObject or a CSkeleton may be created.
    ObjectOnly=False;
    # This is a list of all CExportObject instances, which have this object as parent
    ChildList=[];
    # If BlenderObject is an empty that belongs to a group this boolean saves whether
    # the group has already been resolved
    GroupResolved=False;
    # This is the DFS number of this export object. It is used to bring the objects
    # into an order, where parents have lower indices than their children
    iDFS=-1;


class CModelOptions:
    """This class encapsulates the options that are relevant for reading model 
       data."""

    def __init__(self,ExportSelectionOnly,ExportAnimations,ExportConstraints,ExportCurrentNLA,MaxKeyDistance,ExportAllUVLayers,ReadModifiers,RenderModeModifiers=False,CreateDefaultMaterial=True,ExportObjectCallback=None):
        """This constructor saves the given option values.
        @param ExportSelectionOnly: If this is true only selected objects in 
               the current scene will be read. Otherwise all objects on all 
               layers in the current scene will be read.
        @param ExportAnimations: If this is true animations will be read. If 
               it is false armatures will be regarded as if they were unselected  
               empties and object animations are not read either.
        @param ExportConstraints: If ExportConstraints is True animations are 
               not read from keys directly, but are read from actual bone 
               transformations in the different poses at different frames. This 
               way constraints are supported. Otherwise only explicit key 
               values are regarded.
        @param ExportCurrentNLA: If this option is true, the current NLA of the 
               objects will be read. If it is false all actions will be read. If 
               it is true, exporting constraints gets enabled automatically, 
               because it is a prerequisite.
        @param MaxKeyDistance: If this integer is 0 or smaller the plugin will 
               create keys at frames at which the actions have keys. If it has 
               a positive value this value gives the maximal number of frames 
               between keys and more keys get created accordingly.
        @param ExportAllUVLayers: If this is True all UVLayers will be regarded 
               during reading and thus every vertex may have multiple texture 
               coordinates. This can only work, if all meshes have the same set 
               of UV layer names. The mapping from UV layers to texture 
               coordinate set indices gets written to the console window. If it 
               is False only the active UV layer gets regarded and the vertices 
               will have at most one texture coordinate. If a mesh has no UV 
               layer the vertices for this mesh will have no texture 
               coordinates. See also:
               CModel.UnifyTexCoordSetCount(...)
        @param ReadModifiers: If this is True all meshes are read with all 
               modifiers applied. More generally spoken it reads all meshes  
               just like they are displayed in Blender. Armature modifiers are 
               the only exception. As a consequence of this all object types, 
               that are visualized through meshes can be exported. Those are 
               mesh, surface, curve, meta and text. If it is False all 
               modifiers except for the armature modifier get ignored and only 
               objects of type mesh can be exported to geometry.
        @param RenderModeModifiers: This option is only relevant, if 
               ReadModifiers is True. If this option is True the modifiers will 
               be applied using the settings that are used for rendering in 
               Blender. If it is False the modifiers will be applied using the  
               settings that are used for interactive display in object mode. 
        @param CreateDefaultMaterial: If this is True a special, white, 
               untextured default material gets created and used, if there is 
               no other material. If it is False you might end up having a 
               model with no material at all.
        @param ExportObjectCallback: If this is not None it has to be a callback, 
               which takes a Blender Object as parameter. The callback has to return 
               True, if the Object should be exported in all cases, False, if it 
               must not be exported or None, if the default method is to be used to 
               determine whether the object should be exported."""
        self.ExportSelectionOnly=ExportSelectionOnly;
        self.ExportAnimations=ExportAnimations;
        self.ExportConstraints=ExportConstraints;
        self.ExportCurrentNLA=ExportCurrentNLA;
        if(self.ExportCurrentNLA and not self.ExportConstraints):
            print("Exporting constraints has been automatically enabled to support exporting NLA.");
            self.ExportConstraints=True;
        self.MaxKeyDistance=MaxKeyDistance;
        self.ExportAllUVLayers=ExportAllUVLayers;
        self.ReadModifiers=ReadModifiers;
        self.RenderModeModifiers=RenderModeModifiers;
        self.CreateDefaultMaterial=CreateDefaultMaterial;
        self.ExportObjectCallback=ExportObjectCallback;
    
    # See the comments in __init__ for descriptions of these attributes
    ExportSelectionOnly=False;
    ExportAnimations=True;
    ExportConstraints=True;
    ExportCurrentNLA=False;
    MaxKeyDistance=0;
    ExportAllUVLayers=False;
    ReadModifiers=True;
    RenderModeModifiers=False;
    CreateDefaultMaterial=True;
    ExportObjectCallback=None;


class CModelDataBase:
    """This function provides access to the data of a model in a way that is 
       not specific to a particular Blender version. It is an abstract base 
       class for version specific classes."""
    
    def __init__(self,ModelOptions,BlenderUtility):
        """This constructor initializes this object with default data and with 
           the given options."""
        self.Options=ModelOptions;
        self.BlenderUtility=BlenderUtility;
        
    def GetExportObjectList(self):
        """This function returns a list of export objects (CExportObject), 
           which are to be read for the associated model."""
        self.MarkAbstract();

    def ExportObjectToObject(self,ExportObject,ObjectParent,BlenderMaterialList,MeshList,UVLayerSet,DuplicationRootDict):
        """This function creates a matching CObject for a given CExportObject. 
           The resulting CObject is returned. For this to work the CObject of 
           the parent object has to be given (if there is no parent, pass 
           None). Besides you need to pass the mapping from duplication IDs 
           to duplication root export objects. For creating mesh objects some 
           additional parameters are needed. 
        @param BlenderMaterialList: Has to be a list of all materials Blender 
               materials that are already in use. The function may add to it.
        @param MeshList: Has to be a list of all read CMesh instances. The 
               function may add to it.
        @param UVLayerSet: Has to be a set of UV layer names, which are to be 
               read or None."""
        self.MarkAbstract();
        
    def GetUVLayerSet(self,RelevantObjects):
        """If the set of UV texture names is not the same for all meshes that 
           are associated with the export objects in the given list, this 
           function returns None. Otherwise it returns the unique set of UV 
           layer names."""
        self.MarkAbstract();
    
    def GetMaterialData(self,BlenderMaterial,UVLayerSet):
        """Given a Blender material object and a set of UV layer names (or 
           None) this function returns a corresponding CMaterialData 
           instance. Note that this is quite expensive."""
        self.MarkAbstract();
        
    def CreateExportObject(self,BlenderObject,PureObject,DuplicationID):
        """This function creates a new CExportObject for the given Blender Object. 
           If PureObject is true, it will result in a pure CObject (no derived 
           class). The new CExportObject uses the given duplication ID."""
        self.MarkAbstract();

    def StoreState(self):
        """This function has to be called before calling either ApplyAction(...) or 
           ApplyFrame(...) to ensure that RestoreState(...) works properly."""
        self.MarkAbstract();
    
    def GetNLAFrameList(self,MaxKeyDistance=0.0,DefaultFrame=1.0):
        """This function returns a sorted list of floating point values giving the 
           frames at which keys should be generated. The given maximal distance 
           between subsequent keys is used and if the sequence would be empty 
           otherwise the given default frame value is used."""
        FrameList=self.DoGetNLAFrameList();
        return CompleteFrameList(FrameList,MaxKeyDistance,DefaultFrame);

    def DoGetNLAFrameList(self):
        """This function implements GetNLAFrameList(...). It does not need to take 
           care of MaxKeyDistance, sorting or a DefaultFrame."""
        self.MarkAbstract();
        
    def GetActionNameList(self):
        """This function returns a list of strings, giving the names of all actions, 
           which are accessible through this model data. No name occurs twice."""
        self.MarkAbstract();

    def GetActionSource(self,ActionName):
        """This function returns the Blender object, which represents the action 
           with the given name. It is meant to be saved in a source list."""
        self.MarkAbstract();
        
    def GetActionRange(self,ActionName):
        """This function returns the frame at which the action with the given name 
           begins and the frame at which it ends (as pair of either int or float)."""
        self.MarkAbstract();
        
    def GetActionFrameList(self,ActionName,MaxKeyDistance=0.0,DefaultFrame=1.0):
        """This function returns all frames, which need to be keyed for the action 
           with the given name. If MaxKeyDistance is positive, it defines a maximal 
           distance between subsequent frames, which will be guaranteed. The result 
           is returned as list of sorted floats. If the result would be empty 
           otherwise it consists only of DefaultFrame."""
        # Get the frame list
        FrameList=self.DoGetActionFrameList(ActionName);
        # Apply the maximal key distance and the default frame
        return CompleteFrameList(FrameList,MaxKeyDistance,DefaultFrame);

    def DoGetActionFrameList(self,ActionName):
        """This function returns all frames, which have to be keyed for the action 
           with the given name as list of floats."""
        self.MarkAbstract();
        
    def ApplyAction(self,ActionName):
        """This function applies the action with the given name to all Blender 
           objects in this model and it resets the poses of all armatures. After 
           calling it you can play back the action using ApplyFrame(...) to record 
           it. Use RestoreState() to undo the changes caused by this function."""
        self.MarkAbstract();
        
    def ApplyFrame(self,Frame):
        """This function sets up the given frame for the scene, whose data is 
           accessed by this model. Use RestoreState() to undo the changes caused by 
           this function. Frame should be a floating point value."""
        self.MarkAbstract();
    
    def RestoreState(self):
        """This function reverts the changes caused by either ApplyAction(...) or 
           ApplyFrame(...). It can only work correctly, if StoreState(...) has been 
           called initially."""
        self.MarkAbstract();

    def GetArmatureData(self,Skeleton):
        """Given a CExportObjectBase that refers to a skeleton this function 
           returns a corresponding instance of a derived class of 
           CArmatureDataBase. On failure it returns None. Note that this is quite 
           expensive."""
        self.MarkAbstract();
        
    def IsValid(self):
        """This function returns true, if and only if this model data object is 
           ready to be used for reading data."""
        self.MarkAbstract();
        
    def MarkAbstract(self):
        """This function throws a not implemented exception. It is used to 
           indicate "abstract" functions."""
        raise NotImplemented("CModelDataBase is abstract. Use derived classes to get access to model data in a particular Blender version.");

    # The options that are to be used for reading model data
    Options=None;
    # This is an instance of a derived class of CBlenderUtilityBase
    BlenderUtility=None;


class CModel:
    """This function takes care of collecting all required data from a scene in
       Blender. What data is collected can be controlled through a 
       CModelOptions instance."""

    def __init__(self,ModelData,TotalProgress=1.0):
        """This function collects data from the current scene and saves it to 
           the member variables of this model.
           @param ModelData: An instance of a derived class of CModelDataBase 
                  giving the data from which this model is to be created. 
           @param TotalProgress: Defines the state of the progress bar once this 
                  function has finished."""
        self.CleanUp();
        Options=ModelData.Options;
        BlenderUtility=ModelData.BlenderUtility;
        # If the model data is invalid the whole process fails
        if(not ModelData.IsValid()):
            BlenderUtility.ShowMessage(
               "ERROR: Failed to retrieve model data.",
               "Error","Failed to retrieve model data.");
            self.pMesh=None;
            return;
        
        # Ensure that all temporary changes can be reverted correctly
        if(Options.ExportAnimations and Options.ExportConstraints):
            ModelData.StoreState();
    
        # Prepare some information here:
        BlenderUtility.SetProgress(0.0*TotalProgress,"Getting objects");
        # Get the relevant objects
        RelevantObjects=ModelData.GetExportObjectList();
        if(len(RelevantObjects)==0):
            if(Options.ExportSelectionOnly):
                BlenderUtility.ShowMessage(
                   "ERROR: No exportable objects selected for export.",
                   "Error","No objects selected for export.");
            else:
                BlenderUtility.ShowMessage(
                   "ERROR: No exportable object contained in the scene.",
                   "Error","No exportable object contained in the scene.");
            self.pMesh=None;
            return;
    
        # Resolve groups in the list of relevant objects. Keep track of the mapping 
        # from duplication IDs to duplication root objects.
        DuplicationRootDict=dict();
        GroupsResolved=self.ResolveGroups(RelevantObjects,DuplicationRootDict);
    
        # For every CExportObject find the CExportObject, which belongs to the Blender
        # parent object
        for ExportObject in RelevantObjects:
            if(ExportObject.BlenderParent!=None):
                # Try to find the parent in the same duplication
                for ParentObject in RelevantObjects:
                    if(ExportObject.IsChildOf(ParentObject)):
                        ExportObject.Parent=ParentObject;
                        break;
                # If no parent has been found a new CExportObject is needed
                if(ExportObject.Parent==None):
                    # Check, whether skinning or bone parenting is used
                    CompleteCopyNeeded=(ExportObject.BlenderParentBone!=None or ExportObject.UsesSkinning());
                    NewParent=ModelData.CreateExportObject(ExportObject.BlenderParent,not CompleteCopyNeeded,ExportObject.DuplicationID);
                    if(CompleteCopyNeeded):
                        print("WARNING: The object \""+ExportObject.GetObjectName()+"\" is being duplicated through the duplicate groups setting. However its parent \""+NewParent.GetObjectName()+"\" is not duplicated this way in Blender. Since the parent is needed the exporter has duplicated it anyway.");
                    ExportObject.Parent=NewParent;
                    RelevantObjects.append(NewParent);
            elif(ExportObject.DuplicationID!=""):
                # Duplicated root objects use the duplication root as parent
                ExportObject.Parent=DuplicationRootDict[ExportObject.DuplicationID];
            # Tell the parent about its new child
            if(ExportObject.Parent!=None):
                ExportObject.Parent.ChildList.append(ExportObject);
        
        # The list of CExportObject instances needs to be resorted such that the
        # parent of an object always has a lower index. Perform a DFS and sort by
        # DFS numbers to do this (works, because this is a tree).
        piDFS=[0];
        for ExportObject in RelevantObjects:
            if(ExportObject.Parent==None):
                ExportObject.DoDFS(piDFS);
        for ExportObject in RelevantObjects:
            if(ExportObject.iDFS<0):
                BlenderUtility.ShowMessage(
                   "ERROR: There appears to be a deadlock in your object dependencies. A parent depends upon its child. Exporting has been aborted.",
                   "Error","Deadlock in parenting relations.");
                return;
        SortedRelevantObjects=[None]*len(RelevantObjects);
        for ExportObject in RelevantObjects:
            SortedRelevantObjects[ExportObject.iDFS]=ExportObject;
        RelevantObjects=SortedRelevantObjects;
    
        # If necessary check whether all meshes have identical UV layers
        UVLayerSet=None;
        if(Options.ExportAllUVLayers):
            UVLayerSet=ModelData.GetUVLayerSet(RelevantObjects);
            if(UVLayerSet==None):
                BlenderUtility.ShowMessage(
                   "Failed to export all UV layers. The option has been disabled.","Warning");
                Options.ExportAllUVLayers=False;
        # If the UV layers can be exported output their order
        if(Options.ExportAllUVLayers):
            print("All UV layers will be exported. Here is the mapping from UV layers to texture coordinate set indices:");
            for iTexCoordSet,UVLayer in enumerate(UVLayerSet):
                print(UVLayer+" -> "+str(iTexCoordSet));
            self.UVLayerSet=UVLayerSet;
    
        # Initialize the list of all materials. The meshes append to it.
        BlenderMaterialList=[];
        
        # Create a corresponding CObject for every CExportObject in RelevantObjects
        BlenderUtility.SetProgress(0.1*TotalProgress,"Reading objects");
        ObjectProgress=0.9/max(1,len(RelevantObjects));
        if(Options.ExportAnimations and Options.ExportConstraints):
            ObjectProgress*=0.5/0.9;
        self.pObject=[];
        self.pMesh=[];
        self.pMeshObject=[];
        self.pSkeleton=[];
        for iExportObject,ExportObject in enumerate(RelevantObjects):
            # Get the data that is related to the export object
            print("Reading the Blender object \""+ExportObject.GetObjectName()+"\".");
            BlenderUtility.SetProgress((0.1+ObjectProgress*iExportObject)*TotalProgress,"Reading "+ExportObject.GetObjectName());
            ObjectParent=None;
            if(ExportObject.Parent!=None):
                ObjectParent=ExportObject.Parent.Object;
            # Create the object
            Object=ModelData.ExportObjectToObject(ExportObject,ObjectParent,BlenderMaterialList,self.pMesh,self.UVLayerSet,DuplicationRootDict);
            # Associate the new object with the export object and save it
            ExportObject.Object=Object;
            self.pObject.append(Object);
            if(isinstance(Object,CSkeleton)):
                self.pSkeleton.append(Object);
            if(isinstance(Object,CMeshObject)):
                self.pMeshObject.append(Object);
            # Read animation data from action curves, if requested
            if(Options.ExportAnimations and not Options.ExportConstraints):
                Object.AddKeysFromAllActionCurves(Options.MaxKeyDistance);
            # If the object has a bone as parent find the right CBone
            if(ExportObject.BlenderParentBone!=None):
                if(isinstance(ObjectParent,CSkeleton)):
                    for Bone in ObjectParent.BoneList:
                        if(Bone.Name==ExportObject.GetParentBoneName()):
                            Object.SetParentBone(Bone);
    
        # If constraints should be exported create keys from poses now
        if(Options.ExportAnimations and Options.ExportConstraints):
            # Generate a list of skeletons with corresponding armature data, for 
            # which keys need to be generated 
            SkeletonList=list();
            for SkeletonObject in RelevantObjects:
                if(isinstance(SkeletonObject.Object,CSkeleton)):
                    Skeleton=SkeletonObject.Object;
                    ArmatureData=ModelData.GetArmatureData(SkeletonObject);
                    if(ArmatureData!=None):
                        SkeletonList.append((Skeleton,ArmatureData));
            # Iterate over all actions (or just over the single NLA)
            if(Options.ExportCurrentNLA):
                ActionNameList=["CurrentNLA"];
            else:
                ActionNameList=ModelData.GetActionNameList();
            ActionProgress=0.3/max(1,len(ActionNameList));
            for iAction,ActionName in enumerate(ActionNameList):
                BlenderUtility.SetProgress((0.6+ActionProgress*iAction)*TotalProgress,"Action: "+ActionName);
                # Set this action as current action for all objects (and reset 
                # poses)
                if(not Options.ExportCurrentNLA):
                    ModelData.ApplyAction(ActionName);
                # Get a list of frames, which need to be keyed for the action or the 
                # NLA
                if(Options.ExportCurrentNLA):
                    ActionFrameList=ModelData.GetNLAFrameList(Options.MaxKeyDistance);
                else:
                    ActionFrameList=ModelData.GetActionFrameList(ActionName,Options.MaxKeyDistance);
                # Iterate over all these frames
                for Frame in ActionFrameList:
                    # Make this frame the current frame in the scene
                    ModelData.ApplyFrame(Frame);
                    # Iterate over all skeletons to key the current pose
                    for Skeleton,ArmatureData in SkeletonList:
                        if(not Skeleton.AddKeysFromPose(ActionName,Frame,ArmatureData)):
                            print("Failed to generate some keys for the skeleton \""+Skeleton.Name+"\".");
                    # Iterate over all objects to key the current transformation
                    for Object in self.pObject:
                        Object.AddKeysFromState(ActionName,Frame);
                ModelData.RestoreState();
            # Fix all rotation keys
            for Object in self.pObject:
                Object.FixRotationKeys();

        # Output some information
        print(str(len(self.pMesh))+" different meshes have been read.");
        print(str(len(self.pMeshObject))+" mesh objects have been read.");
        print(str(len(self.pSkeleton))+" skeletons (resp. armatures) have been read.");
        print(str(self.GetBoneCount())+" bones have been read.");
        print(str(len(self.pObject)-len(self.pMeshObject)-len(self.pSkeleton))+" other objects have been read.");
    
        # Read the materials
        BlenderUtility.SetProgress(0.9*TotalProgress,"Reading materials");
        iMaterial=0;
        self.pMaterial=[];
        for BlenderMaterial in BlenderMaterialList:
            MaterialOut=CMaterial(iMaterial,ModelData.GetMaterialData(BlenderMaterial,UVLayerSet));
            self.pMaterial.append(MaterialOut);
            iMaterial+=1;
        if(len(BlenderMaterialList)):
            print(str(self.GetMaterialCount())+" materials have been read.");
        # If there's no material in the material list create a default material
        elif(Options.CreateDefaultMaterial):
            self.pMaterial.append(CMaterial());
            print("The model didn't use any materials. A white, untextured default material has been created.");
        else:
            print("There are no materials at all. This may lead to problems.");
    
        print("All data has been read successfully.");
        
        if(GroupsResolved):
            self.MakeUniqueObjectNames();
            print("Made object names unique.");
        BlenderUtility.SetProgress(1.0*TotalProgress,"Read model");

    def ApplyTransformation(self,Transformation):
        """This function transforms the entire model by the given transformation. It
           does this by multiplying the transformation of all CObjects without a
           parent by the given transformation."""
        for Object in self.pObject:
            if(Object.Parent==None):
                Object.Transformation*=Transformation;


    def TransformTime(self,FrameOffset,FrameFactor=1.0):
        """This function multiplies all frame values in all animations by the 
           given factor and then adds the given offset."""
        for Skeleton in self.pSkeleton:
            Skeleton.TransformTime(FrameOffset,FrameFactor);
        for Object in self.pObject:
            Object.TransformationDict.TransformTime(FrameOffset,FrameFactor);


    def ApplyTransformations(self):
        """This function ensures that every mesh object has an individual CMesh and
           applies the transformation of the mesh object to the vertex data of the
           mesh. The transformation of the mesh object is set to identity and its
           parent is set to None unless it is a CSkeleton."""
        # Apply transformations to the mesh data
        self.pMesh=[];
        for MeshObject in self.pMeshObject:
            MeshObject.Mesh=MeshObject.Mesh.Clone();
            MeshObject.Mesh.iMesh=len(self.pMesh);
            self.pMesh.append(MeshObject.Mesh);
            # Rename the mesh to the name of the mesh object
            MeshObject.Mesh.Name=MeshObject.Name;
            # Apply the mesh transformation
            MeshObject.Mesh.ApplyTransformation(MeshObject.GetTransformation());
        # Clear the transformations and parenting relations
        for MeshObject in self.pMeshObject:
            # Overwrite the transformation in the mesh object with the identity
            # transformation
            MeshObject.Transformation=CMatrix4x4();
            # Set the parent to None unless it is a skeleton
            if(not isinstance(MeshObject.Parent,CSkeleton)):
                MeshObject.Parent=None;


    def ConvertObjectTransformationsToBones(self):
        """This function makes all animated object transformations, become constant 
           identity transformations without losing the animations. It does this by 
           generating a dummy skeleton with a single dummy bone for every animated 
           object transformation. The dummy skeleton is parented to the parent of 
           the object and the object is parented to the bone. The bone stores the 
           animation of the object. This way object animations are reduced to 
           skeletal animations, which makes it easier to export them. It ignores 
           objects, which use skinning, because then the object can not be parented 
           to a bone and object animations would yield odd results anyway. If 
           pObject used to be in parent-first order, this order is kept intact. A 
           list of the indices of newly created skeletons in pObject is returned."""
        DummyIndexList=list();
        # Iterate over all animated objects
        iObject=0;
        iSkeleton=0;
        while(iObject<len(self.pObject)):
            Object=self.pObject[iObject];
            if(Object.IsAnimated()):
                print("Creating a dummy skeleton for the object \""+Object.Name+"\" to represent the object animation.");
                # Create the dummy skeleton
                DummySkeleton=CSkeleton(None,None,None,None,None,None);
                DummySkeleton.MakeDummySkeleton(Object.TransformationDict,Object.Transformation,Object.Name+"AnimationDummy",Object.Name+"AnimationDummy");
                # Parent it to the parent of the object without changing its 
                # transformation
                DummySkeleton.Parent=Object.Parent;
                DummySkeleton.ParentBone=Object.ParentBone;
                # Parent the object to the dummy bone
                Object.Parent=DummySkeleton;
                Object.ParentForSkinning=False;
                Object.ParentBone=DummySkeleton.BoneList[0];
                # Reset the transformation. The parent describes it completely.
                Object.Transformation=CMatrix4x4();
                Object.TransformationDict=CKeyedTransformationDict();
                # Copy over the source list
                DummySkeleton.SourceList=list(Object.SourceList);
                # Store the new skeleton without destroying the parent first order
                self.pObject=self.pObject[0:iObject]+[DummySkeleton]+self.pObject[iObject:];
                self.pSkeleton=self.pSkeleton[0:iSkeleton]+[DummySkeleton]+self.pSkeleton[iSkeleton:];
                DummyIndexList.append(iObject);
                iObject+=1;
                iSkeleton+=1;
            iObject+=1;
            if(isinstance(Object,CSkeleton)):
                iSkeleton+=1;
        # Ensure unique object names
        self.MakeUniqueObjectNames();
        return DummyIndexList;

    def SimplifyModel(self):
        """This function simplifies this model, by combining all meshes and
           skeletons. Once it has finished the meshes described by all mesh objects
           are part of the only remaining CMesh in self.pMesh and this CMesh is used 
           only by a single CMeshObject, which uses the identity transformation. All
           skeletons are combined as well and the weights of the mesh are changed
           accordingly. If there is a skeleton the only remaining mesh object uses it
           as parent."""
        # Get rid of object animations
        self.ConvertObjectTransformationsToBones();
        # If there is a skeleton create a new skeleton to replace it
        NewSkeleton=None;
        if(len(self.pSkeleton)>0):
            NewSkeleton=CSkeleton(None,None,None,None,None,None);
            NewSkeleton.Name="Skeleton";
        # Check if there is a skinned and an unskinned mesh. If so an 
        # untransformed bone is needed for the unskinned mesh.
        SkinningNeeded=(len(self.pSkeleton)>0);
        UnskinnedMeshExists=False;
        for Mesh in self.pMesh:
            if(not Mesh.SkinningUsed):
                UnskinnedMeshExists=True;
        if(SkinningNeeded and UnskinnedMeshExists):
            NewBone=CBone(None,None,None,None,None,None,None,None);
            NewBone.Name="IdentityBone";
            NewSkeleton.BoneList.append(NewBone);
            print("A new bone called \"IdentityBone\" has been added to the skeleton to function as parent for meshes that used to be unskinned.");
        # Iterate through all skeletons and append them to the new skeleton. Remember
        # the new bone indices for every skeleton.
        SkeletonIndexDict=dict([]);
        for SkeletonSource in self.pSkeleton:
            # Copy the skeleton
            Skeleton=SkeletonSource.Clone();
            # Check if there is a parent bone. Store the relative transformation.
            Parent=Skeleton;
            Transformation=CMatrix4x4();
            ParentBone=None;
            while(Parent!=None and ParentBone==None):
                Transformation*=Parent.Transformation;
                ParentBone=Parent.ParentBone;
                Parent=Parent.Parent;
            # Apply the transformation (either relative to the parent bone or
            # relative to the root). Eliminate the parent.
            Skeleton.ApplyDataTransformation(Transformation);
            if(ParentBone!=None):
                Skeleton.Parent=Parent;
                Skeleton.ParentBone=ParentBone;
            else:
                Skeleton.Parent=None;
                Skeleton.ParentBone=None;
            # If there is a parent bone set it as parent for all root bones now
            if(ParentBone!=None):
                iParentBone=SkeletonIndexDict[Skeleton.Parent.Name][ParentBone.iBone];
                for Bone in Skeleton.BoneList:
                    if(Bone.Parent==None):
                        Bone.Parent=NewSkeleton.BoneList[iParentBone];
            # Append the skeleton. Remember the index mapping.
            SkeletonIndexDict[Skeleton.Name]=range(len(NewSkeleton.BoneList),len(NewSkeleton.BoneList)+len(Skeleton.BoneList));
            NewSkeleton.BoneList+=Skeleton.BoneList;
            NewSkeleton.SourceList+=Skeleton.SourceList;
            # Add actions which are not part of the list in this skeleton
            for Action in Skeleton.ActionList:
                if(not Action in NewSkeleton.ActionList):
                    NewSkeleton.ActionList.append(Action);
        # Update bone indices within the new skeleton
        if(NewSkeleton!=None):
            for iBone,Bone in enumerate(NewSkeleton.BoneList):
                Bone.iBone=iBone;
        # If there is a mesh, create a new output mesh
        NewMesh=None;
        if(len(self.pMesh)>0):
            NewMesh=CMesh(None,None);
            NewMesh.Name="Mesh";
            if(SkinningNeeded):
                NewMesh.iMaxBone=len(NewSkeleton.BoneList)-1;
                NewMesh.SkinningUsed=True;
        # Iterate through all mesh objects
        for MeshObject in self.pMeshObject:
            # Find the first parent, which is either a bone, a skeleton or the root.
            # Determine the relative transformation.
            Transformation=CMatrix4x4();
            Parent=MeshObject;
            ParentBone=None;
            ParentForSkinning=False;
            while(Parent!=None and not (isinstance(Parent,CSkeleton) and (ParentForSkinning or ParentBone!=None))):
                Transformation*=Parent.Transformation;
                ParentBone=Parent.ParentBone;
                ParentForSkinning=Parent.ParentForSkinning;
                Parent=Parent.Parent;
            # Create a copy of the mesh
            Mesh=MeshObject.Mesh.Clone();
            # If a parent bone is used a quite special transformation needs to be
            # applied
            if(ParentBone!=None):
                # The final mesh transformation will be:
                # ParentBone.MeshTransformation*ParentBone.BoneToWorldSpace
                # The desired transformation is:
                # Transformation*ParentBone.BoneToWorldSpace
                # Therefore Transformation and the inverse mesh transformation get
                # applied here.
                Mesh.ApplyTransformation(Transformation*ParentBone.MeshTransformation.ComputeInverse());
            # Otherwise the mesh is just brought to world space
            else:
                Mesh.ApplyTransformation(MeshObject.GetTransformation());

            # Update the skinning of the source mesh
            # If the mesh is skinned already, the indices need to be updated
            if(Mesh.SkinningUsed):
                Mesh.UpdateSkinningIndices(SkeletonIndexDict[Parent.Name]);
            # If the mesh did not use skinning, but has a bone as parent it needs to
            # be skinned to this bone
            elif(ParentBone!=None):
                iParentBone=SkeletonIndexDict[Parent.Name][ParentBone.iBone];
                Mesh.SkinToSingleBone(iParentBone);
            # If the mesh is not influenced by any bone, but the destination mesh
            # uses skinning, then it needs to be skinned to the identity bone
            elif(NewMesh.SkinningUsed):
                Mesh.SkinToSingleBone(0);

            # Update the skinning of the destination mesh. If the next mesh uses
            # skinning, but the destination doesn't the destination mesh needs to be
            # skinned to the identity bone.
            if(Mesh.SkinningUsed and not NewMesh.SkinningUsed):
                NewMesh.SkinToSingleBone(0);

            # Ensure that the UV formats match
            Mesh.MakeDefaultTexCoords(max(Mesh.nTexCoordSet,NewMesh.nTexCoordSet));
            NewMesh.MakeDefaultTexCoords(max(Mesh.nTexCoordSet,NewMesh.nTexCoordSet));

            # Append the mesh
            NewMesh.AppendMesh(Mesh);

        # Use the simplified objects
        NewMeshObject=CMeshObject(None,None,None,None,None,None);
        NewMeshObject.Name="MeshObject";
        NewMeshObject.SourceList=[];
        for MeshObject in self.pMeshObject:
            NewMeshObject.SourceList.append(MeshObject.SourceList);
        NewMeshObject.Mesh=NewMesh;
        self.pMeshObject=list([NewMeshObject]);
        self.pMesh=list([NewMesh]);
        self.pObject=list([NewMeshObject]);
        print("All mesh objects have been combined into a single mesh object.");
        if(len(self.pSkeleton)>0):
            self.pObject.append(NewSkeleton);
            self.pSkeleton=list([NewSkeleton]);
            NewMeshObject.Parent=NewSkeleton;
            print("All skeletons have been combined into a single skeleton.");


    def UpdateMeshList(self):
        """This function searches for meshes, which are not used by any mesh object.
           If it finds one, it removes it from the list self.pMesh. CMesh.iMesh is
           updated accordingly. Besides it adds meshes, which are not in the mesh
           list so far."""
        pMeshUsed=[False]*self.GetMeshCount();
        for MeshObject in self.pMeshObject:
            if(MeshObject.Mesh in self.pMesh):
                pMeshUsed[MeshObject.Mesh.iMesh]=True;
        NewMeshList=[];
        for i in range(self.GetMeshCount()):
            if(pMeshUsed[i]):
                self.pMesh[i].iMesh=len(NewMeshList);
                NewMeshList.append(self.pMesh[i]);
        # Add meshes which are not yet part of the list
        for MeshObject in self.pMeshObject:
            if(not MeshObject.Mesh in self.pMesh):
                MeshObject.Mesh.iMesh=len(NewMeshList);
                NewMeshList.append(MeshObject.Mesh);
        self.pMesh=NewMeshList;


    def UnifyTexCoordSetCount(self,nTexCoordSet=0,TexCoordGetter=GetDefaultTexCoord):
        """This function does two things. It ensures that all meshes in this model
           have the same number of texture coordinate sets and it ensures that all of
           them have at least the given number of texture coordinate sets. It does
           this by adding texture coordinate sets using the given getter function.
           See GetDefaultTexCoord(...) for the signature of the needed function."""
        # Get the needed number of texture coordinate sets
        for Mesh in self.pMesh:
            nTexCoordSet=max(nTexCoordSet,Mesh.nTexCoordSet);
        # Ensure that they are available
        for Mesh in self.pMesh:
            Mesh.MakeDefaultTexCoords(nTexCoordSet,TexCoordGetter);


    def MakeUniqueNames(self,List,RegisteredNameList=list()):
        """This function ensures that the objects in the given list have unique
           names. The names have to be accessible through a member variable called
           Name. If a name exists twice an index in the common Blender format .XXX
           gets appended. RegisteredNameList can be used to specify additional names,
           which must not be used."""
        # Create a set of all names and a list of objects with non-unique names,
        # whose names have already been added to the set
        DoubleNameList=[];
        NameSet=set(RegisteredNameList);
        for Object in List:
            if(Object.Name in NameSet):
                DoubleNameList.append(Object);
            else:
                NameSet|=set([Object.Name]);
        # Iterate through the objects with non-unique names to make them unique
        for Object in DoubleNameList:
            # Check whether there already is an index suffix in the name. If so
            # remove it.
            PureName=Object.Name;
            iSuffix=PureName.rfind(".");
            if(iSuffix>=0 and iSuffix==len(PureName)-4):
                if(PureName[iSuffix+1:].isdigit()):
                    PureName=PureName[0:iSuffix];
            # Find an unused suffix (or use no suffix)
            Suffix=0;
            while(True):
                NewName=PureName;
                if(Suffix>0):
                    NewName+=".%03d"%Suffix;
                if(not NewName in NameSet):
                    # If the name is unique used it and mark it as used
                    NameSet|=set([NewName]);
                    Object.Name=NewName;
                    break;
                Suffix+=1;
                
    def MakeUniqueObjectNames(self,RegisteredNameList=list()):
        """This function ensures that all objects in self.pObject have unique names. 
           See MakeUniqueNames(...) for details."""
        self.MakeUniqueNames(self.pObject,RegisteredNameList);

    def MakeUniqueMeshNames(self,RegisteredNameList=list()):
        """This function ensures that all meshes in self.pMesh have unique names. See
           MakeUniqueNames(...) for more details. It is recommended to call it once
           after creating meshes or changing mesh names. RegisteredNameList can be
           used to specify additional names, which must not be used."""
        self.MakeUniqueNames(self.pMesh);

    def MakeUniqueBoneNames(self,RegisteredNameList=list()):
        """This function ensures that all bone names are unique. It does not just
           make them unique within the skeletons, but within the whole model.
           RegisteredNameList can be used to specify additional names, which must not
           be used."""
        # Create a list of all bones
        BoneList=[];
        for Skeleton in self.pSkeleton:
            BoneList+=Skeleton.BoneList;
        # Make their names unique
        self.MakeUniqueNames(BoneList);
        # Ensure that the names of the keyed transformations still match the bone
        # names
        for Bone in BoneList:
            for KeyedTransformation in Bone.KeyDict.values():
                KeyedTransformation.Name=Bone.Name;

    def GetMeshCount(self):
        """This function returns the number of different meshes in this model."""
        return len(self.pMesh);

    def GetMeshObjectCount(self):
        """This function returns the number of instantiations of meshes in this
           model (i.e. how many meshes need to be rendered for this model)."""
        return len(self.pMeshObject);

    def GetAnimationRange(self):
        """This function returns the minimal frame value and the maximal frame 
           value that occur in animations for the skeletons of this model. It 
           returns a float tuple of shape (MinFrame,MaxFrame). If there are no 
           skeletons or no animations it returns (0.0,0.0)."""
        MinFrame=1.0e20;
        MaxFrame=-1.0e20;
        for Skeleton in self.pSkeleton:
            (NextMinFrame,NextMaxFrame)=Skeleton.GetAnimationRange();
            if(NextMinFrame!=None):
                MinFrame=min(NextMinFrame,MinFrame);
                MaxFrame=max(NextMaxFrame,MaxFrame);
        if(MinFrame!=1.0e20):
            return (MinFrame,MaxFrame);
        return (0.0,0.0);

    def GetMaterialCount(self):
        """This function returns the number of different materials in this model."""
        return len(self.pMaterial);

    def GetBoneCount(self):
        """This function returns the total number of bones in this model. This is the
           sum of the bone counts in the single skeletons."""
        nBone=0;
        for Skeleton in self.pSkeleton:
            nBone=len(Skeleton.BoneList);
        return nBone;

    def GetTexCoordSetCount(self):
        """This function returns the number of different UV layers in the meshes of
           this model. It is guaranteed that it is the same for all meshes."""
        if(self.UVLayerSet!=None):
            return len(self.UVLayerSet);
        else:
            return 1;

    def CleanUp(self):
        """This function sets all attributes to reasonable, independent, empty 
           default values."""
        self.pMesh=list();
        self.UVLayerSet=None;
        self.pMeshObject=list();
        self.pSkeleton=list();
        self.pObject=list();
        self.pMaterial=list();

    def IsValid(self):
        """This function returns true if and only if, this model is in a valid
           state."""
        return (self.pMesh!=None and self.pMeshObject!=None and self.pSkeleton!=None and self.pObject!=None
                and (len(self.pMeshObject)>0 or len(self.pSkeleton)>0));
                
    def HasGeometry(self):
        """This function returns true, iff this model is valid and defines at least 
           one mesh object with at least one triangle."""
        return self.IsValid() and len(self.pMeshObject)>0 and len(self.pMesh)>0;

    # This member holds a list of CMesh objects giving all meshes for this model. If
    # the model is invalid it is None or empty.
    pMesh=[];
    # This is a set holding the names of all UV layers, which are used in the meshes
    # of this model or None, if exporting multiple UV layers is disabled. Iterate
    # through it to get the order of the UV layers.
    UVLayerSet=None;
    # This member holds a list of all CMeshObject instances, which define
    # instantiations of meshes in this model. It is a sub list pObject.
    pMeshObject=[];
    # This member holds a list of all skeletons in this model as CSkeleton. It is a
    # sub-list of pObject.
    pSkeleton=[];
    # This member holds a list of all CObject instances, this model consists of. This
    # includes all mesh objects, skeletons and possibly some other dummy objects.
    pObject=[];
    # This member holds a list of CMaterial objects giving the materials for this
    # model
    pMaterial=[];


class CBlenderUtilityBase:
    """This is an abstract base class that provides access to several features 
       of the Blender Python API in a way that is independent from the Blender 
       version."""
    
    def __init__(self):
        """Initializes the base part of this class with default data."""
        pass;
        
    def PrepareExport(self,Backup=True):
        """This function ensures that displaying a progress bar is possible and that 
           the current state of the Blender model can be restored after export. The 
           latter step is only performed, if you pass True for Backup. Use 
           FinalizeExport(...) to restore the original state of the model."""
        self.MarkAbstract();
    
    def FinalizeExport(self):
        """This function undoes everything, which has been done by the last call to 
           PrepareExport(...). If PrepareExport(...) has created a backup, it also 
           undoes everything, which has been done in between."""
        self.MarkAbstract();

    def ShowMessage(self,LongText,Title=None,ShortText=None):
        """This function prints LongText to the console. If ShortText is not 
           None it also shows ShortText in a pop up in Blender's UI. If it is 
           None it will show LongText in a pop up. Title gets used as title for 
           the pop up (if given)."""
        self.MarkAbstract();
        
    def SetProgress(self,Progress,State):
        """This function updates the progres bar. Progress should be a floating 
           point value between 0.0 and 1.0 indicating the current progress. 
           State should be a very short string (around 10 letters) describing 
           the current task."""
        self.MarkAbstract();

    def MarkAbstract(self):
        """This function throws a not implemented exception. It is used to 
           indicate "abstract" functions."""
        raise NotImplemented("CBlenderBase is abstract, use derived classes to get access to the features of a particular Blender version.");


class CEnumeration:
    """This class describes an enumeration in which every value is represented 
       through a string and comes with a name and a precise description."""
       
    def __init__(self):
        """This constructor initializes this enumeration without entries."""
        self._EntryList=list();
        self._NameDict=dict();
        self._DescriptionDict=dict();
        
    def AddEntry(self,EntryID,EntryName,EntryDescription):
        """This function appends an entry with the given ID to the end of this 
           enumeration and associates it with the given human-readable name and 
           description."""
        if(not EntryID in self._NameDict):
            self._EntryList.append(EntryID);
        self._NameDict[EntryID]=EntryName;
        self._DescriptionDict[EntryID]=EntryDescription;
    
    def GetEntryList(self):
        """This function returns a list holding the IDs of all entries in this 
           enumeration in the correct order."""
        return list(self._EntryList);
    
    def GetEntryName(self,EntryID):
        """This function returns the human readable name of the entry with the given 
           ID or None, if there is no such entry."""
        if(EntryID in self._NameDict):
            return self._NameDict[EntryID];
        return None;
    
    def GetEntryDescription(self,EntryID):
        """This function returns the description of the entry with the given ID or 
           None, if there is no such entry."""
        if(EntryID in self._DescriptionDict):
            return self._DescriptionDict[EntryID];
        return None;
    
    # This is a list holding the identifiers of all entries in the correct order
    _EntryList=list();
    # This dictionary maps entry IDs to corresponding human-readable names
    _NameDict=dict();
    # This dictionary maps entry IDs to corresponding descriptions
    _DescriptionDict=dict();


class CProperty:
    """This class represents a property of objects, which can be set by the 
       client and is used during export."""
       
    def __init__(self,ID,Name,Description,Type=str,Default=None,ValueRange=None):
        """This constructor initializes this property using the given values for its 
           attributes. If no default value is given, an empty string, a zero value 
           or the first enumeration entry are used. For integer and float properties 
           you can pass a minimal and maximal value in a tuple to ValueRange."""
        self._ID=ID;
        self._Name=Name;
        self._Description=Description;
        self._Type=Type;
        if(ValueRange!=None and (Type==int or Type==float)):
            self._ValueRange=(Type(ValueRange[0]),Type(ValueRange[1]));
        else:
            self._ValueRange=None;
        if(Default==None):
            if(self._Type in [int,str,bool,float]):
                self._Default=self._Type();
            else:
                self._Default=self._Type.GetEntryList()[0];
        else:
            if(self._Type in [int,str,bool,float]):
                self._Default=self._Type(Default);
            else:
                self._Default=str(Default);
        
    def GetID(self):
        """This function returns the ID, which should be used for variables storing 
           this property."""
        return self._ID;
    
    def GetName(self):
        """This function returns the human readable name of this property."""
        return self._Name;
    
    def GetDescription(self):
        """This function returns the description of this property."""
        return self._Description;
    
    def GetType(self):
        """This function returns the type of this property. It is either bool, int, 
           str or float or an instance of Cenumeration."""
        return self._Type;
    
    def GetDefaultValue(self):
        """This function returns the value, which is used for this property, if the 
           client has not set a value yet."""
        return self._Default;
    
    def GetValueRange(self):
        """If this property is an integer property or a floating point property with 
           a previously defined value range, this function returns this value range. 
           This means that it returns a tuple of the form 
           (MinimalValue,MaximalValue). Otherwise it returns None."""
        return self._ValueRange;
    
    # This method is defined in both Blender version specific modules.    
    #def GetValue(BlenderObject):
    #    """This function returns the value, which is currently set for this 
    #       property in the given Blender object or the default value, if None is 
    #       set. In all cases it returns an object of the type implied by GetType(). 
    #       For enumerations it returns the entry ID."""
    
    # This string gives the identifier, which is used to store this property in 
    # Blender objects
    _ID="";
    # This string gives a short, human readable name for this variable
    _Name="";
    # This string gives a precise and complete description of the meaning of this 
    # property
    _Description="";
    # The type of this property. It is either a Python class object or an instance 
    # of the class CEnumeration  
    _Type=None;
    # The default value of this property. Its type depends upon _Type.
    _Default=None;


class CPropertyManager:
    """This class manages all existing properties and organizes them in groups."""
    
    def __init__(self):
        """This default constructor initializes this class without any properties or 
           groups."""
        self._GroupNameList=list();
        self._GroupPropertyListDict=dict();
        self._PropertyDict=dict();
    
    def AddGroup(self,GroupName):
        """This function adds a new group with the given human readable name. It 
           fails returning False, if a group with this name exists already. 
           Otherwise it returns True."""
        if(GroupName in self._GroupPropertyListDict):
            return False;
        self._GroupNameList.append(GroupName);
        self._GroupPropertyListDict[GroupName]=list();
        return True;
    
    def AddProperty(self,Property,GroupName=None):
        """This function appends the given CProperty to the end of the group with 
           the given name. If GroupName is omitted it adds the property to the 
           group, which has been added last. It fails returning False, if such a 
           group does not exist. Otherwise it returns True."""
        if(GroupName==None):
            if(len(self._GroupNameList)==0):
                return False;
            GroupName=self._GroupNameList[len(self._GroupNameList)-1];
        elif(not GroupName in self._GroupPropertyListDict):
            return False;
        self._GroupPropertyListDict[GroupName].append(Property);
        self._PropertyDict[Property.GetID()]=Property;
        return True;
    
    def GetGroupList(self):
        """This function returns a list holding the names of all property groups in 
           this manager."""
        return list(self._GroupNameList);
    
    def GetGroupPropertyList(self,GroupName):
        """This function returns a list of all properties, which are part of the 
           group with the given name or None, if there is no such group."""
        if(GroupName in self._GroupPropertyListDict):
            return self._GroupPropertyListDict[GroupName];
        return None;
    
    def GetProperty(self,PropertyID):
        """If this property manger handles a property with the given ID, this 
           function returns it. Otherwise it returns None."""
        if(PropertyID in self._PropertyDict):
            return self._PropertyDict[PropertyID];
        return None;
    
    # This is a list holding the (human readable) names of all groups in this 
    # manager in correct order
    _GroupNameList=list();
    # This is a dictionary, which maps group names to corresponding lists of 
    # properties. The lists are in the correct order.
    _GroupPropertyListDict=dict();
    # This dictionary maps property IDs to corresponding properties in this manager. 
    # It is redundant, but useful for efficient look-ups.
    _PropertyDict=dict();


# export_utility_shared END

import Blender
from Blender import Scene
from Blender import Group
from Blender import Mathutils
from Blender import Mesh
from Blender import Modifier
from Blender import Armature
from Blender import Draw
from Blender import Window


def CleanPath(Path):
    """This little utility function brings a path into a normalized form by removing
       any redundant use of ".." or ".". The result is returned. As directory
       separator it uses "/". E.g.:
       CleanPath("c:\\Test/..\\Abc")=="c:/Abc" and
       CleanPath("Directory/NoDirectory\\..")=="Directory/" """
    Separator=Blender.sys.sep;
    return Blender.sys.cleanpath(Path.replace("/",Separator).replace("\\",Separator)).replace(Separator,"/");

def MakeRelativePath(Path,StartPath):
    """This little utility function makes the given path relative to StartPath. The
       resulting path is cleaned automatically (see CleanPath(...))."""
    Path=CleanPath(Path).replace("/",Blender.sys.sep);
    StartPath=CleanPath(StartPath).replace("/",Blender.sys.sep);
    return CleanPath(Blender.sys.relpath(Path,StartPath));

def MakeAbsolutePath(Path):
    """This function makes the given path absolute and cleans it. If it is relative 
       it is made absolute using the directory of the current Blend file."""
    if(Path.find(":")>=0):
        return CleanPath(Path);
    else:
        Path=CleanPath(Path).replace("/",Blender.sys.sep);
        return CleanPath(Blender.sys.expandpath("//"+Path));


def GetBlenderVersion():
    """This function returns a human readable string giving the used version of 
       Blender (e.g. "2.49" or "2.58a")."""
    return str(Blender.Get("version"));


def GetCurrentBlendFile():
    """This function returns the absolute path and filename of the *.blend file, 
       which is currently opened in the Blender instance, which executes this 
       script or an empty string, if no *.blend file is opened."""
    Result=Blender.Get("filename");
    if(Result==None):
        return "";
    else:
        return MakeAbsolutePath(Result);


# class CQuaternion
def CQuaternionMakeRotationAlongAxis(self,Axis,Angle):
    """This function makes this quaternion represent a rotation around the given
       axis by the given angle (in radians)."""
    # This function uses the equivalent Blender.Mathutils function to avoid that
    # the module math needs to be important for sin and cos
    self.Copy(Mathutils.Quaternion([Axis.x,Axis.y,Axis.z],Angle*180.0/3.14159265358979));
    self.Normalize();
CQuaternion.MakeRotationAlongAxis=CQuaternionMakeRotationAlongAxis;

def CQuaternionSlerp(self,Q2,Factor):
    """This function can be used to perform a spherical linear interpolation
       between this quaternion and a given other quaterion. The input quaternions
       have to be normalized! The result will be returned. If a real slerp is
       inadquate a nlerp will be used instead. A new quaternion for the result is
       returned."""
    # Implementing a slerp here would require some trigonometric functions, which
    # are not part of Blender.Mathutils. To avoid that a full Python installation
    # becomes a requirement due to import math this function is implemented
    # through the Slerp function of Blender.
    BlenderResult=Mathutils.Slerp(Mathutils.Quaternion(self.w,self.x,self.y,self.z),Mathutils.Quaternion(Q2.w,Q2.x,Q2.y,Q2.z),Factor);
    # Normalize the resulting quaternion
    Result=CQuaternion(BlenderResult.x,BlenderResult.y,BlenderResult.z,BlenderResult.w);
    Result.Normalize();
    return Result;
CQuaternion.Slerp=CQuaternionSlerp;


def CMatrix4x4Copy(self,RHS):
    """This function makes self become a deep-copy of Matrix where matrix has to
       be a CMatrix4x4 or a Blender matrix."""
    if(isinstance(RHS,CMatrix4x4)):
        self._11=RHS._11; self._12=RHS._12; self._13=RHS._13; self._14=RHS._14;
        self._21=RHS._21; self._22=RHS._22; self._23=RHS._23; self._24=RHS._24;
        self._31=RHS._31; self._32=RHS._32; self._33=RHS._33; self._34=RHS._34;
        self._41=RHS._41; self._42=RHS._42; self._43=RHS._43; self._44=RHS._44;
    else:
        if(RHS.colSize<4 or RHS.rowSize<4):
            raise TypeError("Failed to copy a Blender Matrix into a CMatrix4x4. The Blender matrix has a size of "+str(RHS.rowSize)+"x"+str(RHS.colSize)+". This is too small.");
        else:
            self._11=RHS[0][0]; self._12=RHS[0][1]; self._13=RHS[0][2]; self._14=RHS[0][3];
            self._21=RHS[1][0]; self._22=RHS[1][1]; self._23=RHS[1][2]; self._24=RHS[1][3];
            self._31=RHS[2][0]; self._32=RHS[2][1]; self._33=RHS[2][2]; self._34=RHS[2][3];
            self._41=RHS[3][0]; self._42=RHS[3][1]; self._43=RHS[3][2]; self._44=RHS[3][3];
CMatrix4x4.Copy=CMatrix4x4Copy;


class CMeshData(CMeshDataBase):
    """This class is used to provide access to mesh data of Blender 2.49 in a 
       generalized way."""

    def __init__(self,MeshObject,MaterialList,BoneList,UVLayerSet,ReadModifiers,RenderModeModifiers=False):
        """This constructor makes this object access the data of the given mesh 
           object (needed as Object). The given material and bone list are used 
           for finding indices and if UVLayerSet is not None all UV layers in 
           the given set will be part of the vertex data. If MeshObject is None 
           the mesh data is empty. If ReadModifiers is True the modified mesh 
           gets read. If RenderModeModifiers is True the modifiers will be used 
           with their render settings instead of the edit settings."""
        CMeshDataBase.__init__(self);
        # Save the passed data
        self.MeshObject=MeshObject;
        self.SourceList.append(self.MeshObject);
        self.MaterialList=MaterialList;
        self.BoneList=BoneList;
        self.UVLayerSet=UVLayerSet;
        self.ReadModifiers=ReadModifiers;
        self.RenderModeModifiers=RenderModeModifiers;

    def DoLock(self):
        """Implements behavior of LockMeshData() that is specific for derived 
           classes. It returns true, if and only if the operation was 
           successful."""
        # Retrieve the needed mesh data
        if(self.MeshObject==None):
            return False;
        self.ObjectData=None;
        if(self.MeshObject.getType()=="Mesh"):
            self.ObjectData=self.MeshObject.getData(False,True);
        else:
            self.ObjectData=self.MeshObject.getData(False);
        self.SkinningUsed=self.MeshObject.getType()=="Mesh" and self.BoneList!=None and len(self.BoneList)>0;
        self.MeshData=None;
        if(self.ReadModifiers):
            self.MeshData=Mesh.New("TemporaryExportMesh");
            # Ensure that armatures do not get applied to meshes
            if(self.SkinningUsed):
                # The armature can be set as parent or as modifier. Both needs 
                # to be regarded.
                ArmatureModifier=None;
                for ObjectModifier in self.MeshObject.modifiers:
                    if(ObjectModifier.type==Modifier.Types["ARMATURE"]):
                        ArmatureModifier=ObjectModifier;
                        break;
                if(ArmatureModifier!=None):
                    ArmatureModifierRender=ArmatureModifier[Modifier.Settings["RENDER"]];
                    if(ArmatureModifierRender!=False):
                        ArmatureModifier[Modifier.Settings["RENDER"]]=False;
                    ArmatureModifierRealtime=ArmatureModifier[Modifier.Settings["REALTIME"]];
                    if(ArmatureModifierRealtime!=False):
                        ArmatureModifier[Modifier.Settings["REALTIME"]]=False;
                else:
                    ParentArmature=self.MeshObject.parent;
                    self.MeshObject.clrParent(2);
                # Update the object
                self.MeshObject.makeDisplayList();
            # Get the mesh data with modifiers applied
            if(self.RenderModeModifiers):
                self.MeshData.getFromObject(self.MeshObject,0,1);
            else:
                self.MeshData.getFromObject(self.MeshObject,0,0);
            self.Name=self.MeshObject.getName();
            # If necessary restore the previous state
            if(self.SkinningUsed):
                if(ArmatureModifier!=None):
                    if(ArmatureModifierRender!=False):
                        ArmatureModifier[Modifier.Settings["RENDER"]]=ArmatureModifierRender;
                    if(ArmatureModifierRealtime!=False):
                        ArmatureModifier[Modifier.Settings["REALTIME"]]=ArmatureModifierRealtime;
                elif(ParentArmature!=None):
                    ParentArmature.makeParentDeform([self.MeshObject],0);
                # Update the object
                self.MeshObject.makeDisplayList();
        else:
            if(self.MeshObject.getType()!="Mesh"):
                return False;
            self.MeshData=self.MeshObject.getData(False,True);
            self.Name=self.MeshData.name;
        # Get the number of texture coordinate sets and a list of UV layers to export
        if(self.MeshData.vertexUV or self.MeshData.faceUV):
            self.nTexCoordSet=1;
        else:
            self.nTexCoordSet=0;
        if(self.UVLayerSet!=None):
            self.nTexCoordSet=len(self.UVLayerSet);
        elif(self.nTexCoordSet==1):
            self.UVLayerSet=frozenset([self.MeshData.activeUVLayer]);
        elif(self.nTexCoordSet==0):
            self.UVLayerSet=frozenset([]);
        # Determine whether the mesh is partially solid (in this case the
        # creation of the vertex list is more complicated)
        Solid=False;
        for Face in self.MeshData.faces:
            if(not Face.smooth):
                Solid=True;
                break;
        self.UseFaceVertices=((self.MeshData.vertexUV==False and self.MeshData.faceUV) or Solid);
        return True;
    
    def DoUnlock(self):
        """Implements behavior of UnlockMeshData() that is specific for derived 
           classes."""
        # Just mark this object as unlocked
        self.MeshData=None;
        self.ObjectData=None;

    def GetName(self):
        """This function returns the name of the associated mesh."""
        return self.Name;
    
    def UsesSkinning(self):
        """This function returns true, if and only if the associated mesh uses 
           skinning."""
        return self.SkinningUsed;
    
    def UsesFaceVertices(self):
        """This function returns true, if and only if it is necessary to assume 
           that a single vertex may have different values (normals and/or 
           texture coordinates) for different faces. If it returns true you 
           should use GetFaceVertex(...) instead of GetVertex(...)."""
        return self.UseFaceVertices;

    def GetTexCoordSetCount(self):
        """This function returns the number of texture coordinate sets that can 
           be accessed through this object."""
        return self.nTexCoordSet;

    def GetVertexCount(self):
        """This function returns the number of vertices in this mesh. It is 
           barely relevant if UsesFaceVertices() returns true."""
        return len(self.MeshData.verts);

    def GetVertex(self,iVertex):
        """This function returns a new CVertex holding the data of the vertex with 
           the given index in this mesh. Skinning weights, skinning indices and 
           texture coordinates are NOT included."""
        Vertex=self.MeshData.verts[iVertex];
        return CVertex(
                C3DVector(Vertex.co.x,Vertex.co.y,Vertex.co.z),
                C3DVector(Vertex.no.x,Vertex.no.y,Vertex.no.z));

    def GetFaceVertexIndex(self,iFace,iFaceVertex):
        """Given the indices of a face and one of the vertices of this face 
           this function returns the index (for GetVertex(...)) at which the 
           corresponding vertex may be found. If USesFaceVertices(...) returns 
           true the return value is irrelevant."""
        return self.MeshData.faces[iFace].verts[iFaceVertex].index;
    
    def GetFaceCount(self):
        """This function returns the number of faces in this mesh."""
        return len(self.MeshData.faces);

    def GetFaceVertexCount(self,iFace):
        """This function returns the number of vertices of the face with the 
           given index."""
        return len(self.MeshData.faces[iFace].verts);

    def GetFaceVertex(self,iFace,iFaceVertex):
        """This function returns a new CVertex holding the data of the vertex with 
           the given index of the face with the given index in this mesh. Skinning 
           weights, skinning indices and texture coordinates are NOT included."""
        Face=self.MeshData.faces[iFace];
        # Get the position
        Position=Face.verts[iFaceVertex].co;
        # Get the normal
        if(Face.smooth):
            Normal=Face.verts[iFaceVertex].no;
        else:
            Normal=Face.no;
        # Construct the result
        return CVertex(
                C3DVector(Position.x,Position.y,Position.z),
                C3DVector(Normal.x,Normal.y,Normal.z));
                
    def GetFaceMaterial(self,iFace):
        """This function returns the index of the material (in 
           self.MaterialList) that is to be used for the face with the given 
           index."""
        Face=self.MeshData.faces[iFace];
        #**#TODO: Regard colbits
        if(not isinstance(self.ObjectData,Blender.Types.Text3dType)
        and len(self.ObjectData.materials)>0):
            if(self.ObjectData.materials[Face.mat] in self.MaterialList):
                return self.MaterialList.index(self.ObjectData.materials[Face.mat]);
            else:
                self.MaterialList.append(self.ObjectData.materials[Face.mat]);
                return len(self.MaterialList)-1;
        else:
            return 0;
    
    def DoCollectTextureCoordinates(self,pVertex,GetVertexIndex):
        """Implements CollectTextureCoordinates(...). GetVertexIndex must not 
           be None."""
        # Get the texture coordinates
        PreviousUVLayer=self.MeshData.activeUVLayer;
        for UVLayer in self.UVLayerSet:
            self.MeshData.activeUVLayer=UVLayer;
            if(self.UsesFaceVertices()):
                for i in range(self.GetFaceCount()):
                    for j in range(self.GetFaceVertexCount(i)):
                        iVertexOut=GetVertexIndex(i,j,self);
                        # Get the texture coordinate or use a default value
                        if(iVertexOut>=0):
                            if(self.MeshData.vertexUV):
                                TexCoord=self.MeshData.faces[i].verts[j].uvco
                            elif(self.MeshData.faceUV):
                                TexCoord=self.MeshData.faces[i].uv[j];
                            else:
                                TexCoord=C2DVector(0.0,0.0);
                            pVertex[iVertexOut].pTexCoord.append(C2DVector(TexCoord.x,1.0-TexCoord.y));
            else:
                for i in range(self.GetVertexCount()):
                    iVertexOut=GetVertexIndex(i,self);
                    if(iVertexOut>=0):
                        if(self.MeshData.vertexUV):
                            pVertex[iVertexOut].pTexCoord.append(C2DVector(self.MeshData.verts[i].uvco.x,1.0-self.MeshData.verts[i].uvco.y));
                        else:
                            pVertex[iVertexOut].pTexCoord.append(C2DVector(0.0,0.0));
        if(PreviousUVLayer!=None):
            self.MeshData.activeUVLayer=PreviousUVLayer;
    
    def DoCollectWeights(self,pVertex,GetVertexIndex,MissingBoneSet=None):
        """Implements CollectWeights(...). GetVertexIndex must not be None. 
           MissingBoneSet may be a set that will be filled with the names of 
           bones, that are referenced, but were not found."""
        # Stop right here, if there are no weights
        if(not self.UsesSkinning()):
            return;
        # If modifiers have been applied MeshData is not linked to an object.
        # This makes retrieving weights impossible. For this reason MeshData has
        # to be linked to MeshObject temporarily.
        if(self.ReadModifiers):
            self.MeshObject.link(self.MeshData);
        # Retrieve the weight data
        if(self.UsesFaceVertices()):
            for i in range(self.GetFaceCount()):
                for j in range(self.GetFaceVertexCount(i)):
                    iVertexOut=GetVertexIndex(i,j,self);
                    if(iVertexOut>=0):
                        # Add the weights
                        pVertex[iVertexOut].ProcessInfluences(self.MeshData.getVertexInfluences(self.GetFaceVertexIndex(i,j)),self.BoneList,MissingBoneSet);
        else:
            for i in range(self.GetVertexCount()):
                iVertexOut=GetVertexIndex(i,self);
                if(iVertexOut>=0):
                    # Add the weights
                    pVertex[iVertexOut].ProcessInfluences(self.MeshData.getVertexInfluences(i),self.BoneList,MissingBoneSet);
        # Restore the link to the original mesh
        if(self.ReadModifiers):
            self.MeshObject.link(self.ObjectData);
    
    # This is the Blender mesh object whose data can be accessed through this 
    # object
    MeshObject=None;
    # This is a list of Blender materials in correct ordering. It is used to 
    # determine indices. 
    MaterialList=list();
    # This is a list of Blender bones in correct ordering. It is used to 
    # determine indices. 
    BoneList=list();
    # This is a set of all UV layers, which are to be regarded or None, if 
    # there is just a single UV layer.
    UVLayerSet=set();
    # Whether modifiers are to be regarded.
    ReadModifiers=False;
    # Whether the render settings are to be used for modifiers.
    RenderModeModifiers=False;
    
    # The following members get set when the mesh data is locked
    # The name of the associated mesh
    Name=str();
    # The Blender Mesh object, that holds the data
    MeshData=None;
    # The Blender Mesh Object that is to be associated with the Blender Object
    ObjectData=None;
    # Whether the associated mesh uses skinning
    SkinningUsed=False;
    # The number of texture coordinate sets, which are accessible through this 
    # object
    nTexCoordSet=0;
    # The value returned by UsesFaceVertices()
    UseFaceVertices=False;


# class CObject
def CObjectInit(self,Object,ParentObject=None,SkeletonForSkinning=True,DuplicationTransformation=CMatrix4x4()):
    """See export_utility_shared."""
    self.CleanUp();
    if(Object==None):
        return;
    self.SourceList.append(Object);
    self.Name=Object.getName();
    self.Parent=ParentObject;
    if(isinstance(self.Parent,CSkeleton)):
        self.ParentForSkinning=SkeletonForSkinning;
    # Save the transformation
    self.Transformation.Copy(Object.getMatrix());
    self.Transformation*=DuplicationTransformation;
    if(self.Parent!=None):
        InvParentTransform=self.Parent.GetTransformation().ComputeInverse();
        if(InvParentTransform!=None):
            self.Transformation*=InvParentTransform;
        else:
            print("WARNING: The object \""+str(ParentObject.Name)+"\" has a singular transformation (scaling zero). Since it is parent for \""+self.Name+"\" this is not allowed. Please change it.");
CObject.__init__=CObjectInit;


def CObjectGetCurrentTransformation(self):
    """This function returns the current object to parent transformation of the 
       associated Blender Object (first SourceList entry) of this object as 
       CMatrix4x4. On failure (e.g. if the SourceList is empty) it returns None."""
    # Get the Blender Object
    if(len(self.SourceList)==0 or not isinstance(self.SourceList[0],Blender.Types.ObjectType)):
        return None;
    Object=self.SourceList[0];
    # Return the casted transformation
    Result=CMatrix4x4();
    Result.Copy(Object.getMatrix("localspace"));
    return Result;
CObject.GetCurrentTransformation=CObjectGetCurrentTransformation;


def CObjectAddKeysFromActionCurves(self,ActionName,MaxKeyDistance=0.0):
    """This function just adds a keyed transformation with a key, which matches the 
       current transformation of this object. It may be implemented correctly 
       lateron, but since the Blender 2.49 exporter is about to become discontinued, 
       it is unlikely."""
    #TODO: complete
    self.GetKeyedTransformation(ActionName).AddTransformationKey(0.0,self.Transformation);
CObject.AddKeysFromActionCurves=CObjectAddKeysFromActionCurves;


# class CMeshObject
def CMeshObjectInit(self,MeshObject,ParentObject,MaterialList,BoneList,MeshList,UVLayerSet,ReadModifiers=False,RenderModeModifiers=True,DuplicationTransformation=CMatrix4x4()):
    """This function creates a CMeshObject, which matches the given Blender
       object. The given parent object is used. If the needed CMesh already
       exists in MeshList this version will be used. Otherwise a new CMesh is
       appended to the list. If modifiers are to be read a new CMesh with the 
       name of the mesh object is created anyway and modifiers will be regarded 
       during its creation. If RenderModeModifiers is True modifiers are used
       with their render settings. See CObject for a description of 
       DuplicationTransformation."""
    CObject.__init__(self,MeshObject,ParentObject,True,DuplicationTransformation);
    if(MeshObject==None):
        return;
    # Search for a CMesh with the right name, if modifiers should not be applied
    self.Mesh=None;
    if(not ReadModifiers):
        if(MeshObject.getType()!="Mesh"):
            return;
        BlenderMesh=MeshObject.getData();
        for Mesh in MeshList:
            if(Mesh.Name==BlenderMesh.name):
                self.Mesh=Mesh;
                break;
    # If none has been found create a new one
    if(self.Mesh==None):
        self.Mesh=CMesh(len(MeshList),CMeshData(MeshObject,MaterialList,BoneList,UVLayerSet,ReadModifiers,RenderModeModifiers));
        # Add this new mesh to the list unless it is invalid
        if(self.Mesh.IsValid()):
            MeshList.append(self.Mesh);
        else:
            self.Mesh=None;
CMeshObject.__init__=CMeshObjectInit;


def BlendModeFromBlender(BlenderBlendMode):
    """This function converts a Blender 2.49 blend mode integer to a 
       corresponding CBlendMode and returns the result."""
    for BlendMode in Blender.Texture.BlendModes.keys():
        if(BlenderBlendMode==Blender.Texture.BlendModes[BlendMode]):
            return CBlendMode(BlendMode);


class CMaterialData(CMaterialDataBase):
    """This is a base class that gives access to material data of Blender in a 
       way that is not specific to a particular Blender version."""
       
    def __init__(self,MaterialObject,UVLayerSet):
        """This function initializes this object thus that it makes the data of 
           the given material accessible and uses UV layer indices from the 
           given set."""
        CMaterialDataBase.__init__(self);
        self.MaterialObject=MaterialObject;
        self.SourceList.append(self.MaterialObject);
        self.UVLayerSet=UVLayerSet;
        # Create a list of all relevant textures
        self.piBlenderTexture=list();
        for iTexture in self.MaterialObject.enabledTextures:
            Texture=MaterialObject.getTextures()[iTexture];
            if(Texture==None):
                continue;
            Image=Texture.tex.getImage();
            if(Image==None):
                continue;
            if(not Texture.mapto&Blender.Texture.MapTo["COL"]):
                continue;
            self.piBlenderTexture.append(iTexture);
    
    def GetName(self):
        """This function retrieves the name of the associated material."""
        return self.MaterialObject.name;
    
    def GetColorInfo(self):
        """This function returns a 5-tuple of the form (AmbientColor, 
           DiffuseColor, EmissiveColor, SpecularColor, SpecularPower). This way 
           it gives the colors of this material. The colors are given as 
           CColor, SpecularPower is a float."""
        Data=self.MaterialObject
        return (
            CColor(Data.R*Data.amb,Data.G*Data.amb,Data.B*Data.amb,Data.alpha),
            CColor(Data.R*Data.ref,Data.G*Data.ref,Data.B*Data.ref,Data.alpha),
            CColor(Data.R*Data.emit,Data.G*Data.emit,Data.B*Data.emit,1.0),
            CColor(Data.specR*Data.spec,Data.specG*Data.spec,Data.specB*Data.spec,1.0),
            Data.getHardness());

    def GetTextureCount(self):
        """This function returns the number of textures that are used by this 
           material to effect the color."""
        return len(self.piBlenderTexture);
    
    def DoLockTexture(self,iTexture):
        """This function implements LockTexture. It returns True, if and only 
           if locking succeeded."""
        self.Texture=self.MaterialObject.getTextures()[self.piBlenderTexture[iTexture]];
        self.Image=self.Texture.tex.getImage();
        return True;

    def GetBlendMode(self):
        """This function returns the blend mode for the currently locked 
           texture as CBlendMode."""
        return BlendModeFromBlender(self.Texture.blendmode);
    
    def GetTexCoordSet(self):
        """This function returns the index of the texture coordinate set that 
           is to be used for the currently locked texture."""
        iTexCoordSet=0;
        if(self.UVLayerSet!=None):
            iUVLayer=0
            for UVLayer in self.UVLayerSet:
                if(self.Texture.uvlayer==UVLayer):
                    iTexCoordSet=iUVLayer;
                    break;
                iUVLayer+=1;
        return iTexCoordSet;

    def GetTextureFile(self):
        """This function returns the absolute path to the image file that is to 
           be used for the currently locked texture."""
        return Blender.sys.expandpath(self.Image.getFilename());

    def DoUnlockTexture(self):
        """This function implements UnlockTexture()."""
        self.Texture=None;
        self.Image=None;

    # The Blender material object that functions as data source
    MaterialObject=None;
    # A set holding all exported UV layers
    UVLayerSet=frozenset();
    # This is a list of all Blender texture indices for this material, which 
    # effect the material color
    piBlenderTexture=list();

    # The following members are only relevant, if a texture is locked.
    # The currently locked texture
    Texture=None;
    # The image for the currently locked texture
    Image=None;


class CPose:
    """This is a little utility class. It can be used to save the pose and
       transformation of an armature object and to restore it lateron. It also 
       stores the active action and whether NLA is enabled."""

    def __init__(self,ArmatureObject):
        """This constructor makes a new object, which describes the current pose of
           the given object."""
        self.Armature=ArmatureObject;
        # Save data about the use of animations
        self.Action=ArmatureObject.getAction();
        self.UseNLA=ArmatureObject.enableNLAOverride;
        # Save bone transformations
        self.pBoneTransformation=[];
        if(self.Armature!=None):
            Pose=self.Armature.getPose();
            for Bone in Pose.bones.values():
                self.pBoneTransformation.append(
                    (Bone.loc.copy(),Bone.quat.copy(),Bone.size.copy()));

    def Restore(self):
        """This function restores the pose described by this object."""
        # Restore animation data
        self.Armature.action=self.Action;
        self.Armature.enableNLAOverride=self.UseNLA;
        # Restore the pose
        Pose=self.Armature.getPose();
        iBone=0;
        for Bone in Pose.bones.values():
            Bone.loc=self.pBoneTransformation[iBone][0];
            Bone.quat=self.pBoneTransformation[iBone][1];
            Bone.size=self.pBoneTransformation[iBone][2];
            iBone+=1;
        Pose.update();

    # This is the armature Object for which this object describes a pose
    Armature=None;
    # This is the Action, which is to be associated with the armature object, when 
    # the pose is restored
    Action=None;
    # This boolean is true, if and only if the use of NLA should be enabled for the 
    # object, when the pose is restored
    UseNLA=False;
    # This list holds local delta transformations of all bones as triples in the form
    # (loc,quat,size). They describe the pose that is to be restored.
    pBoneTransformation=[];


class CArmatureData(CArmatureDataBase):
    """This class encapsulates Blender 2.49 armature objects and makes their 
       data accessible through a version independent interface."""
    
    def __init__(self,ArmatureObject):
        """This function initalizes this class with the data from the given Blender 
           Object."""
        if(ArmatureObject==None or not isinstance(ArmatureObject,Blender.Types.ObjectType)
        or ArmatureObject.getType()!="Armature"):
            self.ArmatureObject=None;
            self.Armature=None;
            return;
        # Save the passed data
        self.ArmatureObject=ArmatureObject;
        self.Armature=self.ArmatureObject.getData();
        # Initialize the base class
        CArmatureDataBase.__init__(self,ArmatureObject.getName());
        self.SourceList.append(self.ArmatureObject);
        
    def GetArmatureTransformation(self):
        return self.ArmatureObject.getMatrix();
    
    def BoneExists(self,BoneName):
        return (BoneName in self.Armature.bones);
    
    def GetBoneNameList(self):
        return self.Armature.bones.keys();

    def GetBone(self,BoneName):
        """This function returns the Blender Bone object, which corresponds to the 
           given name or None if there is no such bone in this armature."""
        if(self.BoneExists(BoneName)):
            return self.Armature.bones[BoneName];
        return None;
    
    def GetBoneParent(self,BoneName):
        Bone=self.GetBone(BoneName);
        if(Bone!=None and Bone.parent!=None):
            return Bone.parent.name;
        return None;

    def GetBoneTransformationInherit(self,BoneName):
        Bone=self.GetBone(BoneName);
        if(Bone!=None):
            UseHinge=(Armature.HINGE in Bone.options);
            return (not UseHinge,not UseHinge,False,False);
        return (None,None,None,None);
    
    def GetBoneSource(self,BoneName):
        return self.GetBone(BoneName);
    
    def GetBoneRestTransformation(self,BoneName):
        Bone=self.GetBone(BoneName);
        if(Bone!=None):
            return Bone.matrix["ARMATURESPACE"];
        return None;

    def DoLockBoneAction(self,ActionName,BoneName):
        Bone=self.GetBone(BoneName);
        Action=Blender.NLA.GetActions()[ActionName];
        if(Bone==None or Action==None):
            return False;
        # Get the IPO curves
        try:
            AnimationIpo=Action.getChannelIpo(Bone.name);
        except:
            return False;
        self.pBoneTranslation=list([None,None,None]);
        self.pBoneRotation=list([None,None,None,None]);
        self.pBoneScaling=list([None,None,None]);
        pTranslationName=["PO_LOCX","PO_LOCY","PO_LOCZ","LocX","LocY","LocZ"];
        pRotationName=["PO_QUATW","PO_QUATX","PO_QUATY","PO_QUATZ","QuatW","QuatX","QuatY","QuatZ"];
        pScalingName=["PO_SIZEX","PO_SIZEY","PO_SIZEZ","SizeX","SizeY","SizeZ"];
        for Curve in AnimationIpo:
            for i,CurveName in enumerate(pTranslationName):
                if(Curve.name==CurveName):
                    self.pBoneTranslation[i%3]=Curve;
            for i,CurveName in enumerate(pRotationName):
                if(Curve.name==CurveName):
                    self.pBoneRotation[i%4]=Curve;
            for i,CurveName in enumerate(pScalingName):
                if(Curve.name==CurveName):
                    self.pBoneScaling[i%3]=Curve;
        return True;

    def DoUnlockBoneAction(self):
        # Just unset the curve lists
        self.pBoneTranslation=None;
        self.pBoneRotation=None;
        self.pBoneScaling=None;
    
    def GetBoneActionTranslation(self,Frame):
        Result=None;
        for i,Curve in enumerate(self.pBoneTranslation):
            if(Curve!=None):
                if(Result==None):
                    Result=C3DVector(0.0,0.0,0.0);
                Result[i]=Curve[Frame];
        return Result;
        
    def GetBoneActionRotation(self,Frame):
        Result=None;
        for i,Curve in enumerate(self.pBoneRotation):
            if(Curve!=None):
                if(Result==None):
                    Result=CQuaternion(0.0,0.0,0.0,1.0);
                Result[i]=Curve[Frame];
        return Result;
        
    def GetBoneActionScaling(self,Frame):
        Result=None;
        for i,Curve in enumerate(self.pBoneScaling):
            if(Curve!=None):
                if(Result==None):
                    Result=C3DVector(0.0,0.0,0.0);
                Result[i]=Curve[Frame];
        return Result;
        
    def GetBonePoseTransformation(self,BoneName):
        if(self.BoneExists(BoneName)):
            return self.ArmatureObject.getPose().bones[BoneName].poseMatrix;
        return None;

    def IsValid(self):
        return self.ArmatureObject!=None and self.Armature!=None;

    # The Blender Object that holds the armature for this object
    ArmatureObject=None;
    # The Blender armature that holds the actual armature data for this object
    Armature=None;
    
    # The following data is only relevant, if the armature data is currently 
    # locked through LockBoneAction(...)
    
    # This list holds IPO curves for the X, Y and Z components of the bone 
    # translation
    pBoneTranslation=list([None,None,None]);
    # This list holds IPO curves for the W, X, Y and Z components of the 
    # quaternion for the bone rotation
    pBoneRotation=list([None,None,None,None]);
    # This list holds IPO curves for the X, Y and Z components of the bone 
    # scaling
    pBoneScaling=list([None,None,None]);


class CExportObject(CExportObjectBase):
    """This is a little helper class, which is used to resolve groups and find 
       parents in Blender 2.49. Temporary instances of it get created in 
       CModel.__init__(...)."""

    def __init__(self,BlenderObject,ObjectOnly=False):
        """This constructor initializes this object with the given values. 
           Blender data is retrieved automatically, most other values are set 
           to None to indicate that they are not yet initialized."""
        CExportObjectBase.__init__(self);
        # Save the input data
        self.BlenderObject=BlenderObject;
        self.ObjectOnly=ObjectOnly;
        # Get the Blender parent
        self.BlenderParent=BlenderObject.getParent();
        self.BlenderParentBone=None;
        ParentBoneName=BlenderObject.getParentBoneName();
        if(ParentBoneName!=None and self.BlenderParent.getType()=="Armature"):
            self.BlenderParentBone=self.BlenderParent.getData().bones[ParentBoneName];
        # If there is an armature modifier the armature is used as parent
        for ObjectModifier in BlenderObject.modifiers:
            if(ObjectModifier.type==Modifier.Types["ARMATURE"]):
                Armature=ObjectModifier[Modifier.Settings["OBJECT"]];
                if(Armature!=None):
                    if(Armature!=self.BlenderParent):
                        if(self.BlenderParent!=None):
                            print("WARNING: The object \""+BlenderObject.getName()+"\" uses the armature \""+Armature.getName()+"\", but has a different object set up as parent. The exporter will regard the armature as parent anyway.");
                        # This message is just buggy and unnecessary most of 
                        # the time
                        #else:
                        #    print("WARNING: The object \""+BlenderObject.getName()+"\" uses the armature \""+Armature.getName()+"\", but does not have it set up as parent. The exporter will regard the armature as parent anyway.");
                        if(ObjectModifier[Modifier.Settings["ENVELOPES"]]==True):
                            print("The object \""+BlenderObject.getName()+"\" has envelopes enabled in its armature modifier. It is recommended to disable them because they do not get exported.");
                        # Set up the new parent
                        self.BlenderParent=Armature;
                        self.BlenderParentBone=None;
                        break;

    def GetObjectName(self):
        """This function returns the name of the associated Blender object."""
        return self.BlenderObject.getName();

    def GetParentBoneName(self):
        """This function returns the name of the parent bone of this export 
           object, or None, if it does not have a parent bone."""
        if(self.BlenderParentBone!=None):
            return self.BlenderParentBone.name;
        return None;

    def UsesSkinning(self):
        """This function returns true, iff this export object uses skinning."""
        return (self.ObjectOnly==False and self.BlenderParent!=None 
                and self.BlenderParentBone==None and self.BlenderParent.getType()=="Armature");


class CModelData(CModelDataBase):
    """This class provides access to the data of a model in Blender 2.49."""
    
    def __init__(self,ModelOptions,BlenderUtility):
        """This constructor initializes this model data object thus that it 
           gives access to the current scene using the given options."""
        CModelDataBase.__init__(self,ModelOptions,BlenderUtility);
        # Check if the model data is accessible
        if(Window.EditMode() or Window.PoseMode()):
            self.BlenderUtility.ShowMessage(
                "ERROR: Blender is currently not in object mode. Exporting is only possible from object mode.",
                "Error","Not in object mode.");
            self.Valid=False;
        print("Reading data for export.");
        # Create an undo state, if risky settings are enabled
        if(self.Options.ExportConstraints or self.Options.ReadModifiers):
            Blender.SaveUndoState("Export");
            print("Created an undo state to be certain that the previous state can be recovered.");
        # Determine the list of supported types
        self.SupportedTypeList=["Mesh","Empty"];
        if(self.Options.ExportAnimations):
            self.SupportedTypeList.append("Armature");
        if(self.Options.ReadModifiers):
            self.SupportedTypeList+=["Text","Surf","Curve","MBall"];
        # Initialize the data structures for StoreState()
        self.PreviousFrame=0;
        self.PoseDict=dict();
        self.TransformationDict=dict();
        # Everything was successful
        self.Valid=True;

    def GetExportObjectList(self):
        """This function returns a list of export objects (CExportObject), 
           which are to be read for the associated model."""
        # Get the relevant scene
        SceneObject=Scene.GetCurrent();
        # Get a list of all objects in the scene
        AllObjects=SceneObject.objects;
        # Create a list of the objects, which may need to be exported (if their
        # type is right). Use CExportObject for this purpose.
        RelevantObjects=[];
        for Obj in AllObjects:
            ExportState=None;
            if(self.Options.ExportObjectCallback!=None):
                ExportState=self.Options.ExportObjectCallback(Obj);
            if(Obj.getType() in self.SupportedTypeList and ExportState!=False):
                if(self.Options.ExportSelectionOnly):
                    if(Obj.isSelected() or ExportState==True):
                        RelevantObjects.append(CExportObject(Obj,Obj.getType()=="Empty"));
                else:
                    RelevantObjects.append(CExportObject(Obj,Obj.getType()=="Empty"));
        return RelevantObjects;
    
    def ExportObjectToObject(self,ExportObject,ObjectParent,BlenderMaterialList,MeshList,UVLayerSet,DuplicationRootDict):
        """See base."""
        BlenderObject=ExportObject.BlenderObject;
        # Get the duplication transformation
        DuplicationTransformation=CMatrix4x4();
        if(ExportObject.DuplicationID in DuplicationRootDict):
            DuplicationTransformation.Copy(DuplicationRootDict[ExportObject.DuplicationID].BlenderObject.getMatrix());
        # Create either a CObject,
        if(ExportObject.ObjectOnly
        or not BlenderObject.getType() in self.SupportedTypeList
        or BlenderObject.getType()=="EMPTY"):
            Object=CObject(BlenderObject,ObjectParent,False,DuplicationTransformation);
        # a CSkeleton
        elif(BlenderObject.getType()=="Armature"):
            Object=CSkeleton(self,CArmatureData(BlenderObject),ObjectParent,self.Options.ExportAnimations and not self.Options.ExportConstraints,not self.Options.ExportCurrentNLA,self.Options.MaxKeyDistance,DuplicationTransformation);
        # or a CMeshObject (curves, meta balls, etc. become meshes as well)
        else:
            BoneList=None;
            if(ObjectParent!=None and ExportObject.BlenderParentBone==None):
                if(isinstance(ObjectParent,CSkeleton)):
                    BoneList=ObjectParent.BoneList;
            Object=CMeshObject(BlenderObject,ObjectParent,BlenderMaterialList,BoneList,MeshList,UVLayerSet,self.Options.ReadModifiers,self.Options.RenderModeModifiers,DuplicationTransformation);
            # Empty meshes are converted to pure objects
            if(Object.Mesh==None or not Object.Mesh.IsValid()):
                Object=Object.CloneObject();
        return Object;

    def GetUVLayerSet(self,RelevantObjects):
        """If the set of UV texture names is not the same for all meshes that 
           are associated with the export objects in the given list, this 
           function returns None. Otherwise it returns the unique set of UV 
           layer names."""
        BlenderMeshList=[];
        for ExportObject in RelevantObjects:
            Object=ExportObject.BlenderObject;
            if(Object.getType()=="Mesh"):
                BlenderMeshList.append(Object);
        if(len(BlenderMeshList)==0):
            return frozenset();
        UVLayerSet=frozenset(BlenderMeshList[0].getData(False,True).getUVLayerNames());
        for MeshObject in BlenderMeshList:
            if(frozenset(MeshObject.getData(False,True).getUVLayerNames())!=UVLayerSet):
                print("Failed to determine a unique set of UV layers. The UV layers of the mesh object \""+BlenderMeshList[0].getName()+"\" and of \""+MeshObject.getName()+" are not identical.");
                return None;
        return UVLayerSet;
    
    def GetMaterialData(self,BlenderMaterial,UVLayerSet):
        """Given a Blender material object and a set of UV layer names (or 
           None) this function returns a corresponding CMaterialData 
           instance."""
        return CMaterialData(BlenderMaterial,UVLayerSet)
        
    def CreateExportObject(self,BlenderObject,PureObject,DuplicationID):
        """See base."""
        Result=CExportObject(BlenderObject,PureObject);
        Result.DuplicationID=DuplicationID;
        return Result;

    def DoGetNLAFrameList(self):
        """This function implements GetNLAFrameList(...). It does not need to take 
           care of MaxKeyDistance, sorting or a DefaultFrame."""
        RenderContext=Blender.Scene.GetCurrent().getRenderingContext();
        FirstFrame=float(RenderContext.sFrame);
        LastFrame=float(RenderContext.eFrame);
        return CompleteFrameList([FirstFrame,LastFrame],5.0,1.0);
    
    def GetActionNameList(self):
        """This function returns a list of strings, giving the names of all actions, 
           which are accessible through this model data."""
        return Armature.NLA.GetActions().keys();

    def GetActionSource(self,ActionName):
        """This function returns the Blender object, which represents the action 
           with the given name. It is meant to be saved in a source list."""
        return Armature.NLA.GetActions()[ActionName];

    def GetActionRange(self,ActionName):
        """This function returns the frame at which the action with the given name 
           begins and the frame at which it ends (as pair of either int or float)."""
        Action=Armature.NLA.GetActions()[ActionName];
        if(Action!=None):
            FrameNumbers=Action.getFrameNumbers();
            return (min(FrameNumbers),max(FrameNumbers)+1);
        return (None,None);
    
    def StoreState(self):
        """This function has to be called before calling either ApplyAction(...) or 
           ApplyFrame(...) to ensure that RestoreState(...) works properly."""
        # Save the current frame
        self.PreviousFrame=Scene.getCurrent().getRenderingContext().currentFrame();
        # Iterate over all objects and store poses and transformations
        for Object in Blender.Object.Get():
            # Save the transformation
            self.TransformationDict[Object.getName()]=(Object.getLocation(),Object.getEuler(),Object.getSize());
            # Save the pose for armatures
            if(Object.getType()=="Armature"):
                self.PoseDict[Object.getName()]=CPose(Object);

    def DoGetActionFrameList(self,ActionName):
        """This function returns all frames, which have to be keyed for the action 
           with the given name as list of floats."""
        Action=Armature.NLA.GetActions()[ActionName];
        if(Action==None):
            return list();
        Result=list();
        for Frame in Action.getFrameNumbers():
            Result.append(float(Frame));
        return Result;

    def ApplyAction(self,ActionName):
        """This function applies the action with the given name to all Blender 
           objects in this model and it resets the poses of all armatures. After 
           calling it you can play back the action using ApplyFrame(...) to record 
           it. Use RestoreState() to undo the changes caused by this function."""
        Action=Armature.NLA.GetActions()[ActionName];
        if(Action==None):
            return;
        for Object in Blender.Object.Get():
            # Check, whether the object is an armature
            Pose=Object.getPose();
            if(Object.getType()=="Armature" and Pose!=None):
                # Apply the action
                Object.enableNLAOverride=False;
                Object.action=Action;
                # Reset the pose
                for Bone in Pose.bones.values():
                    Bone.loc=Mathutils.Vector(0.0,0.0,0.0);
                    Bone.quat=Mathutils.Quaternion(1.0,0.0,0.0,0.0);
                    Bone.size=Mathutils.Vector(1.0,1.0,1.0);
                Pose.update();
    
    def ApplyFrame(self,Frame):
        """This function sets up the given frame for the scene, whose data is 
           accessed by this model. Use RestoreState() to undo the changes caused by 
           this function. Frame should be a floating point value."""
        Scene.GetCurrent().getRenderingContext().currentFrame(int(Frame));
    
    def RestoreState(self):
        """This function reverts the changes caused by either ApplyAction(...) or 
           ApplyFrame(...). It can only work correctly, if StoreState(...) has been 
           called initially."""
        # Restore the frame value
        Scene.getCurrent().getRenderingContext().currentFrame(self.PreviousFrame);
        # Iterate over all objects and restore poses and transformations
        for Object in Blender.Object.Get():
            # Restore the pose
            if((Object.getName() in self.PoseDict) and Object.getType()=="Armature"):
                self.PoseDict[Object.getName()].Restore();
            # Restore the transformation
            if(Object.getName() in self.TransformationDict):
                Transformation=self.TransformationDict[Object.getName()];
                Object.setLocation(Transformation[0][0],Transformation[0][1],Transformation[0][2]);
                Object.setEuler(Transformation[1]);
                Object.setSize(Transformation[2][0],Transformation[2][1],Transformation[2][2]);
        
    def GetArmatureData(self,Skeleton):
        """Given a CExportObjectBase that refers to a skeleton this function 
           returns a corresponding instance of a derived class of 
           CArmatureDataBase. On failure it returns None."""
        return CArmatureData(Skeleton.BlenderObject);
        
    def IsValid(self):
        """This function returns true, if and only if this model data object is 
           ready to be used for reading data."""
        return self.Valid;

    # This boolean is true, if and only if this object represents valid model 
    # data
    Valid=False;
    # A list holding the names of all Blender object types, which can be read 
    # with this model data
    SupportedTypeList=list();
    
    # The frame value, which has been set for the scene at the last call to 
    # StoreState()
    PreviousFrame=0;
    # This dictionary maps Blender Object names of objects for armatures to 
    # corresponding instances of CPose, which have been generated by StoreState(). 
    # RestoreState() uses it.
    PoseDict=dict();
    # This dictionary maps object names to object transformations, which have been 
    # saved by StoreState(). RestoreState() uses it. Transformations are saved 
    # relative to the parent using triples of Blender objects in the form 
    # (Location,Rotation,Scaling).
    TransformationDict=dict();
    

# class CModel
def CModelResolveGroups(self,RelevantObjects,DuplicationRootDict):
    """This function iterates through the given list of CExportObjects and checks
       whether it contains objects, which are set to duplicate a group. If so it 
       adds duplicates of the objects in the group as new CExportObjects to the 
       list. It assigns duplication IDs and writes the mapping from duplication IDs 
       to CExportObjects for the root of the duplication to the given dictionary. 
       Besides it deletes empties, which do not duplicate a group. It returns true, 
       iff objects were duplicated."""
    # Resolve groups
    GroupResolved=False;
    # Iterate over all objects, which duplicate a group. Note that this iteration 
    # also passes duplicated objects and thus recursive duplication is supported.
    for ExportObject in RelevantObjects:
        BlenderObject=ExportObject.BlenderObject;
        if(BlenderObject.getType()!="Empty"):
            continue;
        try:
            FoundGroup=Group.Get(BlenderObject.getName());
        except:
            FoundGroup=None;
        if(FoundGroup==None):
            continue;
        # Determine an ID for this duplication
        DuplicationID=ExportObject.DuplicationID;
        if(DuplicationID==""):
            DuplicationID=BlenderObject.name;
        else:
            DuplicationID=DuplicationID+"_$_"+BlenderObject.name;
        DuplicationRootDict[DuplicationID]=ExportObject;
        # Determine a transformation for this duplication
        DuplicationTransformation=CMatrix4x4();
        DuplicationTransformation.Copy(BlenderObject.getMatrix());
        DuplicationTransformation*=ExportObject.DuplicationTransformation;
        # Duplicate the objects contained in the group
        for SubObject in FoundGroup.objects:
            NewExportObject=CExportObject(SubObject);
            NewExportObject.DuplicationTransformation=DuplicationTransformation;
            NewExportObject.DuplicationID=DuplicationID;
            RelevantObjects.append(NewExportObject);
            # Mark the object as resolved and remember that a resolution has 
            # happened
            GroupResolved=True;
            ExportObject.GroupResolved=True;
    # Delete empties, which did not belong to groups
    i=0;
    while(i<len(RelevantObjects)):
        ExportObject=RelevantObjects[i];
        if(ExportObject.BlenderObject.getType()=="Empty"):
            if(not ExportObject.GroupResolved):
                RelevantObjects.remove(ExportObject);
                i-=1;
        i+=1;
    # Return whether a group has been resolved
    return GroupResolved;
CModel.ResolveGroups=CModelResolveGroups;


class CBlenderUtility(CBlenderUtilityBase):
    """This is class implements CBlenderUtilityBase functionality for Blender 
       2.49."""
    
    def __init__(self):
        """This constructor initializes the class with default values."""
        CBlenderUtilityBase.__init__(self);
            
    def PrepareExport(self,Backup=True):
        """See base."""
        pass;
    
    def FinalizeExport(self):
        """See base."""
        pass;

    def ShowMessage(self,LongText,Title=None,ShortText=None):
        """This function prints LongText to the console. If ShortText is not 
           None it also shows ShortText in a pop up in Blender's UI. If it is 
           None it will show LongText in a pop up. Title gets used as title for 
           the pop up (if given)."""
        print(LongText);
        if(ShortText==None):
            ShortText=LongText;
        if(Title!=None):
            ShortText=Title+"%t|"+ShortText;
        Draw.PupMenu(ShortText);
        
    def SetProgress(self,Progress,State):
        """This function updates the progres bar. Progress should be a floating 
           point value between 0.0 and 1.0 indicating the current progress. 
           State should be a very short string (around 10 letters) describing 
           the current task."""
        Window.DrawProgressBar(Progress,State);


# This method is defined in both Blender version specific modules.    
def CPropertyGetValue(self,BlenderObject):
    """This function returns the value, which is currently set for this 
       property in the given Blender object or the default value, if None is 
       set. In all cases it returns an object of the type implied by GetType(). 
       For enumerations it returns the entry ID."""
    if(BlenderObject==None):
        return self._Default;
    for Property in BlenderObject.getAllProperties():
        if(Property.getName()==self._ID):
            PropertyType=Property.getType();
            if((PropertyType=="FLOAT" and self._Type==float) \
            or (PropertyType=="INT" and self._Type==int) \
            or (PropertyType=="STRING" and not self._Type in [int,float,str,bool])):
                Result=Property.getData();
                if(not self._Type in [int,float,str,bool]):
                    if(Result in self._Type.GetEntryList()):
                        return Result;
                    else:
                        return self._Default;
                else:
                    return Result;
            if(PropertyType=="BOOL" and self._Type==bool):
                return bool(Property.getData());
    return self._Default;
CPropertyGetValue=CProperty.GetValue;

# export_utility END

# struct is needed to bring data into a binary format
from struct import pack
# A bit of trigonometry is needed for the normals
from math import asin,atan2

def PackString(String):
    """This utility function converts a string to a null-terminated sequence of 
       ASCII bytes."""
    return pack(str(len(String)+1)+"s",String.encode("ascii","replace"));

class CChunkSizeWriter:
    """This class rembers the offset at which a chunk size should be written
    and writes the chunk size in this place on request. It is used to export
    fileformat version 2.0."""
    def __init__(self,File):
        """This function remembers the current file pointer of the given file
        and moves it 4=sizeof(DWORD) bytes forward by writing zeroes. Later 
        Write() can write a chunk size at this offset."""
        self.oChunkSize=File.tell();
        self.File=File;
        File.write(pack("L",0));
    
    def Write(self):
        """This function writes a DWORD chunk size at the recently marked
        position."""
        oFilePointer=self.File.tell();
        self.File.seek(self.oChunkSize);
        self.File.write(pack("L",oFilePointer-self.oChunkSize-4));
        self.File.seek(oFilePointer);
       
    oChunkSize=0;
    File=None;


def C2DVectorSaveData(self,File):
    """Saves the data of this vector to the given file"""
    File.write(pack("ff",self.x,self.y));
C2DVector.SaveData=C2DVectorSaveData;


def C3DVectorSaveData(self,File):
    """Saves the data of this vector to the given file"""
    File.write(pack("fff",self.x,self.y,self.z));
C3DVector.SaveData=C3DVectorSaveData;


def C3DVectorSaveAsCompressedNormal(self,File,UseV2):
    """Saves a normalized copy of this vector as 16-bit or 32-bit 
       latitude/longitude vector."""
    self.Normalize();
    Latitude=-asin(self.y);
    Longitude=atan2(self.x,self.z);
    if(UseV2):
        Latitude*=20860.12008;
        Longitude*=10430.06004;
        File.write(pack("hh",int(Latitude),int(Longitude)));
    else:
        Latitude*=20.212678;
        Longitude*=40.4253555;
        File.write(pack("bb",int(Latitude),int(Longitude)));
C3DVector.SaveAsCompressedNormal=C3DVectorSaveAsCompressedNormal;


def CMatrix4x4SaveData(self,File):
    """This function saves this matrix to a given binary file."""
    File.write(pack("16f",self._11,self._12,self._13,self._14,self._21,self._22,self._23,self._24,self._31,self._32,self._33,self._34,self._41,self._42,self._43,self._44));
CMatrix4x4.SaveData=CMatrix4x4SaveData;


def CQuaternionSaveData(self,File):
    """This function saves the data of this quaternion to the given file."""
    File.write(pack("4f",self.x,self.y,self.z,self.w));
CQuaternion.SaveData=CQuaternionSaveData;


def CTriangleIndicesSaveData(self,File):
    """Saves the data of this triangle index set to the given file"""
    File.write(pack("HHH",self.A,self.B,self.C));
CTriangleIndices.SaveData=CTriangleIndicesSaveData;


def CColorSaveData(self,File):
    File.write(pack("4f",self.r,self.g,self.b,self.a));
CColor.SaveData=CColorSaveData;


def CModelSaveData(self,File,UseV2):
    """This function writes header data of this model to the given file object"""
    if(UseV2):
        # Write the chunk header for the file header
        File.write(PackString("$U3D_FILE_HEADER"));
        # Remember the offset for the chunk size
        ChunkSize=CChunkSizeWriter(File);
        # Write the file format version
        File.write(pack("LLL",2,0,0));
        # Write the encryption version and the compression version
        File.write(pack("LL",0,0));
        # The chunk is complete
        ChunkSize.Write();

        # Write the chunk header for the model header
        File.write(PackString("$U3D_MODEL_HEADER"));
        ChunkSize=CChunkSizeWriter(File);
        # Write the mesh counts
        MinFrame,MaxFrame=self.GetAnimationRange();
        nFrame=int(1.0+MaxFrame-MinFrame);
        File.write(pack("LLLL",self.GetMeshCount(),self.GetMeshCount(),nFrame,1));
        # Write the material count
        File.write(pack("L",self.GetMaterialCount()));
        # Write the bone count
        File.write(pack("L",self.GetBoneCount()));
        # Vertex tweening is not used
        File.write(pack("b",0));
        # Write the maximal distance of the one and only LoD
        File.write(pack("f",3.402823466e+38));
        # Write the texture coordinate dimensions
        for i in range(max(1,self.GetTexCoordSetCount())):
            File.write(pack("L",2));
        for i in range(max(1,self.GetTexCoordSetCount()),8):
            File.write(pack("L",0));
        # Write the skin weight count
        File.write(pack("L",Choose(self.GetBoneCount()>0,3,0)));
        # Write that there is no shader pack template
        File.write(pack("b",0));
        # The chunk is complete
        ChunkSize.Write();
        
    else:
        # Write the null-terminated file format identification string
        File.write(PackString("U3DModelFile"));
        # Write the file version
        File.write(pack("f",1.0));
        #Write the mesh count (twice)
        File.write(pack("HH",self.GetMeshCount(),self.GetMeshCount()));
        # Write the frame count
        MinFrame,MaxFrame=self.GetAnimationRange();
        nFrame=int(1.0+MaxFrame-MinFrame);
        File.write(pack("L",nFrame));
        # Write the LoD count
        File.write(pack("H",1));
        # Write the material count
        File.write(pack("H",self.GetMaterialCount()));
        # Write the bone count
        File.write(pack("H",self.GetBoneCount()));
        # Write the cel-shading line strength and the cel-shading color
        File.write(pack("5f",0.0,0.0,0.0,0.0,0.0));
        # Write whether this model uses alpha blending
        File.write(pack("b",1));
        # Write whether this model uses vertex tweening
        File.write(pack("b",0));
CModel.SaveData=CModelSaveData;


def CMeshSaveData(self,File,UseV2,BlenderUtility):
    """This function writes the collected data to the given file object"""
    # Check whether the model can be written
    if(self.nVertex>65535 or self.nTriangle>65535):
        BlenderUtility.ShowMessage(
          "Error: The model can not be written to an Ultimate 3D model file, because one of the meshes has more than 65535 triangles or vertices.",
          "Error","Too many triangles/vertices in mesh \""+self.Name+"\" ("+str(self.nTriangle)+"/"+str(self.nVertex)+").");
        return False;
    if(UseV2):
        # Write the chunk header
        File.write(PackString("$U3D_MESH"));
        ChunkSize=CChunkSizeWriter(File);
        # Write the mesh ID values
        File.write(pack("LLL",self.iMesh,0,0));
    else:
        # Write the null-terminated chunk identifier        
        File.write(PackString("$U3D_MESH"));
    # Write the null-terminated mesh name
    File.write(PackString(self.Name));
    if(UseV2==False):
        # Write the frame index for the mesh (0)
        File.write(pack("H",0));
        # Write the lod of this mesh (1.0)
        File.write(pack("f",1.0));
    # Write the normal scalar
    NormalScalar=Choose(self.SkinningUsed,-1.0,1.0);
    File.write(pack("f",NormalScalar));
    # Write whether this mesh contains valid inverse tangent space matrices
    File.write(pack("b",0));
    # Write the vertex count
    File.write(pack(Choose(UseV2,"L","H"),self.nVertex));
    if(UseV2==False):
        # Write the texture coordinate dimension array
        for i in range(self.nTexCoordSet):
            File.write(pack("H",2));
        for i in range(self.nTexCoordSet,8):
            File.write(pack("H",0));
        # Write the number of skin weights per vertex
        File.write(pack("H",3*self.SkinningUsed));
        # Write the maximum index of the bones that are influencing this mesh
        File.write(pack("B",self.iMaxBone));
        # Write the vertex and triangle count of shadow optimized geometry
        File.write(pack("HH",0,0));
    # Write the vertex positions
    for Vertex in self.pVertex:
        Vertex.Position.SaveData(File);
    # Write the vertex normals
    if(UseV2):
        for Vertex in self.pVertex:
            # In version 2.0 the normal scalar is applied to the normals
            # in the file after loading. Since the normals are correct now
            # they need to be identical after loading. Scale them by the
            # normal scalar now (1*1=(-1)*(-1)=1).
            ScaledNormal=Vertex.Normal;
            ScaledNormal.x*=NormalScalar;
            ScaledNormal.y*=NormalScalar;
            ScaledNormal.z*=NormalScalar;
            ScaledNormal.SaveAsCompressedNormal(File,UseV2);
    else:
        for Vertex in self.pVertex:
            Vertex.Normal.SaveAsCompressedNormal(File,UseV2);
    if(UseV2==False):
        # Whether a texture coordinate buffer is to be used and whether the
        # buffer is owned
        File.write(pack("b",0));
        File.write(pack("b",1));
    # Write the vertex texture coordinates
    for i in range(self.nTexCoordSet):
        for Vertex in self.pVertex:
            Vertex.pTexCoord[i].SaveData(File);
    if(UseV2 and self.nTexCoordSet==0):
        NullTexCoord=C2DVector(0.0,0.0);
        for i in range(self.nVertex):
            NullTexCoord.SaveData(File);
    # Export skinning data, if necessary
    if(self.SkinningUsed):
        # Write the skin weights
        for Vertex in self.pVertex:
            File.write(pack("3f",Vertex.pSkinWeight[0],Vertex.pSkinWeight[1],Vertex.pSkinWeight[2]));
        # Write the bone indices for each vertex
        for Vertex in self.pVertex:
            File.write(pack("4B",Vertex.piSkinBone[0],Vertex.piSkinBone[1],Vertex.piSkinBone[2],Vertex.piSkinBone[3]));
    # Write the triangle count
    File.write(pack(Choose(UseV2,"L","H"),self.nTriangle));
    # Whether the triangle indices are owned by this mesh (true)
    File.write(pack("b",1));
    # Write the index buffer
    for Triangle in self.pTriangle:
        Triangle.SaveData(File);
    # Write the triangle material indices
    for i in range(self.nTriangle):
        File.write(pack("H",self.piTriangleMaterial[i]));
    if(UseV2):
        # Whether there is shadow optimized geometry (false)
        File.write(pack("b",0));
        # The chunk is completed
        ChunkSize.Write();
    # The function has succeeded
    return True;
CMesh.SaveData=CMeshSaveData;


# class CBlendMode
def CBlendModeToD3DTexOp(self):
    """This is a little utility function, which converts a given Blender texture
       operation (resp. blend mode) to the D3DTEXTUREOP DWORD value, which matches it
       best."""
    if(self.IsValid()):
        TexOpDict=dict({
            "MIX":13,"ADD":7,"SUBTRACT":10,"MULTIPLY":4,
            "SCREEN":4,"OVERLAY":8,"DIFFERENCE":10,"DIVIDE":4,
            "DARKEN":4,"LIGHTEN":4,
            "HUE":4,"SATURATION":4,"VALUE":4,
            "COLOR":4,"SOFT LIGHT":4,"LINEAR LIGHT":4});
        return TexOpDict[self.GetName()];
    return 1;
CBlendMode.ToD3DTexOp=CBlendModeToD3DTexOp;


# class CMaterial
def CMaterialSaveData(self,File,UseV2):
    """This function writes the collected data to the given file."""
    # Write the null-terminated the chunk header or the material identifier
    if(UseV2):
        File.write(PackString("$U3D_MATERIAL"));
        ChunkSize=CChunkSizeWriter(File);
        # Write the material index
        File.write(pack("L",self.iMaterial));
    else:
        File.write(PackString("$U3D_MAT"));
    # Write the null-terminated material name
    File.write(PackString(self.Name));
    # Write default material colors
    self.AmbientColor.SaveData(File);
    self.DiffuseColor.SaveData(File);
    self.SpecularColor.SaveData(File);
    self.EmissiveColor.SaveData(File);
    # Write the specular power
    File.write(pack("f",self.SpecularPower));
    # Write the parallax information
    if(UseV2):
        File.write(pack("ff",0.0,1.0));
    # Write the number of textures
    if(UseV2==False):
        self.nTexture=min(8,self.nTexture);
        File.write(pack("H",self.nTexture));
    # Write default texture operations
    pTextureOperation=[1,1,1,1,1,1,1,1];
    if(self.nTexture>0):
        pTextureOperation[0]=5;
    for i in range(1,self.nTexture):
        pTextureOperation[i]=self.pBlendMode[i].ToD3DTexOp();
    for i in range(0,8):
        File.write(pack("L",pTextureOperation[i]));
    # Write texture coordinate sets
    for i in range(0,self.nTexture):
        File.write(pack("L",self.piTexCoordSet[i]));
    for i in range(self.nTexture,8):
        File.write(pack("L",0));
    # Write whether the resources are part of the file (false)
    if(UseV2==False):
        File.write(pack("b",0));
    # Write the texture data
    if(UseV2):
        for i in range(8):
            if(i<len(self.pTextureFileName)):
                TextureFile=self.pTextureFileName[i];
                # Write a texture chunk header
                File.write(PackString("$U3D_TEXTURE"));
                ChunkSize2=CChunkSizeWriter(File);
                # Write that a texture is held
                File.write(pack("b",True));
                # Write its dimensions and its type
                File.write(pack("LL",0,0));
                File.write(pack("bb",False,False));
                # Write the height scalar (0.0)
                File.write(pack("f",0.0));
                # Write the file name
                File.write(PackString("*"+TextureFile));
                # The chunk is completed
                ChunkSize2.Write();
            else:
                # Write a header and a False after that
                File.write(PackString("$U3D_TEXTURE"));
                ChunkSize2=CChunkSizeWriter(File);
                File.write(pack("b",False));
                ChunkSize2.Write();
    else:
        nWrittenTexture=0;
        for TextureFile in self.pTextureFileName:
            if(nWrittenTexture==8):
                break;
            File.write(pack("b",True));
            File.write(pack("I",0));
            File.write(PackString(TextureFile));
            File.write(pack("H",0));
            File.write(pack("H",0));
            File.write(PackString(TextureFile));
            nWrittenTexture+=1;
        for i in range(nWrittenTexture,8):
            File.write(pack("b",0));
    # Write that no effect is being used
    File.write(pack("b",0));
    # End the chunk
    if(UseV2):
        ChunkSize.Write();
CMaterial.SaveData=CMaterialSaveData;


def CKeySaveBaseData(self,File,UseV2,oFrame=0.0):
    """Write data of the base part of this key to the given file."""
    if(UseV2):
        File.write(pack("L",int(self.Frame+oFrame)));
    else:
        File.write(pack("H",int(self.Frame+oFrame)));      
        File.write(pack("H",0));
CKey.SaveBaseData=CKeySaveBaseData;

def ScalingKeySaveData(self,File,UseV2,oFrame=0.0):
    """Write the data of this key to the given file."""
    self.SaveBaseData(File,UseV2,oFrame);
    self.Scaling.SaveData(File);
    #**#C3DVector(1.0,1.0,1.0).SaveData(File);
CScalingKey.SaveData=ScalingKeySaveData;

def TranslationKeySaveData(self,File,UseV2,oFrame=0.0):
    """Write the data of this key to the given file."""
    self.SaveBaseData(File,UseV2,oFrame);
    self.Translation.SaveData(File);
    #**#C3DVector(0.0,0.0,0.0).SaveData(File);
CTranslationKey.SaveData=TranslationKeySaveData;

def RotationKeySaveData(self,File,UseV2,oFrame=0.0):
    """Write the data of this key to the given file."""
    self.SaveBaseData(File,UseV2,oFrame);
    self.Rotation.SaveData(File);
    #**#CQuaternion(0.0,0.0,0.0,1.0).SaveData(File);
CRotationKey.SaveData=RotationKeySaveData;


def CKeyedTransformationSaveData(self,File,UseV2,oFrame=0.0):
    """This function saves the keys of this keyed transformation to the given file
       using either U3D 2.x or 3.x format. The given offset is applied to all written
       frame values."""
    # Write the scaling keys
    File.write(pack(Choose(UseV2,"L","H"),len(self.pScalingKey)));
    for ScalingKey in self.pScalingKey:
        ScalingKey.SaveData(File,UseV2,oFrame);
    # Write the translation keys
    File.write(pack(Choose(UseV2,"L","H"),len(self.pTranslationKey)));
    for TranslationKey in self.pTranslationKey:
        TranslationKey.SaveData(File,UseV2,oFrame);
    # Write the rotation keys
    File.write(pack(Choose(UseV2,"L","H"),len(self.pRotationKey)));
    for RotationKey in self.pRotationKey:
        RotationKey.SaveData(File,UseV2,oFrame);
CKeyedTransformation.SaveData=CKeyedTransformationSaveData;


def CBoneSaveData(self,File,UseV2):
    """This function saves all data of this bone to the given file."""
    # Write the chunk header or the null-terminated bone chunk identifier
    ChunkSize=None;
    if(UseV2):
        File.write(PackString("$U3D_BONE"));
        ChunkSize=CChunkSizeWriter(File);
        # Write the bone index
        File.write(pack("L",self.iBone));
    else:
        File.write(PackString("$U3D_BONE"));
    # Write the null-terminated name of this bone
    File.write(PackString(self.Name));
    # Write the index of the parent of this bone
    iParent=65535;
    if(self.Parent!=None):
        iParent=self.Parent.iBone;
    File.write(pack(Choose(UseV2,"L","H"),iParent));
    # Write the bone frame offset (-1.0) and whether this is to effect the
    # child frame
    File.write(pack("f",-1.0));
    File.write(pack("B",0));
    # Write the number of meshes that are appended to this bone
    File.write(pack(Choose(UseV2,"L","H"),1));
    # Write the index of the skinning mesh
    File.write(pack(Choose(UseV2,"L","H"),0));
    # Write the mesh transformation
    self.MeshTransformation.SaveData(File);
    # Output the key frames
    self.KeyDict[""].SaveData(File,UseV2);
    # The chunk is completed
    if(UseV2):
        ChunkSize.Write();
CBone.SaveData=CBoneSaveData;


def CSkeletonSaveData(self,File,UseV2):
    """Saves the skeleton to the given file and writes an action range chunk."""
    ActionRangeList=list();
    iFirstFrame=0.0;
    for Action in self.ActionList:
        (MinFrame,MaxFrame)=self.GetAnimationRange(Action);
        nFrame=1;
        if(MinFrame!=None):
            nFrame=1.0+MaxFrame-MinFrame;
        ActionRangeList.append((Action,int(iFirstFrame),int(nFrame)));
        iFirstFrame+=nFrame;
    for Bone in self.BoneList:
        Bone.SaveData(File,UseV2);
    if(UseV2):
        File.write(PackString("$U3D_ACTION_RANGE"));
        ChunkSize=CChunkSizeWriter(File);
        File.write(pack("L",len(self.ActionList)));
        for ActionRange in ActionRangeList:
            File.write(PackString(ActionRange[0]));
            File.write(pack("L",int(ActionRange[1])));
            File.write(pack("L",int(ActionRange[1]+ActionRange[2]-1)));
        ChunkSize.Write();
CSkeleton.SaveData=CSkeletonSaveData;


class CExportPlugin:
    """This class does the entire export."""

    def __init__(self,BlenderUtility):
        """This constructor prepares the plugin for exporting model data. To do 
           so it needs an instance of a derived class of CBlenderUtilityBase."""
        self.BlenderUtility=BlenderUtility;

    def ExportModel(self,FileName,Options,ExportActionList,RemoveUnusedBones,MergeMeshes,FrameFactor,UseV2):
        """This function initializes all processes that are necessary to export
           the model to a given file. For this purpose it uses the given 
           options."""
        print("");
        print("*** Exporting to an Ultimate 3D model file.***");
        
        self.BlenderUtility.PrepareExport(Options.ExportAnimations and Options.ExportConstraints);

        # Set up the right Blender to output space. It needs to swap the y- and the
        # z-coordinate.
        BlenderToOutputSpace=CMatrix4x4(1.0,0.0,0.0,0.0,
                                        0.0,0.0,1.0,0.0,
                                        0.0,1.0,0.0,0.0,
                                        0.0,0.0,0.0,1.0);
        # Read all data of the model
        ModelData=CModelData(Options,self.BlenderUtility);
        Model=CModel(ModelData,0.7);
        if(not Model.IsValid()):
            self.BlenderUtility.ShowMessage("No data for export read.","Error");
            return False;
        if(Model.IsValid() and not Model.HasGeometry()):
            self.BlenderUtility.ShowMessage("The read model does not contain any geometry.","Error");
            return False;

        # Update the progress bar
        self.BlenderUtility.SetProgress(0.7,"Post-processing");
        # If requested unused bones get eliminated now
        if(RemoveUnusedBones):
            for Skeleton in Model.pSkeleton:
                Skeleton.RemoveUnusedBones(Model,0.05);
        # If there is a skeleton or if it is requested, the model needs to be 
        # simplified
        if(len(Model.pSkeleton)>0):
            Model.SimplifyModel();
            # Ensure that there are always exactly four skin weights
            Model.pMesh[0].LimitWeightCount(4,4);
        elif(MergeMeshes):
            Model.SimplifyModel();
        else:
            # Check, if there are object animations
            ObjectAnimationsUsed=False;
            for Object in Model.pObject:
                ObjectAnimationsUsed=ObjectAnimationsUsed or Object.IsAnimated();
            # If so the model needs to be simplified to include them
            if(ObjectAnimationsUsed):
                Model.SimplifyModel();
        # Transform the model using the Blender to output space
        Model.ApplyTransformation(BlenderToOutputSpace);
        # Applying transformations always makes export easier
        if(len(Model.pSkeleton)==0):
            Model.ApplyTransformations();
        else:
            Model.pSkeleton[0].ApplyTransformation();
            Model.pMesh[0].FlipTriangles();

        if(len(Model.pSkeleton)>0):
            # Ensure that every vertex has exactly four weights, if there is a 
            # skeleton
            for Mesh in Model.pMesh:
                Mesh.LimitWeightCount(4,4);
            # Initialize a text file to store the action list
            ActionOffsetDict=None;
            if(not Options.ExportCurrentNLA and Options.ExportAnimations):
                ActionListFileName=FileName[0:max(0,len(FileName)-4)]+".txt";
                try:
                    ActionList=open(ActionListFileName,"w");
                except:
                    print("Failed to open the file for the action list: "+ActionListFileName);
                    ActionList=None;
                print("Action ranges:");
                # Determine an appropriate frame offset for every action
                ActionOffsetDict=dict();
                ActionMinFrameDict=dict();
                ActionMaxFrameDict=dict();
                CurrentNextFrame=0.0;
                for ActionName in ModelData.GetActionNameList():
                    # Get the range of the action
                    ActionMinFrame,ActionMaxFrame=ModelData.GetActionRange(ActionName);
                    FrameNumbers=ModelData.GetActionFrameList(ActionName,0.0,ActionMinFrame);
                    ActionMinFrame=min(ActionMinFrame,min(FrameNumbers));
                    ActionMaxFrame=max(ActionMaxFrame,max(FrameNumbers));
                    # Compute a corresponding offset and update the next frame value
                    ActionOffset=CurrentNextFrame-ActionMinFrame;
                    CurrentNextFrame+=1.0+ActionMaxFrame-ActionMinFrame;
                    # Store the results
                    ActionMinFrameDict[ActionName]=ActionMinFrame;
                    ActionMaxFrameDict[ActionName]=ActionMaxFrame;
                    ActionOffsetDict[ActionName]=ActionOffset;
                    # Write them to the action list
                    StartFrame=ActionMinFrame+ActionOffset;
                    EndFrame=ActionMaxFrame+ActionOffset;
                    ActionData="\""+ActionName+"\" "+str(StartFrame)+" "+str(EndFrame);
                    if(ActionList!=None):
                        ActionList.write(ActionData+"\r\n");
                    print(ActionData);
                # The action list is complete
                if(ActionList!=None):
                    ActionList.close();
                    ActionList=None;
            # Combine all actions
            for Bone in Model.pSkeleton[0].BoneList:
                Bone.KeyDict.CombineTransformations(ActionOffsetDict);

        # Check whether the number of bones is valid
        BoneList=[];
        if(len(Model.pSkeleton)>0):
            BoneList=Model.pSkeleton[0].BoneList;
        if(len(BoneList)>=256):
            self.BlenderUtility.ShowMessage(
                "Your scene contains more than 255 bones ("+str(len(BoneList))+"). Since Ultimate 3D and the Ultimate 3D model file format can't handle that many bones in combination with skinning you have to reduce the number of bones. The exporter will terminate now. No file has been created.",
                "Error","Maximum of 255 bones exceeded with "+str(len(BoneList))+" bones.");
            return False;
        
        # Change the animation speed, if requested
        if(FrameFactor!=1.0):
            Model.TransformTime(0.0,FrameFactor);

        # Try to open the file
        File=None;
        Failure=None;
        try:
            File=open(FileName,"wb");
        except Exception as ThrownException:
            File=None;
            Failure=ThrownException;
        if(File==None):
            ExceptionString="";
            if(Failure!=None):
                ExceptionString="\nThe exception is described as:\n"+str(Failure);
            self.BlenderUtility.ShowMessage(
                "Failed to open the file \""+FileName+"\". Please make sure that you have write access and try again."+ExceptionString,
                "Error","Failed to open \""+FileName+"\".");
            return False;
        print("Opened the file "+FileName+" successfully.");

        # Update the progress bar
        self.BlenderUtility.SetProgress(0.8,"Saving data");
        # Export the model header data
        Model.SaveData(File,UseV2);

        # Export the meshes
        for ReadMesh in Model.pMesh:
            ReadMesh.SaveData(File,UseV2,self.BlenderUtility);

        # Export the material list
        for ReadMaterial in Model.pMaterial:
            ReadMaterial.SaveData(File,UseV2);

        # Export the skeleton
        if(len(Model.pSkeleton)>0):
            Model.pSkeleton[0].SaveData(File,UseV2);

        # Close the file
        File.close();

        # Update the progress bar and indicate success
        self.BlenderUtility.SetProgress(1.0,"Done");
        self.BlenderUtility.ShowMessage("The model has been exported successfully.","Success");
        self.BlenderUtility.FinalizeExport();

        return True;


# export_u3d_shared END



InputWindow=None;
def FileSelectorCallback(FileName):
    """This is the callback function for the file selector."""
    global InputWindow;
    ExportPlugin=InputWindow.ExportPlugin;
    Input=InputWindow;
    if(len(FileName)>3):
        ExportPlugin.ExportModel(FileName,
                CModelOptions(
                    Input.ExportSelectionOnly.val,
                    Input.ExportAnimations.val,
                    Input.ExportConstraints.val,
                    False,
                    Input.MaxKeyDistance.val,
                    Input.ExportAllUVLayers.val,
                    Input.ReadModifiers.val,
                    Input.RenderModeModifiers.val),
                Input.ExportActionList.val,
                Input.RemoveUnusedBones.val,
                Input.MergeMeshes.val,
                Input.FrameFactor.val,
                Input.UseV2.val);


class CInputWindow:
    """This class creates a window in which the user has to enter
       information about how he wants the file to be exported."""

    def __init__(self,ExportPlugin):
        """This function creates a user interface to get the input."""
        global InputWindow;
        self.ExportPlugin=ExportPlugin;
        # Display a neat little pop up with the options
        FormContent=[];
        FormContent.append(("Selection only",self.ExportSelectionOnly,"If checked only selected objects will be exported, otherwise the whole scene will be exported."));
        FormContent.append(("Animations",self.ExportAnimations,"Export armatures, skin weights and actions?"));
        FormContent.append(("Constraints",self.ExportConstraints,"Cycle through all frames to export constraints?"));
        FormContent.append(("Action list",self.ExportActionList,"Export a list of all actions with their corresponding frame values?"));
        FormContent.append(("All UV textures",self.ExportAllUVLayers,"Export all UV textures or only the active layers. If enabled all meshes must have the same UV layers with the same names."));
        FormContent.append(("Max key dist",self.MaxKeyDistance,0.0,100.0,"The maximal distance between the frame values of two neighboring keys or 0 for no maximum."));
        FormContent.append(("Eliminate bones",self.RemoveUnusedBones,"If this is checked the plugin will search for bones, which do not effect any geometry to remove them."));
        FormContent.append(("Modifiers",self.ReadModifiers,"If this option is enabled modifiers are used and curves, texts, meta balls and surface objects can be exported."));
        FormContent.append(("Render settings",self.RenderModeModifiers,"If this option is enabled modifiers get exported using the render mode settings."));
        FormContent.append(("Merge meshes",self.MergeMeshes,"If this option is enabled all meshes get combined into a single mesh automatically."));
        FormContent.append(("Frame factor",self.FrameFactor,0.0,100.0,"This option defines a factor for the speed of the animation."));
        FormContent.append(("Use v2.0 (U3D 3.0)",self.UseV2,"Use format version 2.0 instead of version 1.0? Version 2.0 is new in Ultimate 3D 3.0."));
        PopUpReturnValue=Draw.PupBlock("Ultimate 3D export settings",FormContent);
        # If OK has been clicked show the file selection dialog
        if(PopUpReturnValue):
            # Suggest a filename that makes sense
            ExportFileSuggestion=Blender.Get("filename");
            ExportFileSuggestion=ExportFileSuggestion[0:ExportFileSuggestion.rfind(".")+1]+"u3d";
            if(len(ExportFileSuggestion)==3):
                ExportFileSuggestion="Model.u3d";
            InputWindow=self;
            Window.FileSelector(FileSelectorCallback,"Export *.u3d model",ExportFileSuggestion);

    # The values for the buttons
    ExportSelectionOnly=Draw.Create(False);
    ExportAnimations=Draw.Create(True);
    ExportConstraints=Draw.Create(True);
    ExportActionList=Draw.Create(False);
    ExportAllUVLayers=Draw.Create(False);
    MaxKeyDistance=Draw.Create(0.0);
    RemoveUnusedBones=Draw.Create(False);
    ReadModifiers=Draw.Create(True);
    RenderModeModifiers=Draw.Create(False);
    MergeMeshes=Draw.Create(False);
    FrameFactor=Draw.Create(1.0);
    UseV2=Draw.Create(False);

# Create an input window
CInputWindow(CExportPlugin(CBlenderUtility()));
